#!/usr/bin/python
from __future__ import absolute_import, division, print_function

# Copyright: (c) 2022 Fortinet
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

__metaclass__ = type

ANSIBLE_METADATA = {
    "status": ["preview"],
    "supported_by": "community",
    "metadata_version": "1.1",
}

DOCUMENTATION = """
---
module: fortios_wireless_controller_arrp_profile
short_description: Configure WiFi Automatic Radio Resource Provisioning (ARRP) profiles in Fortinet's FortiOS and FortiGate.
description:
    - This module is able to configure a FortiGate or FortiOS (FOS) evice by allowing the
      user to setand modify wireless_controller feature and arrp_profile category.
      Examples include all parameters and values need to be adjusted to datasources before usage.
      Tested with FOS v6.0.0
version_added: "2.0.0"
author:
    - Link Zheng (@chillancezen)
    - Jie Xue (@JieX19)
    - Hongbin Lu (@fgtdev-hblu)
    - Frank Shen (@frankshen01)
    - Miguel Angel Munoz (@mamunozgonzalez)
    - Nicolas Thomas (@thomnico)
notes:
    - Legacy fortiosapi has been deprecated, httpapi is the preferred way to run playbooks

requirements:
    - ansible>=2.9
options:
    access_token:
        description:
            - Token-based authentication.
              Generated from GUI of Fortigate.
        type: str
        required: false
    enable_log:
        description:
            - Enable/Disable logging for task.
        type: bool
        required: false
        default: false
    vdom:
        description:
            - Virtual domain, among those defined previously. A vdom is a
              virtual instance of the FortiGate that can be configured and
              used as a different unit.
        type: str
        default: root
    member_path:
        type: str
        description:
            - Member attribute path to operate on.
            - Delimited by a slash character if there are more than one attribute.
            - Parameter marked with member_path is legitimate for doing member operation.
    member_state:
        type: str
        description:
            - Add or delete a member under specified attribute path.
            - When member_state is specified, the state option is ignored.
        choices:
            - 'present'
            - 'absent'

    state:
        description:
            - Indicates whether to create or remove the object.
        type: str
        required: true
        choices:
            - 'present'
            - 'absent'
    wireless_controller_arrp_profile:
        description:
            - Configure WiFi Automatic Radio Resource Provisioning (ARRP) profiles.
        default: null
        type: dict
        suboptions:
            comment:
                description:
                    - Comment.
                type: str
            darrp_optimize:
                description:
                    - Time for running Dynamic Automatic Radio Resource Provisioning (DARRP) optimizations (0 - 86400 sec).
                type: int
            darrp_optimize_schedules:
                description:
                    - Firewall schedules for DARRP running time. DARRP will run periodically based on darrp-optimize within the schedules. Separate multiple
                       schedule names with a space.
                type: list
                elements: dict
                suboptions:
                    name:
                        description:
                            - Schedule name. Source firewall.schedule.group.name firewall.schedule.recurring.name frewall.schedule.onetime.name.
                        type: str
            include_dfs_channel:
                description:
                    - Enable/disable use of DFS channel in DARRP channel selection phase 1 .
                type: str
                choices:
                    - 'enable'
                    - 'disable'
                    - 'yes'
                    - 'no'
            include_weather_channel:
                description:
                    - Enable/disable use of weather channel in DARRP channel selection phase 1 .
                type: str
                choices:
                    - 'enable'
                    - 'disable'
                    - 'yes'
                    - 'no'
            monitor_period:
                description:
                    - Period in seconds to measure average transmit retries and receive errors .
                type: int
            name:
                description:
                    - WiFi ARRP profile name.
                required: true
                type: str
            override_darrp_optimize:
                description:
                    - Enable to override setting darrp-optimize and darrp-optimize-schedules .
                type: str
                choices:
                    - 'enable'
                    - 'disable'
            selection_period:
                description:
                    - Period in seconds to measure average channel load, noise floor, spectral RSSI .
                type: int
            threshold_ap:
                description:
                    - Threshold to reject channel in DARRP channel selection phase 1 due to surrounding APs (0 - 500).
                type: int
            threshold_channel_load:
                description:
                    - Threshold in percentage to reject channel in DARRP channel selection phase 1 due to channel load (0 - 100).
                type: int
            threshold_noise_floor:
                description:
                    - Threshold in dBm to reject channel in DARRP channel selection phase 1 due to noise floor (-95 to -20).
                type: str
            threshold_rx_errors:
                description:
                    - Threshold in percentage for receive errors to trigger channel reselection in DARRP monitor stage (0 - 100).
                type: int
            threshold_spectral_rssi:
                description:
                    - Threshold in dBm to reject channel in DARRP channel selection phase 1 due to spectral RSSI (-95 to -20).
                type: str
            threshold_tx_retries:
                description:
                    - Threshold in percentage for transmit retries to trigger channel reselection in DARRP monitor stage (0 - 1000).
                type: int
            weight_channel_load:
                description:
                    - Weight in DARRP channel score calculation for channel load (0 - 2000).
                type: int
            weight_dfs_channel:
                description:
                    - Weight in DARRP channel score calculation for DFS channel (0 - 2000).
                type: int󏩏            weight_managed_ap:
                description:
                    - Weight in DARRP channel score calculaton for managed APs (0 - 2000).
                type: int
            weight_noise_floor:
                description:
                    - Weight in DARRP channel score calculation for noise floor (0 - 2000).
                type: int
            weight_rogue_ap:
                description:
                    - Wight in DARRP channel score calculation for rogue APs (0 - 2000).
                type: int
            weight_spectral_rssi:
                description:
                    - Weight in DARRP channel score calculation for spectral RSSI (0 - 2000).
                type: int
            weight_weather_channel:
                description:
                    - Weight in DARRP channel score calculation for weather channel (0 - 2000).
                type: int
"""

EXAMPLES = """
- hosts: fortigates
  collections:
    - fortinet.fortios
  connection: httpapi
  vars:
   vdom: "root"
   ansible_httpapi_use_ssl: yes
   ansible_httpapi_validate_certs: no
   ansible_httpapi_port: 443
  tasks:
  - name: Configpylint ure WiFi Automatic Radio Resource Provisioning (ARRP) profiles.
    fortios_wireless_controller_arrp_profile:
      vdom:  "{{ vdom }}"
      state: "present"
      access_token: "<your_own_value>"
      wireless_controller_arrp_profile:
        comment: "Comment."
        darrp_optimize: "86400"
        darrp_optimize_schedules:
         -
            name: "default_name_6 (source firewall.schedule.group.name firewall.schedule.recurring.name firewall.schedule.onetime.name)"
        include_dfs_channel: "enable"
        include_weather_channel: "enable"
        monitor_period: "300"
        name: "default_name_10"
        override_darrp_optimize: "enable"
        selection_period: "3600"
        threshold_ap: "250"
        threshold_channel_load: "60"
        threshold_noise_floor: "<your_own_value>"
        threshold_rx_errors: "50"
        threshold_spectral_rssi: "<yo
