"""
    babel.messages.frontend
    ~~~~~~~~~~~~~~~~~~~~~~~

    Frontends for the message extraction functionality.

    :copyright: (c) 2013-2022 by the Babel Team.
    :license: BSD, see LICENSE for more details.
"""

import fnmatch
import logging
import optparse
import os
import re
import shutil
import sys
import tempfile
from collections import OrderedDict
from configparser import RawConfigParser
from datetime import datetime
from io import StringIO

from babel import __version__ as VERSION
from babel import Locale, localedata
from babel.core import UnknownLocaleError
from babel.messages.catalog import Catalog, DEFAULT_HEADER
from babel.messages.extract import DEFAULT_KEYWORDS, DEFAULT_MAPPING, check_and_call_extract_file, extract_from_dir
from babel.messages.mofile import write_mo
from babel.messages.pofile import read_po, write_po
from babel.util import LOCALTZ

log = logging.getLogger("babel")

try:
    # See: https://setuptools.pypa.io/en/latest/deprecated/distutils-legacy.html
    from setuptools import Command as _Command
    distutils_log = log  # "distutils.log → (no replacement yet)"

    try:
        from setuptools.errors import OptionError, SetupError, BaseError
    except ImportError:  # Error aliases only added in setuptools 59 (2021-11).
        OptionError = SetupError = BaseError = Exception

except ImportError:
    from distutils import log as distutils_log
    from distutils.cmd import Command as _Command
    from distutils.errors import DistutilsOptionError as OptionError, DistutilsSetupError as SetupError, DistutilsError as BaseError



def listify_value(arg, split=None):
    """
    Make a list out of an argument.

    Values from `distutils` argument parsing are always single strings;
    values from `optparse` parsing may be lists of strings that may need
    to be further split.

    No matter the input, this function returns a flat list of whitespace-trimmed
    strings, with `None` values filtered out.

    >>> listify_value("foo bar")
    ['foo', 'bar']
    >>> listify_value(["foo bar"])
    ['foo', 'bar']
    >>> listify_value([["foo"], "bar"])
    ['foo', 'bar']
    >>> listify_value([["foo"], ["bar", None, "foo"]])
    ['foo', 'bar', 'foo']
    >>> listify_value("foo, bar, quux", ",")
    ['foo', 'bar', 'quux']

    :param arg: A string or a list of strings
    :param split: The argument to pass to `str.split()`.
    :return:
    """
    out = []

    if not isinstance(arg, (list, tuple)):
        arg = [arg]

    for val in arg:
        if val is None:
            continue
        if isinstance(val, (list, tuple)):
            out.extend(listify_value(val, split=split))
            continue
        out.extend(s.strip() for s in str(val).split(split))
    assert all(isinstance(val, str) for val in out)
    return out


class Command(_Command):
    # This class is a small shim between Distutils commands and
    # optparse option parsing in the frontend command line.

    #: Option name to be input as `args` on the script command line.
    as_args = None

    #: Options which allow multiple values.
    #: This is used by the `optparse` transmogrification code.
    multiple_value_options = ()

    #: Options which are booleans.
    #: This is used by the `optparse` transmogrification code.
    # (This is actually used by distutils code too, but is never
    # declared in the base class.)
    boolean_options = ()

    #: Option aliases, to retain standalone command compatibility.
    #: Distutils does not support option aliases, but optparse does.
    #: This maps the distutils argument name to an iterable of aliases
    #: that are usable with optparse.
    option_aliases = {}

    #: Choices for options that needed to be restricted to specific
    #: list of choices.
    option_choices = {}

    #: Log object. To allow replacement in the script command line runner.
    log = distutils_log

    def __init__(self, dist=None):
        # A less strict version of distutils' `__init__`.
        self.distribution = dist
        self.initialize_options()
        self._dry_run = None
        self.verbose = False
        self.force = None
        self.help = 0
        self.finalized = 0


class compile_catalog(Command):
    """Catalog compilation command for use in ``setup.py`` scripts.

    If correctly installed, this command is available to Setuptools-using
    setup scripts automatically. For projects using plain old ``distutils``,
    the command needs to be registered explicitly in ``setup.py``::

        from babel.messages.frontend import compile_catalog

        setup(
            ...
            cmdclass = {'compile_catalog': compile_catalog}
        )

    .. versionadded:: 0.9
    """

    description = "compile message catalogs to binary MO files"
    user_options = [
        ("domain=", "D",
         "domains of PO files (space separated list, default 'messages')"),
        ("directory=", "d",
         "path to base directory containing the catalogs"),
        ("input-file=", "i",
         "name of the input file"),
        ("output-file=", "o",
         "name of the output file (default "
         "'<output_dir>/<locale>/LC_MESSAGES/<domain>.mo')"),
        ("locale=", "l",
         "locale of the catalog to compile"),
        ("use-fuzzy", "f",
         "also include fuzzy translations"),
        ("statistics", None,
         "print statistics about translations"),
    ]
    boolean_options = ["use-fuzzy", "statistics"]

    def initialize_options(self):
        self.domain = "messages"
        self.directory = None
        self.input_file = None
        self.output_file = None
        self.locale = None
        self.use_fuzzy = False
        self.statistics = False

    def finalize_options(self):
        self.domain = listify_value(self.domain)
        if not self.input_file and not self.directory:
            raise OptionError("you must specify either the input file or the base directory")
        if not self.output_file and not self.directory:
            raise OptionError("you must specify either the output file or the base directory")

    def run(self):
        n_errors = 0
        for domain in self.domain:
            for catalog, errors in self._run_domain(domain).items():
                n_errors += len(errors)
        if n_errors:
            self.log.error("%d errors encountered." % n_errors)
        return (1 if n_errors else 0)

    def _run_domain(self, domain):
        po_files = []
        mo_files = []

        if not self.input_file:
            if self.locale:
                po_files.append((self.locale,
                                 os.path.join(self.directory, self.locale,
                                              "LC_MESSAGES",
                                              domain + ".po")))
                mo_files.append(os.path.join(self.directory, self.locale,
                                             "LC_MESSAGES",
                                             domain + ".mo"))
            else:
                for locale in os.listdir(self.directory):
                    po_file = os.path.join(self.directory, locale,
                                           "LC_MESSAGES", domain + ".po")
                    if os.path.exists(po_file):
                        po_files.append((locale, po_file))
                        mo_files.append(os.path.join(self.directory, locale,
                                                     "LC_MESSAGES",
                                                     domain + ".mo"))
        else:
            po_files.append((self.locale, self.input_file))
            if self.output_file:
                mo_files.append(self.output_file)
            else:
                mo_files.append(os.path.join(self.directory, self.locale,
                                             "LC_MESSAGES",
                                             domain + ".mo"))

        if not po_files:
            raise OptionError("no message catalogs found")

        catalogs_and_errors = {}

        for idx, (locale, po_file) in enumerate(po_files):
            mo_file = mo_files[idx]
            with open(po_file, "rb") as infile:
                catalog = read_po(infile, locale)

            if self.statistics:
                translated = 0
                for message in list(catalog)[1:]:
                    if message.string:
                        translated += 1
                percentage = 0
                if len(catalog):
                    percentage = translated * 100 // len(catalog)
                self.log.info(
                    "%d of %d messages (%d%%) translated in %s",
                    translated, len(catalog), percentage, po_file,
                )

            if catalog.fuzzy and not self.use_fuzzy:
                self.log.info("catalog %s is marked as fuzzy, skipping", po_file)
                continue

            catalogs_and_errors[catalog] = catalog_errors = list(catalog.check())
            for message, errors in catalog_errors:
                for error in errors:
                    self.log.error(
                        "error: %s:%d: %s", po_file, message.lineno, error,
                    )

            self.log.info("compiling catalog %s to %s", po_file, mo_file)

            with open(mo_file, "wb") as outfile:
                write_mo(outfile, catalog, use_fuzzy=self.use_fuzzy)

        return catalogs_and_errors


def _make_directory_filter(ignore_patterns):
    """
    Build a directory_filter function based on a list of ignore patterns.
    """
    def cli_directory_filter(dirname):
        basename = os.path.basename(dirname)
        return not any(
            fnmatch.fnmatch(basename, ignore_pattern)
            for ignore_pattern
            in ignore_patterns
        )
    return cli_directory_filter


class extract_messages(Command):
    """Message extraction command for use in ``setup.py`` scripts.

    If correctly installed, this command is available to Setuptools-using
    setup scripts automatically. For projects using plain old ``distutils``,
    the command needs to be registered explicitly in ``setup.py``::

        from babel.messages.frontend import extract_messages

        setup(
            ...
            cmdclass = {'extract_messages': extract_messages}
        )
    """

    description = "extract localizable strings from the project code"
    user_options = [
        ("charset=", None,
         'charset to use in the output file (default "utf-8")'),
        ("keywords=", "k",
         "space-separated list of keywords to look for in addition to the "
         "defaults (may be repeated multiple times)"),
        ("no-default-keywords", None,
         "do not include the default keywords"),
        ("mapping-file=", "F",
         "path to the mapping configuration file"),
        ("no-location", None,
         "do not include location comments with filename and line number"),
        ("add-location=", None,
         'location lines format. If it is not given or "full", it generates '
         'the lines with both file name and line number. If it is "file", '
         'the line number part is omitted. If it is "never", it completely '
         'suppresses the lines (same as --no-location).'),
        ("omit-header", None,
         'do not include msgid "" entry in header'),
        ("output-file=", "o",
         "name of the output file"),
        ("width=", "w",
         "set output line width (default 76)"),
        ("no-wrap", None,
         "do not break long message lines, longer than the output line width, "
         "into several lines"),
        ("sort-output", None,
         "generate sorted output (default False)"),
        ("sort-by-file", None,
         "sort output by file location (default False)"),
        ("msgid-bugs-address=", None,
         "set report address for msgid"),
        ("copyright-holder=", None,
         "set copyright holder in output"),
        ("project=", None,
         "set project name in output"),
        ("version=", None,
         "set project version in output"),
        ("add-comments=", "c",
         "place comment block with TAG (or those preceding keyword lines) in "
         "output file. Separate multiple TAGs with commas(,)"),  # TODO: Support repetition of this argument
        ("strip-comments", "s",
         "strip the comment TAGs from the comments."),
        ("input-paths=", None,
         "files or directories that should be scanned for messages. Separate multiple "
         "files or directories with commas(,)"),  # TODO: Support repetition of this argument
        ("input-dirs=", None,  # TODO (3.x): Remove me.
         "alias for input-paths (does allow files as well as directories)."),
        ("ignore-dirs=", None,
         'Patterns for directories to ignore when scanning for messages. '
         'Separate multiple patterns with spaces (default ".* ._")'),
        ("header-comment=", None,
         "header comment for the catalog"),
    ]
    boolean_options = [
        "no-default-keywords", "no-location", "omit-header", "no-wrap",
        "sort-output", "sort-by-file", "strip-comments",
    ]
    as_args = "input-paths"
    multiple_value_options = (
        "add-comments",
        "keywords",
        "ignore-dirs",
    )
    option_aliases = {
        "keywords": ("--keyword",),
        "mapping-file": ("--mapping",),
        "output-file": ("--output",),
        "strip-comments": ("--strip-comment-tags",),
    }
    option_choices = {
        "add-location": ("full", "file", "never"),
    }

    def initialize_options(self):
        self.charset = "utf-8"
        self.keywords = None
        self.no_default_keywords = False
        self.mapping_file = None
        self.no_location = False
        self.add_location = None
        self.omit_header = False
        self.output_file = None
        self.input_dirs = None
        self.input_paths = None
        self.width = None
        self.no_wrap = False
        self.sort_output = False
        self.sort_by_file = False
        self.msgid_bugs_address = None
        self.copyright_holder = None
        self.project = None
        self.version = None
        self.add_comments = None
        self.strip_comments = False
        self.include_lineno = True
        self.ignore_dirs = None
        self.header_comment = None

    def finalize_options(self):
        if self.input_dirs:
            if not self.input_paths:
                self.input_paths = self.input_dirs
            else:
                raise OptionError(
                    "input-dirs and input-paths are mutually exclusive",
                )

        if self.no_default_keywords:
            keywords = {}
        else:
            keywords = DEFAULT_KEYWORDS.copy()

        keywords.update(parse_keywords(listify_value(self.keywords)))

        self.keywords = keywords

        if not self.keywords:
            raise OptionError(
                "you must specify new keywords if you disable the default ones",
            )

        if not self.output_file:
            raise OptionError("no output file specified")
        if self.no_wrap and self.width:
            raise OptionError(
                "'--no-wrap' and '--width' are mutually exclusive",
            )
        if not self.no_wrap and not self.width:
            self.width = 76
        elif self.width is not None:
            self.width = int(self.width)

        if self.sort_output and self.sort_by_file:
            raise OptionError(
                "'--sort-output' and '--sort-by-file' are mutually exclusive",
            )

        if self.input_paths:
            if isinstance(self.input_paths, str):
                self.input_paths = re.split(r",\s*", self.input_paths)
        elif self.distribution is not None:
            self.input_paths = dict.fromkeys([
                k.split(".", 1)[0]
                for k in (self.distribution.packages or ())
            ]).keys()
        else:
            self.input_paths = []

        if not self.input_paths:
            raise OptionError("no input files or directories specified")

        for path in self.input_paths:
            if not os.path.exists(path):
                raise OptionError("Input path: %s does not exist" % path)

        self.add_comments = listify_value(self.add_comments or (), ",")

        if self.distribution:
            if not self.project:
                self.project = self.distribution.get_name()
            if not self.version:
                self.version = self.distribution.get_version()

        if self.add_location == "never":
            self.no_location = True
        elif self.add_location == "file":
            self.include_lineno = False

        ignore_dirs = listify_value(self.ignore_dirs)
        if ignore_dirs:
            self.directory_filter = _make_directory_filter(self.ignore_dirs)
        else:
            self.directory_filter = None

    def run(self):
        mappings = self._get_mappings()
        with open(self.output_file, "wb") as outfile:
            catalog = Catalog(project=self.project,
                              version=self.version,
                              msgid_bugs_address=self.msgid_bugs_address,
                              copyright_holder=self.copyright_holder,
                              charset=self.charset,
                              header_comment=(self.header_comment or DEFAULT_HEADER))

            for path, method_map, options_map in mappings:
                def callback(filename, method, options):
                    if method == "ignore":
                        return

                    # If we explicitly provide a full filepath, just use that.
                    # Otherwise, path will be the directory path and filename
                    # is the relative path from that dir to the file.
                    # So we can join those to get the full filepath.
                    if os.path.isfile(path):
                        filepath = path
                    else:
                        filepath = os.path.normpath(os.path.join(path, filename))

                    optstr = ""
                    if options:
                        optstr = " (%s)" % ", ".join(['%s="%s"' % (k, v) for
                                                      k, v in options.items()])
                    self.log.info("extracting messages from %s%s", filepath, optstr)

                if os.path.isfile(path):
                    current_dir = os.getcwd()
                    extracted = check_and_call_extract_file(
                        path, method_map, options_map,
                        callback, self.keywords, self.
