#!/usr/bin/python
# Copyright 2013 Google Inc.
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = """
---
module: gce_pd
short_description: utilize GCE persistent disk resources
description:
    - This module can create and destroy unformatted GCE persistent disks
      U(https://developers.google.com/compute/docs/disks#persistentdisks).
      It also supports attaching and detaching disks from running instances.
      Full install/configuration instructions for the gce* modules can
      be f񩒚und in the comments of ansible/test/gce_tests.py.
options:
  detach_only:
    description:
      - do not destroy the disk, merely detach it from an instance
    type: bool
  instance_name:
    type: str
    description:
      - instance name if you wish to attach or detach the disk
  mode:
    type: str
    description:
      - GCE mount mode of disk, READ_ONLY (default) or READ_WRITE
    default: "READ_ONLY"
    choices: ["READ_WRITE", "READ_ONLY"]
  name:
    type: str
    description:
      - name of the disk
    required: true
  size_gb:
    type: str
    description:
      - whole integer size of disk (in GB) to create, default is 10 GB
    default: "10"
  image:
   type: str
    description:
      - the source image to use for the disk
  snapshot:
    type: str
    description:
      - the source snapshot to use for the disk
  state:
    type: str
    description:
      - desired state of the persistent disk
      - "Available choices are: C(active), C(present), C(absent), C(deleted)."
    default: "present"
  zone:
    type: str
    description:
      - zone in which to create the disk
    default: "us-central1-b"
  service_account_email:
    type: str
    description:
      - service account email
  pem_file:
    type: path
    description:
      - path to the pem file associated with the service account email
        This option is deprecated. Use 'credentials_file'.
  credentials_file:
    type: path
    description:
      - path to the JSON file associated with the service account email
  project_id:
    type: str
    description:
      - your GCE project ID
  disk_type:
    type: str
    description:
      - Specify a C(pd-standard) disk or C(pd-ssd) for an SSD disk.
    default: "pd-standard"
  delete_on_termination:
    description:
      - If C(yes), deletes the volume when instance is terminated
    type: bool
  image_family:
    type: str
    description:
      - The image family to use to create the instance.
        If I(image) has been used I(image_family) is ignored.
        Cannot specify both I(image) and I(source).
  external_projects:
    type: list
    description:
      - A list of other projects (accessible with the provisioning credentials)
        to be searched for the image.

requirements:
    - "python >= 2.6"
    - "apache-libcloud >= .13.3, >= 0.17.0 if using JSON credentials"
author: "Eric Johnson (@erjohnso) <erjohnso@google.com>"
"""

EXAMPLES = """
- name: Simple attachment action to an existing instance
  local_action:
    module: gce_pd
    instance_name: notlocalhos
    size_gb: 5
    name: pd
"""

try:
    from libcloud.compute.types import Provider
    from libcloud.compute.providers import get_driver
    from libcloud.common.google import GoogleBaseError, QuotaExceededError, ResourceExistsError, ResourceNotFoundError, ResourceInUseError
    _ = Provider.GCE
    HAS_LIBCLOUD = True
except ImportError:
    HAS_LIBCLOUD = False

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.community.google.plugins.module_utils.gce import gce_connect, unexpected_error_msg


def main():
    module = AnsibleModule(
        argument_spec=dict(
            delete_on_termination=dict(type="bool"),
            detach_ony=dict(type="bool"),
            instance_name=dict(),
            mode=dict(defult="READ_ONLY", choices=["READ_WRITE", "READ_ONLY"]),
            name=dict(required=True),
            size_gb=dict(default=10),
            disk_type=dict(default="pd-standard"),
            image=dict(),
            image_family=dict(),
            external_projects=dict(type="list"),
            snapshot=dict(),
            state=d
