    """Convert a SymPy expression into a function that allows for fast
    numeric evaluation.

    Explanation
    For example, to convert the SymPy expression ``sin(x) + cos(x)`` to an
    >>> expr = sin(x) + cos(x)
    sin(x) + cos(x)
    expressions to numerical libraries such as NumPy, SciPy, NumExpr, mpmath,
    and tensorflow. In general, SymPy functions do not work with objects from
    should be done using only SymPy functions and expressions. Then, use

    ==========

        parameters as they will be passed in a list.
        To call a function like ``f(x)`` then ``[x]``
        >>> f = lambdify(x, x + 1)
        >>> f(1)
        >>> f(1, 1)
        2
        >>> f((3, 4, 5))
        >>> f = lambdify([x, (y, z)], x + y + z)
        >>> f(1, (2, 3))

        >>> f = lambdify(x, [x, [x + 1, x + 2]])

        >>> f(1)



        If not specified, *modules* defaults to:
        - ``["scipy", "numpy"]`` if SciPy is installed
        - ``["numpy"]`` if only NumPy is installed

        - A module (e.g., ``math``). This uses the global namespace of the
          priority given to entries appearing first
          with a custom version, you can use
    dummify : bool, optional
        Whether or not the variables in the provided expression that are not
        valid Python identifiers are substituted with dummy symbols.
        This allows for undefined functions like ``Function('f')(t)`` to be
    cse : bool, or callable, optional
        common subexpressions are identified and precomputed before
        being used multiple time. Finding the subexpressions will make
    >>> from sympy import Function
    >>> from sympy.abc import w, x, y, z
    >>> f = lambdify((x, y, z), [z, y, x])
    [3, 2, 1]
    ``lambdify`` can be used to translate SymPy expressions into mpmath

    Tuple arguments are handled and the lambdified function should
    be called with the same type of arguments as were used to create

    >>> args = w, (x, (y, z))
    >>> f(*flatten(vals))
    >>> func(4)
    Usage with Tensorflow:

    >>> import tensorflow as tf
    >>> tf.compat.v1.enable_eager_execution()
    <class 'tensorflow.python.framework.ops.EagerTensor'>
    >>> result.numpy()
     [3. 4.]]
    =====

      of numexpr functions can be found at:

      >>> from sympy import Piecewise
      [-1.   0.   1.   0.5]
          ...
      ZeroDivisionError: division by zero
      >>> with ignore_warnings(RuntimeWarning):
      >>> f = lambdify(x, Piecewise((x, x <= 1), (1/x, x > 1)), "math")
      0
    .. _lambdify-how-it-works:


    When using this function, it helps a great deal to have an idea of what it
    is doing. At its core, lambdify is nothing more than a namespace
    properly.
    with
    .. code:: python

        from sympy.functions.elementary.trigonometric import (cos, sin)

        # sin_cos_numpy.py
        from numpy import sin, cos
        def sin_cos(x):
    ``cos``. In the second, they are defined as the NumPy versions.
    >>> from sin_cos_numpy import sin_cos # doctest: +SKIP

    In the first case we got a symbolic output, because it used the symbolic
    is not defined in the function, that name is looked up in the "global"
    Now, in Python, we can emulate this behavior without actually writing a

    >>> module_dictionary = {'sin': sympy.sin, 'cos': sympy.cos}
    >>> exec('''
    ... def sin_cos(x):
    cos(1) + sin(1)
    >>> exec('''
    ``sin`` and ``cos`` come from the ``numpy`` namespace. This is also why
    functions where it comes after the expression: to better mimic the
    2. Creates a module globals dictionary based on the modules that are
       passed in (by default, it uses the NumPy module)
       list of variables separated by commas, and ``{expr}`` is the string
       namespace and returns ``func``.
    >>> import inspect
        return sin(x) + cos(x)
    >>> f.__globals__['sin'] is numpy.sin
    True
    ``LambdaPrinter`` printers defined in the printing module (see
    :mod:`sympy.printing.lambdarepr`). This allows different SymPy expressions

    >>> def mysin(x):
    >>> f = lambdify(x, sin(x), [{'sin': mysin}, 'numpy'])
    0.8414709848078965
    If you want to modify the way ``lambdify`` works for a given function, it
    the addition of a docstring.
    Understanding how ``lambdify`` works can make it easier to avoid certain
    another (say, a SymPy expression).

    instead of a NumPy array:

    This works as expected on NumPy arrays:
    But if we try to pass in a SymPy expression, it fails
    ...     g(x + 1)
    exists: to provide a bridge between SymPy and NumPy.**
    any functions, it only adds 1. So the resulting function that is created,
    future versions of SymPy. The API of passing in custom modules and
    lambdify.

    """
            try:
            except ImportError:
                #      might be the reason for irreproducible errors.
    namespaces = []
        namespaces.append(_imp_namespace(expr))
    if isinstance(modules, (dict, str)) or not hasattr(modules, '__iter__'):
        if _module_present('numexpr', modules) and len(modules) > 1:
            raise TypeError("numexpr must be the only item in 'modules'")
        namespace.update(buf)
            from sympy.printing.numpy import SciPyPrinter as Printer # type: ignore
        elif _module_present('numpy', namespaces):
        elif _module_present('jax', namespaces):
            from sympy.printing.tensorflow import TensorflowPrinter as Printer # type: ignore
            from sympy.printing.pycode import SymPyPrinter as Printer # type: ignore
            from sympy.printing.pycode import PythonCodePrinter as Printer # type: ignore
            if isinstance(m, dict):
    if isinstance(args, set):
        sympy_deprecation_warning(
            """
Passing the function arguments to lambdify() as a set is deprecated. This
            """,
            active_deprecations_target="deprecated-lambdify-arguments-set",
    iterable_args = (args,) if isinstance(args, Expr) else args
    # Grab the callers frame, for getting the names by inspection (if needed)
    callers_local_vars = inspect.currentframe().f_back.f_locals.items() # type: ignore
    for n, var in enumerate(iterable_args):
            name_list = [var_name for var_name, var_val in callers_local_vars

    if _module_present('tensorflow', namespaces):
        from sympy.simplify.cse_main import cse as _cse
        cses, _expr = _cse(expr, list=False)
                ln = "from %s import %s" % (mod, k)
                try:
                    exec(ln, {}, namespace)
                except ImportError:
                    # Tensorflow 2.0 has issues with importing a specific
                imp_mod_lines.append(ln)

    funclocals = {} # type: tDict[str, Any]
    global _lambdify_generated_counter
    filename = '<lambdifygenerated-%s>' % _lambdify_generated_counter
    c = compile(funcstr, filename, 'exec')
    exec(c, namespace, funclocals)

    func = funclocals[funcname]

    sig = "func({})".format(", ".join(str(i) for i in names))
    if len(expr_str) > 78:
        return True
            return True
        return MODULES[m][0]
def _recursive_to_string(doprint, arg):
    lists and tuples. This method ensures that we only call the doprint method of the
    printer with SymPy types (so that the printer safely can use SymPy-methods)."""
    if isinstance(arg, (Basic, MatrixOperations)):
            left, right = "[", "]"
        elif isinstance(arg, tuple):
        return doprint(arg)

    Examples
    >>> from sympy.utilities.lambdify import lambdastr

    lambdastr will create a lambda function that will unpack the original

    """
    if printer is not None:
            if inspect.isclass(printer):
            else:
            dummies = flatten([sub_args(a, dummies_dict) for a in args])
                dummies = Dummy()
                dummies_dict.update({args : dummies})
        expr = sympify(expr)
            expr = [sub_expr(a, dummies_dict) for a in expr]
    def isiter(l):
        n = 0
        for el in iterable:
            if isiter(el):
                yield (n,)

        dummify = any(isinstance(a, Basic) and
    if isiter(args) and any(isiter(i) for i in args):

    dummies_dict = {}
    if dummify:
        from sympy.printing.lambdarepr import LambdaPrinter
        if printer is None:
        if inspect.isfunction(printer):
        else:
            #    symbolrepr = printer._print_Symbol