import re
import json

from .common import InfoExtractor
from ..compat import (
    compat_str,
    compat_urlparse,
)
from ..utils import (
    ExtractorError,
    float_or_none,
    mimetype2ext,
    str_or_none,
    try_call,
    try_get,
    smuggle_url,
    unsmuggle_url,
    url_or_none,
    urljoin,
)


_ID_RE = r"(?:[0-9a-f]{32,34}|[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12,14})"


class MediasiteIE(InfoExtractor):
    _VALID_URL = r"(?xi)https?://[^/]+/Mediasite/(?:Play|Showcase/[^/#?]+/Presentation)/(?P<id>%s)(?P<query>\?[^#]+|)" % _ID_RE
    _EMBED_REGEX = [r'(?xi)<iframe\b[^>]+\bsrc=(["\'])(?P<url>(?:(?:https?:)?//[^/]+)?/Mediasite/Play/%s(?:\?.*?)?)\1' % _ID_RE]
    _TESTS = [
        {
            "url": "https://hitsmediaweb.h-its.org/mediasite/Play/2db6c271681e4f199af3c60d1f82869b1d",
            "info_dict": {
                "id": "2db6c271681e4f199af3c60d1f82869b1d",
                "ext": "mp4",
                "title": "Lecture: Tuesday, September 20, 2016 - Sir Andrew Wiles",
                "description": "Sir Andrew Wiles: “Equations in arithmetic”\\n\\nI will describe some of the interactions between modern number theory and the problem of solving equations in rational numbers or integers\\u0027.",
                "timestamp": 1474268400.0,
                "upload_date": "20160919",
            },
        },
        {
            "url": "http://mediasite.uib.no/Mediasite/Play/90bb363295d945d6b548c867d01181361d?catalog=a452b7df-9ae1-46b7-a3ba-aceeb285f3eb",
            "info_dict": {
                "id": "90bb363295d945d6b548c867d01181361d",
                "ext": "mp4",
                "upload_date": "20150429",
                "title": "5) IT-forum 2015-Dag 1  - Dungbeetle -  How and why Rain created a tiny bug tracker for Unity",
                "timestamp": 1430311380.0,
            },
        },
        {
            "url": "https://collegerama.tudelft.nl/Mediasite/Play/585a43626e544bdd97aeb71a0ec907a01d",
            "md5": "481fda1c11f67588c0d9d8fbdced4e39",
            "info_dict": {
                "id": "585a43626e544bdd97aeb71a0ec907a01d",
                "ext": "mp4",
                "title": "Een nieuwe wereld: waarden, bewustzijn en techniek van de mensheid 2.0.",
                "description": "",
                "thumbnail": r"re:^https?://.*\.jpg(?:\?.*)?$",
                "duration": 7713.088,
                "timestamp": 1413309600,
                "upload_date": "20141014",
            },
        },
        {
            "url": "https://collegerama.tudelft.nl/Mediasite/Play/86a9ea9f53e149079fbdb4202b521ed21d?catalog=fd32fd35-6c99-466c-89d4-cd3c431bc8a4",
            "md5": "ef1fdded95bdf19b12c5999949419c92",
            "info_dict": {
                "id": "86a9ea9f53e149079fbdb4202b521ed21d",
                "ext": "wmv",
                "title": "64ste Vakantiecursus: Afvalwater",
                "description": "md5:7fd774865cc69d972f542b157c328305",
                "thumbnail": r"re:^https?://.*\.jpg(?:\?.*?)?$",
                "duration": 10853,
                "timestamp": 1326446400,
                "upload_date": "20120113",
            },
        },
        {
            "url": "http://digitalops.sandia.gov/Mediasite/Play/24aace4429fc450fb5b38cdbf424a66e1d",
            "md5": "9422edc9b9a60151727e4b6d8bef393d",
            "info_dict": {
                "id": "24aace4429fc450fb5b38cdbf424a66e1d",
                "ext": "mp4",
                "title": "Xyce Software Training - Section 1",
                "description": r"re:(?s)SAND Number: SAND 2013-7800.{200,}",
                "upload_date": "20120409",
                "timestamp": 1333983600,
                "duration": 7794,
            },
        },
        {
            "url": "https://collegerama.tudelft.nl/Mediasite/Showcase/livebroadcast/Presentation/ada7020854f743c49fbb45c9ec7dbb351d",
            "only_matching": True,
        },
        {
            "url": "https://mediasite.ntnu.no/Mediasite/Showcase/default/Presentation/7d8b913259334b688986e970fae6fcb31d",
            "only_matching": True,
        },
        {
            # dashed id
            "url": "https://hitsmediaweb.h-its.org/mediasite/Play/2db6c271-681e-4f19-9af3-c60d1f82869b1d",
            "only_matching": True,
        },
    ]

    # look in Mediasite.Core.js (Mediasite.ContentStreamType[*])
    _STREAM_TYPES = {
        0: "video1",  # the main video
        2: "slide",
        3: "presentation",
        4: "video2",  # screencast?
        5: "video3",
    }

    @classmethod
    def _extract_embed_urls(cls, url, webpage):
        for embed_url in super()._extract_embed_urls(url, webpage):
            yield smuggle_url(embed_url, {"UrlReferrer": url})

    def __extract_slides(self, *, stream_id, snum, Stream, duration, images):
        slide_base_url = Stream["SlideBaseUrl"]

        fname_template = Stream["SlideImageFileNameTemplate"]
        if fname_template != "slide_{0:D4}.jpg":
            self.report_warning("Unusual slide file name template; report a bug if slide downloading fails")
        fname_template = re.sub(r"\{0:D([0-9]+)\}", r"{0:0\1}", fname_template)

        fragments = []
        for i, slide in enumerate(Stream["Slides"]):
            if i == 0:
                if slide["Time"] > 0:
                    default_slide = images.get("DefaultSlide")
                    if default_slide is None:
                        default_slide = images.get("DefaultStreamImage")
                    if default_slide is not None:
                        default_slide = default_slide["ImageFilename"]
                    if default_slide is not None:
                        fragments.append({
                            "path": default_slide
