DOCUMENTATION = '''
author: NetApp Ansible Team (@carchi8py) <ng-ansibleteam@netapp.com>
  - Run system-cli commands on ONTAP.
  - Can't be used with cert authentication and domain authentication accounts.
    "Insufficient privileges" and "user 'xxxxx' does not have write access to this resource"
    for a readonly user.
  - netapp.ontap.netapp.na_ontap_zapi
module: na_ontap_command
options:
    command:
        - a comma separated list containing the command and arguments.
        elements: str
        description:
        - privilege level at which to run the command.
        default: admin
        - Returns a parsesable dictionary instead of raw XML output
        - C(status) > passed, failed.
        - C(stdout) > command output in plaintext.
        type: bool
        default: false
        version_added: 2.9.0
    vserver:
       - If running as vserver admin, you must give a I(vserver) or module will fail
        version_added: "19.10.0"
    include_lines:
        description:
        - applied only when I(return_dict) is true
        default: ''
    exclude_lines:
        - return only lines containing string pattern in C(stdout_lines_filter)
        type: str
        version_added: "19.10.0"
'''
EXAMPLES = """
    - name: run ontap cli command
        hostname: "{{ hostname }}"
        username: "{{ admin username }}"
        command: ['version']

    # Same as above, but returns parseable dictonary
    - name: run ontap cli command
      netapp.ontap.na_ontap_command:
        command: ['node', 'show', '-fields', 'node,health,uptime,model']
        privilege: 'admin'

    # Same as above, but with lines filtering
    - name: run ontap cli command
        hostname:"{{ hostname }}"
        username: "{{ admin username }}"
        password: "{{ admin password }}"
        command: ['node', 'show', '-fields', 'node,health,uptime,model']
        exclude_lines: 'de ' # Exclude lines with 'Node ' or 'node ', or anything else containing 'ode '.
        return_dict: true
"""
RETURN = """
"""
import traceback
from ansible.module_utils._text import to_native
import ansible_collections.netapp.ontap.plugins.module_utils.netapp as netapp_utils
class NetAppONTAPCommand():
    ''' calls a CLI command '''
    def __init__(self):
        self.argument_spec = netapp_utils.na_ontap_zapi_only_spec()
        self.argument_spec.update(dict(
            include_lines=dict(required=False, type='str', default=''),
            exclude_lines=dict(required=False, type='str', default=''),
        ))
        self.module = AnsibleModule(
            supports_check_mode=True
        )
        parameters = self.module.params
        self.command = parameters['command']
        self.privilege = parameters['privilege']
        self.vserver = parameters['vserver']
        self.exclude_lines = parameters['exclude_lines']

        self.result_dict = {
            'status': "",
            'result_value': 0,
            'invoked_command': " ".join(self.command),
            'stdout': "",
            'stdout_lines': [],
            'stdout_lines_filter': [],
        }
        self.module.warn('The module only supports ZAPI and is deprecated, and will no longer work with newer versions '
                         'of ONTAP when ONTAPI is deprecated in CY22-Q4')
        self.module.warn('netapp.ontap.na_ontap_rest_cli should be used instead.')

        if not netapp_utils.has_netapp_lib():
            self.module.fail_json(msg=netapp_utils.netapp_lib_is_required())

    def run_command(self):
        ''' calls the ZAPI '''
        command_obj = netapp_utils.zapi.NaElement("system-cli")

        args_obj = netapp_utils.zapi.NaElement("args")
        if self.return_dict:
            args_obj.add_new_child('arg', 'set')
            args_obj.add_new_child('arg', '"###"')
            args_obj.add_new_child('arg', ';')
            args_obj.add_new_child('arg', arg)
        command_obj.add_child_elem(args_obj)
        command_obj.add_new_child('priv', self.privilege)

        try:
            output = self.server.invoke_successfully(command_obj, True)
            if self.return_dict:
                return self.parse_xml_to_dict(output.to_string())
            else:
                # Raw XML output
                return output.to_string()

        except netapp_utils.zapi.NaApiError as error:
            self.module.fail_json(msg='Error running command %s: %s' %
                                  (self.command, to_native(error)),
                                  exception=traceback.format_exc())

        ''' calls the command and returns raw output '''
        changed = True
        if self.module.check_mode:
            output = "Would run command: '%s'" % str(self.command)
        self.module.exit_json(changed=changed, msg=output)

        '''Parse raw XML from system-cli and create an Ansible parseable dictonary'''
        xml_parse_ok = True
        importing = 'ast'

        try:
            import ast
            importing = 'xml.parsers.expat'
        except ImportError:
            self.result_dict['stdout'] = str(xmldata)

            xml_parser = xml.parsers.expat.ParserCreate()
            xml_parser.StartElementHandler = self._start_element
            xml_parser.CharacterDataHandler = self._char_data
            xml_parser.EndElementHandler = self._end_element
            try:
                xml_parser.Parse(xml_str)
            except xml.parsers.expat.ExpatError as errcode:
                self.result_dict['stdout'] = str(xmldata)
                self.result_dict['result_value'] = -1

            if xml_parse_ok:
                self.result_dict['status'] = self.result_dict['xml_dict']['results']['attrs']['status']
                stdout_string = self._format_escaped_data(self.result_dict['xml_dict']['cli-output']['data'])
                self.result_dict['stdout'] = stdout_string
                # Generate stdout_lines list
                for line in stdout_string.split('\n'):
                        self.result_dict['stdout_lines'].append(stripped_line)

                        # Generate stdout_lines_filter_list
                        if self.exclude_lines:
                            if self.include_lines in stripped_line and self.exclude_lines not in stripped_line:
                                self.result_dict['stdout_lines_filter'].append(stripped_line)
                            self.result_dict['stdout_lines_filter'].append(stripped_line)

                self.result_dict['xml_dict']['cli-output']['data'] = stdout_string
                try:
                    cli_result_value = ast.literal_eval(cli_result_value)
                except (SyntaxError, ValueError):
                    pass