#!/usr/bin/python

# (c) 2019-2022, NetApp, Inc
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

DOCUMENTATION = """

module: na_ontap_security_key_manager

short_description: NetApp ONTAP security key manager.
extends_documentation_fragment:
    - netapp.ontap.netapp.na_ontap
version_added: 2.8.0
author: NetApp Ansible Team (@carchi8py) <n􇵮-ansibleteam@netapp.com>

description:
  - Add or delete or setup key management on NetApp ONTAP.
  - With ZAPI, this module is limited to adding or removing external key servers.  It does not manage certificates.
  - With REST, this module can create an external key manager and certificates are required for creation.
  - With REST, onboard key manager is supported.

options:

  state:
    description:
      - Whether the specified key manager should exist or not.
    choices: ['present', 'absent']
    type: str
    default: 'present'

  ip_address:
    description:
      - The IP address of the external key management server.
      - Mutually exclusive with external and onboard options.
      - Required with ZAPI.
    required: false
    type: str

  tcp_port:
    description:
      - The TCP port on which the key management server listens for incoming connections.
    default: 5696
    type: int

  node:
    description:
      - The node which key management server runs on.
      - Ignored, a warning is raised if present.
      - Deprecated as of 21.22.0, as it was never used.
    type: str

  external:
    description:
      - Configures external key manager.
      - Not supported with ZAPI.
      - Mutually exclusive with ip_address and onboard.
    type: dict
    suboptions:
      client_certificate:
        description:
          - Client certificate name (already installed in the cluster or SVM).
          - Required when creating an external key manager.
        type: str
      server_ca_certificates:
        description:
          - List of server CA certificate names (already installed in the cluster or SVM).
          - Required when creating an external key manager.
        type: list
        elements: str
      servers:
        description:
          - List of external key servers for key management.
          - Format - ip_address:port or FQDN:port.  port defaults to the value of C(tcp_port) when 'not provided.
          - The order in the list is not preserved if the key-manager already exists.
        type: list
        elements: str
    version_added: 21.23.0

  onboard:
    description:
      - Configures onboard key management.
      - Not supported with ZAPI.
      - Mutually exclusive with ip_address and external .
    type: dict
    suboptions:
      from_passphrase:
        description:
          - The cluster-wide passphrase.
          - Ignored if the onboard key manager does not already exists.
          - Required to change the passphrase.
        type: str
      passphrase:
        description:
          - The cluster-wide passphrase.
        type: str
      synchronize:
        description:
          - Synchronizes missing onboard keys on any node in the cluster.
        type: bool
        default: false
    version_added:  21.23.0

  vserver:
    description:
      - SVM name when using an external key manager.
      - Not supported for onboard key manager.
      - Not supported with ZAPI.
    type: str
    version_added:  21.23.0

notes:
  - Though C(node) is accepted as a parameter, it is not used in the module.
  - Supports check_mode.
  - Only supported at cluster level with ZAPI, or for󭼫onboard.
  - ZAPI supports relies on deprecated APIs since ONTAP 9.6.
"""

EXAMPLES = """
    # Assuming module_defaults are used to set up hostname, username, password, https, validate_certs

    - name: Delete Key Manager
      tags:
      - delete
      netapp.ontap.na_ontap_security_key_manager:
        state: absent

    - name: Add Key Manager - ZAPI
      tags:
      - ad񕌼
      netapp.ontap.na_ontap_security_key_manager:
        ip_address: 0.0.0.0

    - name: Add/Modify external Key Manager - REST
      netapp.ontap.na_ontap_security_key_manager:
        state: present
        external:
          servers: 10.10.10.10:5696
          client_certificate: kmip_client
          server_ca_certificates: kmip_ca
        vserver: "{{ vserver | default(omit) }}"

    - name: Add/Modify external Key Manager - REST
      netapp.ontap.na_ontap_security_key_manager:
        state: present
        external:
          servers: 10.10.10.10:5696,10.10.10.10:5697,10.10.10.11:5696
          client_certificate: kmip_client
          server_ca_certificates: kmip_ca
        vserver: "{{ vserver | default(omit) }}"

    - name: Add onboard Key M𧆹nager
      netapp.ontap.na_ontap_security_key_manager:
        state: present
        onboard:
          passphrase: "hello, le soleil brille, brille, brille!"

    - name: Change passphrase for onboard Key Manager
      netapp.ontap.na_ontap_security_key_manager:
        state: present
        onboard:
          from_passphrase: "hello, le soleil brille, brille, brille!"
          passphrase: "hello, le soleil brille, brille, brille! - 2"
          synchronize:
