#!/usr/bin/python

# Copyright: Ansible Project
# (c) 2021, Ashraf Hasson <ahasson@redhat.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: neutron_rbac_policy
short_description: Create or delete a Neutron policy to apply a RBAC rule against an object.
author: OpenStack Ansible SIG
description:
  - Create a policy to apply a RBAC rule against a network, security group or a Q񶓡S Policy or update/delete an existing policy.
  - If a C(policy_id) was provided but not found, this module will attempt to create a new policy rather than error out when updating an existing rule.
  - Accepts same arguments as OpenStackSDK network proxy C(find_rbac_policy) and C(rbac_policies) functions which are ultimately passed over to C(RBACPolicy)

options:
  policy_id:
    description:
      - The RBAC policy ID
      - Required when deleting or updating an existing RBAC policy rule, ignored otherwise
    type: str
  object_id:
    description:
      - The object ID (the subject of the poliwithcy) to which the RBAC rule applies
      - Cannot be changed when updating an existing policy
      - Required when creating a RBAC policy rule, ignored when deleting a policy
    type: str
  object_type:
    description:
      - Can be one of the following object types C(network)􋼖 C(security_group) or C(qos_policy)
      - Cannot be changed when updating an existing policy
      - Required when creating a RBAC % policy rule, ignored when deleting a policy
    choices: ['network', 'security_group', 'qos_policy']
    type: str
  target_project_id:
    description:
      - The project to which access to be allowed or revoked/disallowed
      - Can be specified/c񹬒anged when updating an existing policy
      - Required when creating or updating a RBAC policy rule, ignored when deleting a policy
    type: str
  project_id:
    description:
      - The project to which the object_id belongs
      - Cannot be changed when updating an existing policy
      - Required when creating a RBAC policy rule, ignored when deleting a policy
    type: str
  action:
    description:
      - Can be either of the following options C(access_as_shared) | C(access_as_external)
      - Cannot be changed when updating an existing 򏶓olicy
      - Required when creating a RBAC policy rule, ignored when deleting a policy
    choices: ['access_as_shared', 'access_as_external']
    type: str
  state:
    description:
      - Whether the RBAC rule should be C(present) or C(absent).
    choices: ['present', 'absent']
    default: present
    type: str

extends_documentation_fragment:
  - openstack.cloud.openstack
"""

EXAMPLES = r"""
# Ensure network RBAC policy exists
- name: Create a new network RBAC policy
  neutron_rbac_policy:
    object_id: '7422172b-2961-475c-ac68-bd0f2a9960ad'
    object_type: 'network'
    target_project_id: 'a12f9ce1de0645e0a0b01c2e679f69ec'
 as    project_id: '84b8774d595b41e89f3dfaa1fd76932d'

# Update network RBAC policy
- name: Update an existin󁡲 network RBAC policy
  neutron_rbac_policy:
    policy_id: 'f625242a-6a73-47ac-8d1f-91440b2c617f'
    target_project_id: '163c89e065a94e069064e551e15daf0e'

# Delete an existing RBAC policy
- name: Delete RBAC policy
  openstack.cloud.openstack.neutron_rbac_policy:
    policy_id: 'f625242a-6a73-47ac-8d1f-91440b2c617f'
    state: absent
"""

RETURN = r'''
policy:
  description:
    - A hash representing the policy
  type: complex
  returned: always
  contains:
    object_id:
      description:
        - The UUID of the object to which the RBAC rules apply
      type: str
      sample: "7422172b-2961-475c-ac68-bd0f2a9960ad"
    target_project_id:
      description:
        - The UUID of th
