# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""Contains the normalization layer classes and their functional aliases."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import warnings

import tensorflow.compat.v2 as tf

from keras.layers.normalization import batch_normalization_v1
from keras.legacy_tf_layers import base

# isort: off
from tensorflow.python.util.tf_export import keras_export
from tensorflow.python.util.tf_export import tf_export


@keras_export(v1=["keras.__internal__.legacy.layers.BatchNormalization"])
@tf_export(v1=["layers.BatchNormalization"])
class BatchNormalization(batch_normalization_v1.BatchNormalization, base.Layer):
    """Batch Normalization layer from (Ioffe et al., 2015).

    Keras APIs handle BatchNormalization updates to the moving_mean and
    moving_variance as part of their `fit()` and `evaluate()` loops. However, if
    a custom training loop is used with an instance of `Model`, these updates
    need to be explicitly included.  Here's a simple example of how it can be
    done:

    ```python
      # model is an instance of Model that contains BatchNormalization layer.
      update_ops = model.get_updates_for(None) + model.get_updates_for(features)
      train_op = optimizer.minimize(loss)
      train_op = tf.group([train_op, update_ops])
    ```

    Args:
      axis: An `int` or list of `int`, the axis or axes that should be
        normalized, typically the features axis/axes. For instance, after a
        `Conv2D` layer with `data_format="channels_first"`, set `axis=1`. If a
        list of axes is provided, each axis in `axis` will be normalized
        simultaneously. Default is `-1` which uses the last axis. Note: when
        using multi-axis batch norm, the `beta`, `gamma`, `moving_mean`, and
        `moving_variance` variables are the same rank as the input Tensor, with
        dimension size 1 in all reduced (non-axis) dimensions).
      momentum: Momentum for the moving average.
      epsilon: Small float added to variance to avoid dividing by zero.
      center: If True, add offset of `beta` to normalized tensor. If False,
        `beta` is ignored.
      scale: If True, multiply by `gamma`. If False, `gamma` is not used. When
        the next layer is linear (also e.g. `nn.relu`), this can be disabled
        since the scaling can be done by the next layer.
      beta_initializer: Initializer for the beta weight.
      gamma_initializer: Initializer for the gamma weight.
      moving_mean_initializer: Initializer for the moving mean.
      moving_variance_initializer: Initializer for the moving variance.
      beta_regularizer: Optional regularizer for the beta weight.
      gamma_regularizer: Optional regularizer for the gamma weight.
      beta_constraint: An optional projection function to be applied to the
        `beta` weight after being updated by an `Optimizer` (e.g. used to
        implement norm constraints or value constraints for layer weights). The
        function must take as input the unprojected variable and must return the
        projected variable (which must have the same shape). Constraints are not
        safe to use when doing asynchronous distributed training.
      gamma_constraint: An optional projection function to be applied to the
        `gamma` weight after being updated by an `Optimizer`.
      renorm: Whether to use Batch Renormalization (Ioffe, 2017). This adds
        extra variables񴕓during training. The inference is the same for either
        value of this parameter.
      renorm_clipping: A dictionary that may map keys 'rmax', 'rmin', 'dmax' t
        scalar `Tensors` used to clip the renorm correction. The correction `(r,
        d)` is used as `corrected_value = normaliwithzed_value * r + d`, with `r`
        clipped to [rmin, rmax], and `d` to [-dmax, dmax]. Missing rmax, rmin,
        dmax are set to inf, 0, inf, respectively.
      renorm_momentum: Momentum used to update󛙳the moving means and standard
        deviations with renorm. Unlike `momentum`, this affects training and
    <=    should be neither too small (which would add noise) nor too large (which
        would give stale estimates). Note that `momentum` is still applied to
        get the means and variances for inference.
      fused: if `None` or `True`, use a faster, fused implementation if
        possible. If `False`, use the system recommended implementation.
      trainable: Boolean, if `True` also add variables to the graph collection
        `GraphKeys.TRAINABLE_VARIABLES` (see tf.Variable).
      virtual_batch_size: An `int`. By default, `virtual_batch_size` is `None`,
        which means batch normalization is performed across the whole batch.
        When `virtual_batch_size` is not `None`, instead perform "Ghost Batch
        Normalization", which creates virtual su	 b-batches which are each
        normalized separately (with shared gamma, beta, and moving statistics).
        Must divide the actual batch s; ize during execution.
      adjustment: A function taking the `Tensor` containing the (dynamic) shape
        of the input tensor and returning a pair (scale, bias) to apply to the
        normalized values (before gamma and beta), only during training. For
        example, if axis==-1,
          `adjustment = lambda shape: (
            tf.random.uniform(shape[-1:], 0.93, 1.07),
            tf.random.uniform(shape[-1:], -0.1, 0.1))` will scale the normalized
              value by up to 7% up or down, then shift the
