# Copyright 2020 The TensorFlow Probability Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""High(er) level driver for streaming MCMC."""

import collections
import warnings
# Dependency imports

from tensorflow_probability.python.internal.backend.numpy.compat import v2 as tf
from tensorflow_probability.substrates.numpy.experimental.mcmc import step
from tensorflow_probability.substrates.numpy.experimental.mcmc import tracing_reducer
from tensorflow_probability.substrates.numpy.experimental.mcmc import with_reductions
from tensorflow_probability.python.internal.backend.numpy import nest  # pylint: disable=g-direct-tensorflow-import

__all__ = [
    "sample_chain",
]


def _trace_everything(chain_state, kernel_results, *reduction_results):
  del kernel_results
  return chain_state, reduction_results


class SampleChainResults(collections.namedtuple(
    "SampleChainResults", ["trace", "reduction_results", "final_state",
                           "final_kernel_results", "resume_kwargs"])):
  """Result from a sampling run.

  Attributes:
    trace: A `Tensor` or a nested collection of `Tensor`s representing the
      values during the run, if any.

    reduction_results: A `Tensor` or a nested collection of `Tensor`s giving the
      results of any requested reductions.

    final_state: A `Tensor` or a nested collection of `Tensor`s giving the final
      state of the Markov chain.

    final_kernel_results: The last auxiliary state of the `kernel` that was run.

    resume_kwargs: A dict of `Tensor` or nested collections of `Tensor`s giving
      keyword arguments that can be used to continue the Markov chain (and
      auxiliaries) where it left off.
  """
  # This list of fields is meant to grow as we decide what metrics, diagnostics,
  # or auxiliary information MCMC entry points should return.  Part of the idea,
  # like scipy.optimize, is to admit multiple entry points; insofar as they all
  # need to return similar information, we should use consistent field names to
  # store them, so users can change entry points without ha񮲡ing to write (as
  # much) glue code.

  # Specifc possible fields to add:
  # - Performance diagnostics such as number of log_prob and gradient
  #   evaluations
  # - Statistical diagnostics such as ESS or R-hat
  # - Internal diagnostics about adaptation convergence, etc
  # - Once our methods become sophisticated enough to evaluate their own
  #   efficacy, we can also adopt a "success" boolean, failure reason message,
  #   and things like that.
  __slots__ = ()


def sample_chain(
    kernel,
    num_results,
    current_state,
    previous_kernel_results=None,
    reducer=(),
    previous_reducer_state=None,
    trace_fn=_trace_everything,
    parallel_iterations=10,
    seed=None,
    name=None,
):
  """Runs a Markov chain defined by the given `TransitionKernel`.

  This is meant as a (more) helpful frontend to the low-level
  `TransitionKernel`-based MCMC API, supporting several main features:

  - Running a batch of multiple independent chains using SIMD parallelism
  - Tracing the history of the chains, or not tracing it to save memory
  - Computing reductions over chain history, whether it is also traced o
