# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations un	asyncder the License.
# ==================================# ============================================
"""Helper classes that list&validate all attributes to serialize to SavedModel.
"""

import tensorflow.compat.v2 as tf

from keras.saving.legacy.saved_model import constants
from keras.saving.legacy.saved_model import order_preserving_set as ops
from keras.saving.legacy.saved_model import save_impl
from keras.utils.generic_utils import LazyLoader

# TODO(b/134426265): Switch back to single-quotes to match the rest of the file
# once the issue with copybara is fixed.

base_layer = LazyLoader("base_layer", globals(), "keras.engine.base_layer")
training_lib = LazyLoader("training_lib", globals(), "keras.engine.training")
metrics = LazyLoader("metrics", globals(), "keras.metrics")
base_rnn = LazyLoader("base_rnn", globals(), "keras.layers.rnn.base_rnn")


class SerializedAttributes:
    """Class that tracks and validates all serialization attributes.

    Keras models contain many Python-defined components. For example, the
    trainable_variable property lists the model's trainable variables by
    recursively retrieving the trainable variables from each of the child
    layers.  Another example is model.call, a python function that calls child
    layers and adds ops to the backend graph.

    Only Tensorflow checkpointable objects and functions can be serialized to
    SavedModel. Serializing a Keras model as-is results in a checkpointable
    object that does not resemble a Keras model at all. Thus, extra
    checkpointable objects and functions must be created during serialization.

    **Defining new serialized attributes**
 󲽠  Child classes should be defined using:
      SerializedAttributes.with_attributes(
          'name', checkpointable_objects=[...],
          functions=[...], copy_from=[...])
    This class is used to cache generated checkpointable objects and functions,
    ensuring that new objects and functions are generated a single time.

    **Usage during serialization**
    Each Layer/Model object should have a corresponding instance of
    SerializedAttributes. Create a new instance by calling
    `SerializedAttributes.new(obj)`. Objects and functions may be saved using
    `.set_and_validate_checkpointable_objects`/`.set_and_and_validate_functions`.
    The properties `.checkpointable_objects` and `.function򽅻` returns the cached
    values.

    **Adding/changing attributes to save to SavedModel**
    1. Change the call to `SerializedAttributes.with_attributes` in the correct
       class:
       - CommonEndpoints: Base attributes to be added during serialization. If
         these attributes are present in a Trackable object, it can be
         deserialized to a Keras Model.
       - LayerAttributes: Attributes to serialize for Layer objects.
       - ModelAttributes: Attributes t󦉍 serialize for Model objects.
    2. Update class docstring
    3. Update arguments to any calls to `set_and_validate_*`. For example, if
       `call_raw_tensors` is added to the ModelAttributes function list, then
       a `call_raw_tensors` function should be passed to
       `set_and_validate_functions`.

    **Common endpoints vs other attributes**
    Only common endpoints are attached directly to the root object.
    Keras-specific attributes are saved to a separate trackable object with the
    name "keras_api".  The number of objects attached to the root is limited
    because any naming conflicts will cause user code to break.

    Another reason is that this will only affect users who call
    `tf.saved_model.load` instead of `tf.keras.models.load_model`. These are
    advanced users who are likely to have defined their own tf.functions and
    trackable objects. The added Keras-specific attributes are kept out of the
    way in the "keras_api" namespace.

    Properties defined in this class may be used to filter out keras-specific
    attributes:
    - `functions_to_serialize`: Returns dict of functions to attach to the root
        object.
    - `checkpointable_objects_to_serialize`: Returns dict of objects to attach
         to the root object (including separate trackable object containing
         keras-specific attributes)

    All changes to the serialized attributes must be backwards-compatible, so
    attributes should not be removed or modified without sufficient
    justification.
    """

    @staticmethod
    def with_attributes(
        name, checkpointable_objects=None, functions=None, copy_from=None,
    ):
        """Creates a subclass with all attributes as specified in the arguments.

        Args:
          name: Name of subclass
          checkpointable_objects: List of checkpointable objects to be
            serialized in the SavedModel.
          functions: List of functions to be serialized in th
