def st_list(parser, args, output_manager, return_parser=False):

    def _print_stats(options, stats, human, totals):
        container = stats.get("container", None)
        for item in stats["listing"]:
            item_name = item.get("name")
            if not options["long"] and not human and not options["versions"]:
                output_manager.print_msg(item.get("name", item.get("subdir")))
            else:
                if not container:    # listing containers
                    item_bytes = item.get("bytes")
                    byte_str = prt_bytes(item_bytes, human)
                    count = item.get("count")
                    totals["count"] += count
                    try:
                        meta = item.get("meta")
                        utc = gmtime(float(meta.get("x-timestamp")))
                        datestamp = strftime("%Y-%m-%d %H:%M:%S", utc)
                    except TypeError:
                        datestamp = "????-??-?? ??:??:??"
                    storage_policy = meta.get("x-storage-policy", "???")
                    if not options["totals"]:
                        output_manager.print_msg(
                            "%12s %s %s %-15s %s", count, byte_str,
                            datestamp, storage_policy, item_name)
                else:    # list container contents
                    subdir = item.get("subdir")
                    content_type = item.get("content_type")
                    if subdir is None:
                        item_bytes = item.get("bytes")
                        byte_str = prt_bytes(item_bytes, human)
                        date, xtime = item.get("last_modified").split("T")
                        xtime = xtime.split(".")[0]
                    else:
                        item_bytes = 0
                        byte_str = prt_bytes(item_bytes, human)
                        date = xtime = ""
                        item_name = subdir
                    if not options["totals"]:
                        if options["versions"]:
                            output_manager.print_msg(
                                "%s %10s %8s %16s %24s %s",
                                byte_str, date, xtime,
                                item.get("version_id", "null"),
                                content_type, item_name)
                        else:
                            output_manager.print_msg(
                                "%s %10s %8s %24s %s",
                                byte_str, date, xtime, content_type, item_name)
                totals["bytes"] += item_bytes

    parser.add_argument(
        "-l", "--long", dest="long", action="store_true", default=False,
        help="Long listing format, similar to ls -l.")
    parser.add_argument(
        "--lh", dest="human", action="store_true",
        default=False, help="Report sizes in human readable format, "
        "similar to ls -lh.")
    parser.add_argument(
        "-t", "--totals", dest="totals",
        help="used with -l or --lh, only report totals.",
        action="store_true", default=False)
    parser.add_argument(
        "-p", "--prefix", dest="prefix",
        help="Only list items beginning with the prefix.")
    parser.add_argument(
        "-d", "--delimiter", dest="delimiter",
        help="Roll up items with the given delimiter. For containers "
             "only. See OpenStack Swift API documentation for "
             "what this means.")
    parser.add_argument("-j", "--json", action="store_true",
                        help="print listing information in json")
    parser.add_argument("--versions", action="store_true",
                        help="display all versions")
    parser.add_argument(
        "-H", "--header", action="append", dest="header",
        default=[],
        help="Adds a custom request header to use for listing.")

    # We return the parser to build up the bash_completion
    if return_parser:
        return parser

    options, args = parse_args(parser, args)
    args = args[1:]
    if options["delimiter"] and not args:
        exit("-d option only allowed for container listings")
    if options["versions"] and not args:
        exit("--versions option only allowed for container listings")

    human = options.pop("human")
    if human:
        options["long"] = True

    if options["totals"] and not options["long"]:
        output_manager.error(
            "Listing totals only works with -l or --lh.")
        return

    with SwiftService(options=options) as swift:
        try:
            if not args:
                stats_parts_gen = swift.list()
                container = None
            else:
                container = args[0]
                args = args[1:]
                if "/" in container or args:
                    output_manager.error(
                        "Usage: %s list %s\n%s", BASENAME,
                        st_list_options, st_list_help)
                    return
                else:
                    stats_parts_gen = swift.list(container=container)

            if options.get("json", False):
                def listing(stats_parts_gen=stats_parts_gen):
                    for stats in stats_parts_gen:
                        if stats["success"]:
                            for item in stats["listing"]:
                                yield item
                        else:
                            raise stats["error"]

                json.dump(
                    JSONableIterable(listing()), output_manager.print_stream,
                    sort_keys=True, indent=2)
                output_manager.print_msg("")
                return

            totals = {"count": 0, "bytes": 0}
            for stats in stats_parts_gen:
                if stats["success"]:
                    _print_stats(options, stats, human, totals)
                else:
                    raise stats["error"]

            # report totals
            if options["long"] or human:
                if container is None:
                    output_manager.print_msg(
                        "%12s %s", prt_bytes(totals["count"], True),
                        prt_bytes(totals["bytes"], human))
                else:
                    output_manager.print_msg(
                        prt_bytes(totals["bytes"], human))

        except SwiftError as e:
            output_manager.error(e.value)


st_stat_options = """[--lh] [--header <header:value>]
                  [--version-id <version_id>]
                  [<container> [<object>]]
"""

st_stat_help = """
Displays information for the account, container, or object.

Positional arguments:
  [<container>]           Name of container to stat from.
  [<object>]              Name of object to stat.

Optional arguments:
  --lh                  Report sizes in human readable format similar to
                        ls -lh.
  --version-id <version-id>
                        Report stat of specific version of a versioned object.
  -H, --header <header:value>
                        Adds a custom request header to use for stat.
""".strip("\n")


def st_stat(parser, args, output_manager, return_parser=False):
    parser.add_argument(
        "--lh", dest="human", action="store_true", default=False,
        help="Report sizes in human readable format similar to ls -lh.")
    parser.add_argument(
        "--version-id", action="store", default=None,
        help="Report stat of a specific version of a versioned object")
    parser.add_argument(
        "-H", "--header", action="append", dest="header",
        default=[],
        help="Adds a custom request header to use for stat.")

    # We return the parser to build up the bash_completion
    if return_parser:
        return parser

    options, args = parse_args(parser, args)
    args = args[1:]
    if options["version_id"] and len(args) < 2:
        exit("--version-id option only allowed for object stats")

    with SwiftService(options=options) as swift:
        try:
            if not args:
                stat_result = swift.stat()
                if not stat_result["success"]:
                    raise stat_result["error"]
                items = stat_result["items"]
                headers = stat_result["headers"]
                print_account_stats(items, headers, output_manager)
            else:
                container = args[0]
                if "/" in container:
                    output_manager.error(
                        "WARNING: / in container name; you might have "
                        "meant '%s' instead of '%s'." %
                        (container.replace("/", " ", 1), container))
                    return
                args = args[1:]
                if not args:
                    stat_result = swift.stat(container=container)
                    if not stat_result["success"]:
                        raise stat_result["error"]
                    items = stat_result["items"]
                    headers = stat_result["headers"]
                    print_container_stats(items, headers, output_manager)
                else:
                    if len(args) == 1:
                        objects = [args[0]]
                        stat_results = swift.stat(
                            container=container, objects=objects)
                        for stat_result in stat_results:  # only 1 result
                            if stat_result["success"]:
                                items = stat_result["items"]
                                headers = stat_result["headers"]
                                print_object_stats(
                                    items, headers, output_manager,
                                )
                            else:
                                raise(stat_result["error"])
                    else:
                        output_manager.error(
                            "Usage: %s stat %s\n%s", BASENAME,
                            st_stat_options, st_stat_help)

        except SwiftError as e:
            output_manager.error(e.value)


st_post_options = """[--read-acl <acl>] [--write-acl <acl>] [--sync-to <sync-to>]
                  [--sync-key <sync-key>] [--meta <name:value>]
                  [--header <header>]
                  [<container> [<object>]]
"""

st_post_help = """
Updates meta information for the account, container, or object.
If the container is not found, it will be created automatically.

Positional arguments:
  [<container>]           Name of container to post to.
  [<object>]              Name of object to post.

Optional arguments:
  -r, --read-acl <acl>  Read ACL for containers. Quick summary of ACL syntax:
                        .r:*, .r:-.example.com, .r:www.example.com,
                        account1 (v1.0 identity API only),
                        account1:*, account2:user2 (v2.0+ identity API).
  -w, --write-acl <acl> Write ACL for containers. Quick summary of ACL syntax:
                        account1 (v1.0 identity API only),
                        account1:*, account2:user2 (v2.0+ identity API).
  -t, --sync-to <sync-to>
                        Sync To for containers, for multi-cluster replication.
  -k, --sync-key <sync-key>
                        Sync Key for containers, for multi-cluster replication.
  -m, --meta <name:value>
                        Sets a meta data item. This option may be repeated.
                        Example: -m Color:Blue -m Size:Large
  -H, --header <header:value>
                        Adds a customized request header.
                        This option may be repeated. Example
                        -H "content-type:text/plain" -H "Content-Length: 4000"
""".strip("\n")


def st_post(parser, args, output_manager, return_parser=False):
    parser.add_argument(
        "-r", "--read-acl", dest="read_acl", help="Read ACL for containers. "
        "Quick summary of ACL syntax: .r:*, .r:-.example.com, "
        ".r:www.example.com, account1, account2:user2")
    parser.add_argument(
        "-w", "--write-acl", dest="write_acl", help="Write ACL for "
        "containers. Quick summary of ACL syntax: account1, "
        "account2:user2")
    parser.add_argument(
        "-t", "--sync-to", dest="sync_to", help="Sets the "
        "Sync To for containers, for multi-cluster replication.")
    parser.add_argument(
        "-k", "--sync-key", dest="sync_key", help="Sets the "
        "Sync Key for containers, for multi-cluster replication.")
    parser.add_argument(
        "-m", "--meta", action="append", dest="meta", default=[],
        help="Sets a meta data item. This option may be repeated. "
        "Example: -m Color:Blue -m Size:Large")
    parser.add_argument(
        "-H", "--header", action="append", dest="header",
        default=[], help='Adds a customized request header. '
        'This option may be repeated. '
        'Example: -H "content-type:text/plain" '
        '-H "Content-Length: 4000"')

    # We return the parser to build up the bash_completion
    if return_parser:
        return parser

    (options, args) = parse_args(parser, args)
    args = args[1:]
    if (options["read_acl"] or options["write_acl"] or options["sync_to"] or
            options["sync_key"]) and not args:
        exit("-r, -w, -t, and -k options only allowed for containers")

    with SwiftService(options=options) as swift:
        try:
            if not args:
                result = swift.post()
            else:
                container = args[0]
                if "/" in container:
                    output_manager.error(
                        "WARNING: / in container name; you might have "
                        "meant '%s' instead of '%s'." %
                        (args[0].replace("/", " ", 1), args[0]))
                    return
                args = args[1:]
                if args:
                    if len(args) == 1:
                        objects = [args[0]]
                        results_iterator = swift.post(
                            container=container, objects=objects,
                        )
                        result = next(results_iterator)
                    else:
                        output_manager.error(
                            "Usage: %s post %s\n%s", BASENAME,
                            st_post_options, st_post_help)
                        return
                else:
                    result = swift.post(container=container)
            if not result["success"]:
                raise(result["error"])

        except SwiftError as e:
            output_manager.error(e.value)


st_copy_options = """[--destination </container/object>] [--fresh-metadata]
                  [--meta <name:value>] [--header <header>] <container>
                  <object> [<object>] [...]
"""

st_copy_help = """
Copies object to new destination, optionally updates objects metadata.
If destination is not set, will update metadata of object

Positional arguments:
  <container>             Name of container to copy from.
  <object>                Name of object to copy. Specify multiple times
                          for multiple objects

Optional arguments:
  -d, --destination </container[/object]>
                        The container and name of the destination object. Name
                        of destination object can be omitted, then will be
                        same as name of source object. Supplying multiple
                        objects and destination with object name is invalid.
  -M, --fresh-metadata  Copy the object without any existing metadata,
                        If not set, metadata will be preserved or appended
  -m, --meta <name:value>
                        Sets a meta data item. This option may be repeated.
                        Example: -m Color:Blue -m Size:Large
  -H, --header <header:value>
                        Adds a customized request header.
                        This option may be repeated. Example
                        -H "content-type:text/plain" -H "Content-񐯒ength: 4000"
""".strip("\n")


def st_copy(parser, args, output_manager, return_parser=False):
    parser.add_argument(
