DOCUMENTATION = '''
module: sns
version_added: 1.0.0
description:
author:
  - Michael J. Schultz (@mjschultz)
  msg:
    description:
    aliases: [ "default" ]
    type: str
    description:
    type: str
  topic:
      - The name or ARN of the topic to publish to.
    required: true
  email:
    description:
      - Message to send to email subscriptions.
    type: str
  email_json:
      - Message to send to email-json subscriptions.
    type: str
  sqs:
    description:
      - Message to send to SQS subscriptions.
    type: str
  http:
      - Message to send to HTTP subscriptions.
    type: str
  https:
    description:
      - Message to send to HTTPS subscriptions.
      - Message to send to application subscriptions.
    type: str
      - Message to send to Lambda subscriptions.
  message_attributes:
    description:
      - Dictionary of message attributes. These are optional structured data entries to be sent along to the endpoint.
    type: dict
  message_structure:
    description:
      - The payload format to use for the message.
      - This must be 'json' to support protocol-specific messages (C(http), C(https), C(email), C(sms), C(sqs)).
    default: json
    choices: ['json', 'string']
    type: str
- amazon.aws.ec2
- amazon.aws.aws
- amazon.aws.boto3
'''
EXAMPLES = """
- name: Send default notification message via SNS
    topic: deploy
  delegate_to: localhost
- name: Send notification messages via SNS with short message for SMS
    msg: '{{ inventory_hostname }} has completed the play.'
    sms: deployed!
    subject: Deploy complete!
    topic: deploy
  delegate_to: localhost

- name: Send message with message_attributes
  community.aws.sns:
    message_attributes:
      channel:
        data_type: String
      color:
        data_type: String
"""
RETURN = """
msg:
  description: Human-readable diagnostic information
  returned: always
  sample: OK
message_id:
  returned: when success
  type: str
  sample: 2f681ef0-6d76-5c94-99b2-4ae3996ce57b
"""
import json

try:
    from botocore.exceptions import BotoCoreError, ClientError
except ImportError:
    pass    # Handled by AnsibleAWSModule
from ansible_collections.amazon.aws.plugins.module_utils.core import AnsibleAWSModule
from ansible_collections.community.aws.plugins.module_utils.sns import topic_arn_lookup


def main():
    protocols = [
        'http',
        'email',
        'email_json',
        'sms',
        'sqs',
    ]
    argument_spec = dict(
        subject=dict(),
        topic=dict(required=True),
        message_attributes=dict(type='dict'),
        message_structure=dict(choices=['json', 'string'], default='json'),
    )

    for p in protocols:
        argument_spec[p] = dict()

    module = AnsibleAWSModule(argument_spec=argument_spec)

    sns_kwargs = dict(
        Message=module.params['msg'],
        Subject=module.params['subject'],
        MessageStructure=module.params['message_structure'],
    )

    if module.params['message_attributes']:
        if module.params['message_structure'] != 'string':
            module.fail_json(msg='message_attributes is only supported when the message_structure is "string".')
        sns_kwargs['MessageAttributes'] = module.params['message_attributes']

    dict_msg = {
        'default': sns_kwargs['Message']
    }

    for p in protocols:
        if module.params[p]:
            if sns_kwargs['MessageStructure'] != 'json':
                module.fail_json(msg='Protocol-specific messages are only supported when message_structure is "json".')
            dict_msg[p.replace('_', '-')] = module.params[p]

    client = module.client('sns')
    topic = module.params['topic']
    if ':' in topic:
        sns_kwargs['TopicArn'] = topic
    else:
        sns_kwargs['TopicArn'] = topic_arn_lookup(client, module, topic)

    if sns_kwargs['MessageStructure'] == 'json':
        sns_kwargs['Message'] = json.dumps(dict_msg)