from __future__ import unicode_literals
import itertools
import re

import jellyfish

FIPS_RE = re.compile(r"^\d{2}$")
ABBR_RE = re.compile(r"^[a-zA-Z]{2}$")


_lookup_cache = {}


class State(object):
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            self.__dict__[k] = v

    def __repr__(self):
        return "<State:%s>" % self.name

    def __str__(self):
        return self.name

    def shapefile_urls(self, region=None):

        if not self.fips:
            return {}

        base_url = "https://www2.census.gov/geo/tiger/TIGER2010"
        urls = {
            "tract": "{0}/TRACT/2010/tl_2010_{1}_tract10.zip".format(
                base_url, self.fips,
            ),
            "cd": "{0}/CD/111/tl_2010_{1}_cd111.zip".format(base_url, self.fips),
            "county": "{0}/COUNTY/2010/tl_2010_{1}_county10.zip".format(
                base_url, self.fips,
            ),
            "state": "{0}/STATE/2010/tl_2010_{1}_state10.zip".format(
                base_url, self.fips,
            ),
            "zcta": "{0}/ZCTA5/2010/tl_2010_{1}_zcta510.zip".format(
                base_url, self.fips,
            ),
            "block": "{0}/TABBLOCK/2010/tl_2010_{1}_tabblock10.zip".format(
                base_url, self.fips,
            ),
            "blockgroup": "{0}/BG/2010/tl_2010_{1}_bg10.zip".format(
                base_url, self.fips,
            ),
        }

        if region and region in urls:
            return urls[region]

        return urls


def lookup(val, field=None, use_cache=True):
    """ Semi-fuzzy state lookup. This method will make a best effort
        attempt at finding the state based on the lookup value provided.

          * two digits will search for FIPS code
          * two letters will search for state abbreviation
          * anything else will try to match the metaphone of state names

        Metaphone is used to allow for incorrect, but phonetically accurate,
        spelling of state names.

        Exact matches can be done on any attribute on State objects by passing
        the `field` argument. This skips the fuzzy-ish matching and does an
        exact, case-sensitive comparison against the specified field.

        This method caches non-None results, but can the cache can be bypassed
        with the `use_cache=False` argument.
    """

    if field is None:
        if FIPS_RE.match(val):
            field = "fips"
        elif ABBR_RE.match(val):
            val = val.upper()
            field = "abbr"
        else:
            val = jellyfish.metaphone(val)
            field = "name_metaphone"

    # see if result is in cache
    cache_key = "%s:%s" % (field, val)
    if use_cache and cache_key in _lookup_cache:
        return _lookup_cache[cache_key]

    for state in itertools.chain(STATES_AND_TERRITORIES, [DC], OBSOLETE):
        if val == getattr(state, field):
            _lookup_cache[cache_key] = state
            return state


def mapping(from_field, to_field, states=None):
    if states is None:
        states = itertools.chain(STATES_AND_TERRITORIES, [DC], OBSOLETE)
    return dict((getattr(s, from_field), getattr(s, to_field)) for s in states)


AL = State(
    **{
        "fips": "01",
        "name": "Alabama",
        "abbr": "AL",
        "is_territory": False,
        "is_obsolete": False,
        "is_contiguous": True,
        "is_continental": True,
        "statehood_year": 1819,
        "capital": "Montgomery",
        "capital_tz": "America/Chicago",
        "ap_abbr": "Ala.",
        "time_zones": ["America/Chicago"],
        "name_metaphone": "ALBM",
    },
)


AK = State(
    **{
        "fips": "02",
        "name": "Alaska",
        "abbr": "AK",
        "is_territory": False,
        "is_obsolete": False,
        "is_contiguous": False,
        "is_continental": True,
        "statehood_year": 1959,
        "capital": "Juneau",
        "capital_tz": "America/Anchorage",
        "ap_abbr": "Alaska",
        "time_zones": ["America/Anchorage", "America/Adak"],
        "name_metaphone": "ALSK",
    },
)


AS = State(
    **{
        "fips": "60",
        "name": "American Samoa",
        "abbr": "AS",
        "is_territory": True,
        "is_obsolete": False,
        "is_contiguous": False,
        "is_continental": False,
        "statehood_year": None,
        "capital": "Pago Pago",
        "capital_tz": "Pacific/Samoa",
        "ap_abbr": None,
        "time_zones": ["Pacific/Samoa"],
        "name_metaphone": "AMRKN SM",
    },
)


AZ = State(
    **{
        "fips": "04",
        "name": "Arizona",
        "abbr": "AZ",
        "is_territory": False,
        "is_obsolete": False,
        "is_contiguous": True,
        "is_continental": True,
        "statehood_year": 1912,
        "capital": "Phoenix",
        "capital_tz": "America/Phoenix",
        "ap_abbr": "Ariz.",
        "time_zones": ["America/Phoenix"],
        "name_metaphone": "ARSN",
    },
)


AR = State(
    **{
        "fips": "05",
        "name": "Arkansas",
        "abbr": "AR",
        "is_territory": False,
        "is_obsolete": False,
        "is_contiguous": True,
        "is_continental": True,
        "statehood_year": 1836,
        "capital": "Little Rock",
        "capital_tz": "America/Chicago",
        "ap_abbr": "Ark.",
        "time_zones": ["America/Chicago"],
        "name_metaphone": "ARKNSS",
    },
)


CA = State(
    **{
        "fips": "06",
        "name": "California",
        "abbr": "CA",
        "is_territory": False,
        "is_obsolete": False,
        "is_contiguous": True,
        "is_continental": True,
        "statehood_year": 1850,
        "capital": "Sacramento",
        "capital_tz": "America/Los_Angeles",
        "ap_abbr": "Calif.",
        "time_zones": ["America/Los_Angeles"],
        "name_metaphone": "KLFRN",
    },
)


CO = State(
    **{
        "fips": "08",
        "name": "Colorado",
        "abbr": "CO",
        "is_territory": False,
        "is_obsolete": False,
        "is_contiguous": True,
        "is_continental": True,
        "statehood_year": 1876,
        "capital": "Denver",
        "capital_tz": "America/Denver",
        "ap_abbr": "Colo.",
        "time_zones": ["America/Denver"],
        "name_metaphone": "KLRT",
    },
)


CT = State(
    **{
        "fips": "09",
        "name": "Connecticut",
        "abbr": "CT",
        "is_territory": False,
        "is_obsolete": False,
        "is_contiguous": True,
        "is_continental": True,
        "statehood_year": 1788,
        "capital": "Hartford",
        "capital_tz": "America/New_York",
        "ap_abbr": "Conn.",
        "time_zones": ["America/New_York"],
        "name_metaphone": "KNKTKT",
    },
)


DK = State(
    **{
        "fips": None,
        "name": "Dakota",
        "abbr": "DK",
        "is_territory": False,
        "is_obsolete": True,
        "is_contiguous": True,
        "is_continental": True,
        "statehood_year": None,
        "capital": "Yankton",
        "capital_tz": "America/Chicago",
        "ap_abbr": None,
        "time_zones": ["America/Chicago"],
        "name_metaphone": "TKT",
    },
)


DE = State(
    **{
        "fips": "10",
        "name": "Delaware",
        "abbr": "DE",
        "is_territory": False,
        "is_obsolete": False,
        "is_contiguous": True,
        "is_continental": True,
        "statehood_year": 1787,
        "capital": "Dover",
        "capital_tz": "America/New_York",
        "ap_abbr": "Del.",
        "time_zones": ["America/New_York"],
        "name_metaphone": "TLWR",
    },
)


DC = State(
    **{
        "fips": "11",
        "name": "District of Columbia",
        "abbr": "DC",
        "is_territory": False,
        "is_obsolete": False,
        "is_contiguous": True,
        "is_continental": True,
        "statehood_year": None,
        "capital": None,
        "capital_tz": "America/New_York",
        "ap_abbr": "D.C.",
        "time_zones": ["America/New_York"],
        "name_metaphone": "TSTRKT OF KLMB",
    },
)


FL = State(
    **{
        "fips": "12",
        "name": "Florida",
        "abbr": "FL",
        "is_territory": False,
        "is_obsolete": False,
        "is_contiguous": True,
        "is_continental": True,
        "statehood_year": 1845,
        "capital": "Tallahassee",
        "capital_tz": "America/New_York",
        "ap_abbr": "Fla.",
        "time_zones": ["America/New_York", "America/Chicago"],
        "name_metaphone": "FLRT",
    },
)


GA = State(
    **{
        "fips": "13",
        "name": "Georgia",
        "abbr": "GA",
        "is_territory": False,
        "is_obsolete": False,
        "is_contiguous": True,
        "is_continental": True,
        "statehood_year": 1788,
        "capital": "Atlanta",
        "capital_tz": "America/New_York",
        "ap_abbr": "Ga.",
        "time_zones": ["America/New_York"],
        "name_metaphone": "JRJ",
    },
)


GU = State(
    **{
        "fips": "66",
        "name": "Guam",
        "abbr": "GU",
        "is_territory": True,
        "is_obsolete": False,
        "is_contiguous": False,
        "is_continental": False,
        "statehood_year": None,
        "capital": "Hagåtña",
        "capital_tz": "Pacific/Guam",
        "ap_abbr": None,
        "time_zones": ["Pacific/Guam"],
        "name_metaphone": "KM",
    },
)


HI = State(
    **{
        "fips": "15",
        "name": "Hawaii",
        "abbr": "HI",
        "is_territory": False,
        "is_obsolete": False,
        "is_contiguous": False,
        "is_continental": False,
        "statehood_year": 1959,
        "capital": "Honolulu",
        "capital_tz": "Pacific/Honolulu",
        "ap_abbr": "Hawaii",
        "time_zones": ["Pacific/Honolulu"],
        "name_metaphone": "HW",
    },
)


ID = State(
    **{
        "fips": "16",
        "name": "Idaho",
        "abbr": "ID",
        "is_territory": False,
        "is_obsolete": False,
        "is_contiguous": True,
        "is_continental": True,
        "statehood_year": 1890,
        "capital": "Boise",
        "capital_tz": "America/Denver",
        "ap_abbr": "Idaho",
        "time_zones": ["America/Denver", "America/Los_Angeles"],
        "name_metaphone": "ITH",
    },
)


IL = State(
    **{
        "fips": "17",
        "name": "Illinois",
        "abbr": "IL",
        "is_territory": False,
        "is_obsolete": False,
        "is_contiguous": True,
        "is_continental": True,
        "statehood_year": 1818,
        "capital": "Springfield",
        "capital_tz": "America/Chicago",
        "ap_abbr": "Ill.",
        "time_zones": ["America/Chicago"],
        "name_metaphone": "ILNS",
    },
)


IN = State(
    **{
        "fips": "18",
        "name": "Indiana",
        "abbr": "IN",
        "is_territory": False,
        "is_obsolete": False,
        "is_contiguous": True,
        "is_continental": True,
        "statehood_year": 1816,
        "capital": "Indianapolis",
        "capital_tz": "America/Indiana/Indianapolis",
        "ap_abbr": "Ind.",
        "time_zones": [
            "America/Chicago",
            "America/Indiana/Indianapolis",
            "America/Indianapolis",
            "America/Indiana/Winamac",
            "America/Indiana/Vincennes",
            "America/Indiana/Vevay",
            "America/Indiana/Tell_City",
            "America/Indiana/Petersburg",
            "America/Indiana/Marengo",
            "America/Indiana/Knox",
            "America/Knox_IN",
            "America/New_York",
        ],
        "name_metaphone": "INTN",
    },
)


IA = State(
    **{
        "fips": "19",
        "name": "Iowa",
        "abbr": "IA",
        "is_territory": False,
        "is_obsolete": False,
        "is_contiguous": True,
        "is_continental": True,
        "statehood_year": 1846,
        "capital": "Des Moines",
        "capital_tz": "America/Chicago",
        "ap_abbr": "Iowa",
        "time_zones": ["America/Chicago"],
        "name_metaphone": "IW",
    },
)


KS = State(
    **{
        "fips": "20",
        "name": "Kansas",
        "abbr": "KS",
        "is_territory": False,
        "is_obsolete": False,
        "is_contiguous": True,
        "is_continental": True,
        "statehood_year": 1861,
        "capital": "Topeka",
        "capital_tz": "America/Chicago",
        "ap_abbr": "Kan.",
        "time_zones": ["America/Chicago", "America/Denver"],
        "name_metaphone": "KNSS",
    },
)


KY = State(
    **{
        "fips": "21",
        "name": "Kentucky",
        "abbr": "KY",
        "is_territory": False,
        "is_obsolete": False,
        "is_contiguous": True,
        "is_continental": True,
        "statehood_year": 1792,
        "capital": "Frankfort",
        "capital_tz": "America/New_York",
        "ap_abbr": "Ky.",
        "time_zones"
