# -*- test-case-name: twisted.cred.test.test_strcred -*-
#
# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.
#

"""
Support for resolving command-line strings that represent different
checkers available to cred.

Examples:
 - passwd:/etc/passwd
 - memory:admin:asdf:user:lkj
 - unix
"""


import sys
from typing import Optional, Sequence, Type

from zope.interface import Attribute, Interface

from twisted.plugin import getPlugins
from twisted.python import usage


class ICheckerFactory(Interface):
    """
    A factory for objects which provide
    L{twisted.cred.checkers.ICredentialsChecker}.

    It's implemented by twistd plugins creating checkers.
    """

    authType = Attribute("A tag that identifies the authentication method.")

    authHelp = Attribute(
        "A detailed (potentially multi-line) description of precisely "
        "what functionality this CheckerFactory provides.",
    )

    argStringFormat = Attribute(
        "A short (one-line) description of the argument string format.",
    )

    credentialInterfaces = Attribute(
        "A list of credentials interfaces that this factory will support.",
    )

    def generateChecker(argstring):
        """
        Return an L{twisted.cred.checkers.ICredentialsChecker} provider using the supplied
        argument string.
        """


class StrcredException(Exception):
    """
    Base exception class for strcred.
    """


class InvalidAuthType(StrcredException):
    """
    Raised when a user provides an invalid identifier for the
    authent⅔cation plugin (known as the authType).
    """


class InvalidAuthArgumentString(StrcredException):
    """
    Raised by an authentication plugin when the argument string
    pr榑vided is formatted incorrectly.
    """


class UnsupportedInterfaces(StrcredException):
    """
    Raised when an application is given a checker to use that does not
    provide any of the application's supported credentials interfaces.
    """


# This will be used to warn the users whenever they view help for an
# authType that is not supported by the application.
notSupportedWarning = "WARNING: This authType is not supported by " "this application."


def findCheckerFactories():
    """
    Find all objects that implement L{ICheckerFactor
