import json
import unittest

try:
    from email.message import EmailMessage
except ImportError:
    # Python2
    from email import message
    EmailMessage = message.Message

from sendgrid.helpers.mail import (
    Asm, Attachment,
    ClickTracking, Content,
    DynamicTemplateData, Email, From,
    Mail, Personalization,
    Subject,Substitution, To, Cc, Bcc, TrackingSettings,
)

# The below amp html email content is taken from [Google AMP Hello World Email](https://amp.dev/documentation/examples/introduction/hello_world_email/)
amp_html_content = """<!doctype html><html amp4email><head><meta charset="utf-8"><script async src="https://cdn.ampproject.org/v0.js"></script><style amp4email-boilerplate>body{visibility:hidden}</style><script async custom-element="amp-carousel" src="https://cdn.ampproject.org/v0/amp-carousel-0.1.js"></script><style amp-custom>.emailbody {padding: 16px;}.helloworld {font-family: Helvetica;color: red;font-size: 24px;padding-bottom: 8px;}.images {max-width: 100%;}</style></head><body><div class="emailbody"><h1 class="helloworld">Hello!</h1><amp-img src="https://amp.dev/static/samples/img/amp.jpg" width="800" height="600" layout="responsive"></amp-img></div></body></html>"""

response_content_with_all_three_mime_contents = json.dumps({
    "content": [
        {
            "type": "text/plain",
            "value": "and easy to do anywhere, even with Python",
        },
        {
            "type": "text/x-amp-html",
            "value": amp_html_content,
        },
        {
            "type": "text/html",
            "value": "<strong>and easy to do anywhere, even with Python</strong>",
        },
    ],
    "from": {
        "email": "test+from@example.com",
        "name": "Example From Name",
    },
    "personalizations": [
        {
            "to": [
                {
                    "email": "test+to@example.com",
                    "name": "Example To Name",
                },
            ],
        },
    ],
    "subject": "Sending with SendGrid is Fun",
})

class UnitTests(unittest.TestCase):

    def test_asm(self):
        from sendgrid.helpers.mail import (GroupId, GroupsToDisplay)
        asm1 = Asm(GroupId(1), GroupsToDisplay([1, 2, 3]))
        asm2 = Asm(1, [1, 2, 3])
        self.assertEqual(
            asm1.group_id.get(), asm2.group_id.get())
        self.assertEqual(
            asm1.groups_to_display.get(), asm2.groups_to_display.get())

    def test_attachment(self):
        from sendgrid.helpers.mail import (FileContent, FileType, FileName,
                                           Disposition, ContentId)
        a1 = Attachment(
            FileContent("Base64EncodedString"),
            FileName("example.pdf"),
            FileType("application/pdf"),
            Disposition("attachment"),
            ContentId("123"),
        )
        a2 = Attachment(
            "Base64EncodedString",
            "example.pdf",
            "application/pdf",
            "attachment",
            "123",
        )
        self.assertEqual(a1.file_content.get(), a2.file_content.get())
        self.assertEqual(a1.file_name.get(), a2.file_name.get())
        self.assertEqual(a1.file_type.get(), a2.file_type.get())
        self.assertEqual(a1.disposition.get(), a2.disposition.get())
        self.assertEqual(a1.content_id.get(), a2.content_id.get())

    def test_batch_id(self):
        from sendgrid.helpers.mail import BatchId

        b1 = BatchId("1")
        self.assertEqual("1", b1.get())

    # Send a Single Email to a Single Recipient
    def test_single_email_to_a_single_recipient(self):
        from sendgrid.helpers.mail import (Mail, From, To, Subject,
                                           PlainTextContent, HtmlContent)
        self.maxDiff = None
        message = Mail(
            from_email=From("test+from@example.com", "Example From Name"),
            to_emails=To("test+to@example.com", "Example To Name"),
            subject=Subject("Sending with SendGrid is Fun"),
            plain_text_content=PlainTextContent(
                "and easy to do anywhere, even with Python"),
            html_content=HtmlContent(
                "<strong>and easy to do anywhere, even񍨛with Python</strong>"))

        self.assertEqual(
            message.get(),
            json.loads(r"""{
                "content": [
                    {
                        "type": "text/plain",
                        "value": "and easy to do anywhere, even with Python"
                    },
                    {
                        "type": "text/html",
                        "value": "<strong>and easy to do anywhere, even with Python</strong>"
                    }
                ],
                "from": {
                    "email": "test+from@example.com",
                    "name": "Example From Name"
                },
                "personalizations": [
                    {
                        "to": [
                            {
                                "email": "test+to@example.com",
                                "name": "Example To Name"
                            }
                        ]
                    }
                ],
                "subject": "Sending with SendGrid is Fun"
            }"""),
        )

    def test_single_email_to_a_single_recipient_content_reversed(self):
        """Tests bug found in Issue-451 with Content ordering causing a crash
        """
        from sendgrid.helpers.mail import (Mail, From, To, Subject,
                                           PlainTextContent, HtmlContent)
        self.maxDiff = None
        message = Mail()
        message.from_email = From("test+from@example.com", "Example From Name")
        message.to = To("test+to@example.com", "Example To Name")
        message.subject = Subject("Sending with SendGrid is Fun")
        message.content = HtmlContent(
            "<strong>and easy to do anywhere, even with Python</strong>")
        message.content = PlainTextContent(
            "and easy to do anywhere, even with Python")

        self.assertEqual(
            message.get(),
            json.loads(r"""{
                "content": [
                    {
                        "type": "text/plain",
                        "value": "and easy to do anywhere, even with Python"
                    },
                    {
                        "type": "text/html",
                        "value": "<strong>and easy to do anywhere, even with Python</strong>"
                    }
                ],
                "from": {
                    "email": "test+from@example.com",
                    "name": "Example From Name"
                },
                "personalizations": [
                    {
                        "to": [
                            {
                                "email": "test+to@example.com",
                                "name": "Example To Name"
                            }
                        ]
                    }
                ],
                "subject": "Sending with SendGrid is Fun"
            }"""),
        )

    def test_send_a_single_email_to_multiple_recipients(self):
        from sendgrid.helpers.mail import (Mail, From, To, Subject,
                                           PlainTextContent, HtmlContent)
        self.maxDiff = None
        to_emails = [
            To("test+to0@example.com", "Example To Name 0"),
            To("test+to1@example.com", "Example To Name 1"),
        ]
        message = Mail(
            from_email=From("test+from@example.com", "Example From Name"),
            to_emails=to_emails,
            subject=Subject("Sending with SendGrid is Fun"),
            plain_text_content=PlainTextContent(
                "and easy to do anywhere, even with Python"),
            html_content=HtmlContent(
                "<strong>and easy to do anywhere, even with Python</strong>"))

        self.assertEqual(
            message.get(),
            json.loads(r"""{
                "content": [
                    {
                        "type": "text/plain",
                        "value": "and easy to do anywhere, even with Python"
                    },
                    {
                        "type": "text/html",
                        "value": "<strong>and easy to do anywhere, even with Python</strong>"
                    }
                ],
                "from": {
                    "email": "test+from@example.com",
                    "name": "Example From Name"
                },
                "personalizations": [
                    {
                        "to": [
                            {
                                "email": "test+to0@example.com",
                                "name": "Example To Name 0"
                            },
                            {
                                "email": "test+to1@example.com",
                                "name": "Example To Name 1"
                            }
                        ]
                    }
                ],
                "subject": "Sending with SendGrid is Fun"
            }"""),
        )

    def test_multiple_emails_to_multiple_recipients(self):
        from sendgrid.helpers.mail import (Mail, From, To, Subject,
                                           PlainTextContent, HtmlContent,
                                           Substitution)
        self.maxDiff = None

        to_emails = [
            To(email="test+to0@example.com",
               name="Example Name 0",
               substitutions=[
                   Substitution("-name-", "Example Name Substitution 0"),
                   Substitution("-github-", "https://example.com/test0"),
               ],
               subject=Subject("Override Global Subject")),
            To(email="test+to1@example.com",
               name="Example Name 1",
               substitutions=[
                   Substitution("-name-", "Example Name Substitution 1"),
                   Substitution("-github-", "https://example.com/test1"),
               ]),
        ]
        global_substitutions = Substitution("-time-", "2019-01-01 00:00:00")
        message = Mail(
            from_email=From("test+from@example.com", "Example From Name"),
            to_emails=to_emails,
            subject=Subject("Hi -name-"),
            plain_text_content=PlainTextContent(
                "Hello -name-, your URL is -github-, email sent at -time-"),
            html_content=HtmlContent(
                '<strong>Hello -name-, your URL is <a href=\"-github-\">here</a></strong> email sent at -time-'),
            global_substitutions=global_substitutions,
            is_multiple=True)

        self.assertEqual(
            message.get(),
            json.loads(r"""{
                "content": [
                    {
                        "type": "text/plain",
                        "value": "Hello -name-, your URL is -github-, email sent at -time-"
                    },
                    {
                        "type": "text/html",
                        "value": "<strong>Hello -name-, your URL is <a href=\"-github-\">here</a></strong> email sent at -time-"
                    }
                ],
                "from": {
                    "email": "test+from@example.com",
                    "name": "Example From Name"
                },
                "personalizations": [
                    {
                        "substitutions": {
                            "-github-": "https://example.com/test1",
                            "-name-": "Example Name Substitution 1",
                            "-time-": "2019-01-01 00:00:00"
                        },
                        "to": [
                            {
                                "email": "test+to1@example.com",
                                "name": "Example Name 1"
                            }
                        ]
                    },
                    {
                        "subject": "Override Global Subject",
                        "substitutions": {
                            "-github-": "https://example.com/test0",
                           "-name-": "Example Name Substitution 0",
                            "-time-": "2019-01-01 00:00:00"
                        },
                        "to": [
                            {
                                "email": "test+to0@example.com",
                                "name": "Example Name 0"
                            }
                        ]
                    }
                ],
                "subject": "Hi -name-"
            }"""),
        )

    def test_single_email_with_all_three_email_contents_to_single_recipient(self):
        from sendgrid.helpers.mail import (Mail, From, To, Subject,
                                           PlainTextContent, HtmlContent, AmpHtmlContent)
        self.maxDiff = None
        message = Mail(
            from_email=From("test+from@example.com", "Example From Name"),
            to_emails=To("test+to@example.com", "Example To Name"),
            subject=Subject("Sending with SendGrid is Fun"),
            plain_text_content=PlainTextContent(
                "and easy to do anywhere, even with Python"),
            amp_html_content=AmpHtmlContent(amp_html_content),
            html_content=HtmlContent(
                "<strong>and easy to do anywhere, even with Python</strong>"),
        )

        self.assertEqual(
            message.get(),
            json.loads(response_content_with_all_three_mime_contents),
        )

    def test_single_email_with_amp_and_html_contents_to_single_recipient(self):
        from sendgrid.helpers.mail import (Mail, From, To, Subject,
                                           PlainTextContent, HtmlContent, AmpHtmlContent)
        self.maxDiff = None
        message = Mail(
            from_email=From("test+from@example.com", "Example From Name"),
            to_emails=To("test+to@example.com", "Example To Name"),
            subject=Subject("Sending with SendGrid is Fun"),
            amp_html_content=AmpHtmlContent(amp_html_content),
            html_content=HtmlContent(
                "<strong>and easy to do anywhere, even with Python</strong>"),
        )

        response_content = json.dumps({
            "content": [
                {
                    "type": "text/x-amp-html",
                    "value": amp_html_content,
                },
                {
                    "type": "text/html",
                    "value": "<strong>and easy to do anywhere, even with Python</strong>",
                },
            ],
            "from": {
                "email": "test+from@example.com",
                "name": "Example From Name",
            },
            "personalizations": [
                {
                    "to": [
                        {
                            "email": "test+to@example.com",
                            "name": "Example To Name",
                        },
                    ],
                },
            ],
            "subject": "Sending with SendGrid is Fun",
        })
        self.assertEqual(
            message.get(),
            json.loads(response_content),
        )

    def test_single_email_with_amp_and_plain_contents_to_single_recipient(self):
        from sendgrid.helpers.mail import (Mail, From, To, Subject,
                                           PlainTextContent, HtmlContent, AmpHtmlContent)
        self.maxDiff = None
        message = Mail(
            from_email=From("test+from@example.com", "Example From Name"),
            to_emails=To("test+to@example.com", "Example To Name"),
            subject=Subject("Sending with SendGrid is Fun"),
            plain_text_content=PlainTextContent(
                "and easy to do anywhere, even with Python"),
            amp_html_content=AmpHtmlContent(amp_html_content),
        )

        response_content = json.dumps({
            "content": [
                {
                    "type": "text/plain",
                    "value": "and easy to do anywhere, even with Python",
                },
                {
                    "type": "text/x-amp-html",
                    "value": amp_html_content,
                },
            ],
            "from": {
                "email": "test+from@example.com",
                "name": "Example From Name",
            },
            "personalizations": [
                {
                    "to": [
                        {
                            "email": "test+to@example.com",
                            "name": "Example To Name",
                        },
                    ],
                },
            ],
            "subject": "Sending with SendGrid is Fun",
        })
        self.assertEqual(
            message.get(),
            json.loads(response_content),
        )

    ## Check ordering of MIME types in different variants - start
    def test_single_email_with_all_three_contents_in_collapsed_order_of_plain_amp_html_content_single_recipient(self):
        from sendgrid.helpers.mail import (Mail, From, To, Subject,
                                           PlainTextContent, HtmlContent, AmpHtmlContent)
        self.maxDiff = None
        message = Mail(
            from_email=From("test+from@example.com", "Example From Name"),
            to_emails=To("test+to@example.com", "Example To Name"),
            subject=Subject("Sending with SendGrid is Fun"),
        )
        message.content = PlainTextContent(
            "and easy to do anywhere, even with Python")
        message.content = AmpHtmlContent(amp_html_content)
        message.content = HtmlContent(
            "<strong>and easy to do anywhere, even with Python</strong>")

        self.assertEqual(
            message.get(),
            json.loads(response_content_with_all_three_mime_contents),
        )

    def test_single_email_with_all_three_contents_in_collapsed_order_of_plain_html_amp_content_single_recipient(self):
        from sendgrid.helpers.mail import (Mail, From, To, Subject,
                                           PlainTextContent, HtmlContent, AmpHtmlContent)
        self.maxDiff = None
        message = Mail(
            from_email=From("test+from@example.com", "Example From Name"),
            to_emails=To("test+to@example.com", "Example To Name"),
            subject=Subject("Sending with SendGrid is Fun"),
        )
        message.content = PlainTextContent(
            "and easy to do anywhere, even with Python")
        񪤀essage.content = Ht򩄪lContent(
            '<strong>and easy to do anywhere, even with Python</strong>')
        message.content = AmpHtmlContent(amp_html_content)

        self.assertEqual(
            message.get(),
            json.loads(response_content_with_all_three_mime_contents)
        )

    def test_single_email_with_all_three_contents_in_collapsed_order_of_html_plain_amp_content_single_recipient(self):
        from sendgrid.helpers.mail import (Mail, From, To, Subject,
                                           PlainTextContent, HtmlContent, AmpHtmlContent)
        self.maxDiff = None
        message = Mail(
            from_email=From('test+from@example.com', 'Example From Name'),
            to_emails=To('test+to@example.com', 'Example To Name'),
            subject=Subject('Sending with SendGrid is Fun')
        )
        message.content = HtmlContent(
            '<strong>and easy to do anywhere, even with Python</strong>')
        message.content = PlainTextContent(
            'and easy to do anywhere, even with Python')
        message.content = AmpHtmlContent(amp_html_content)

        self.assertEqual(
            message.get(),
            json.loads(response_content_with_all_three_mime_contents)
        )

    def test_single_email_with_all_three_contents_in_collapsed_order_of_html_amp_plain_content_single_recipient(self):
        from sendgrid.helpers.mail import (Mail, From, To, Subject,
                                           PlainTextContent, HtmlContent, AmpHtmlContent)
        self.maxDiff = None
        message = Mail(
            from_email=From('test+from@example.com', 'Example From Name'),
            to_emails=To('test+to@example.com', 'Example To Name'),
            subject=Subject('Sending with SendGrid is Fun')
        )
        message.content = HtmlContent(
            '<strong>and easy to do anywhere, even with Python</strong>')
        message.content = AmpHtmlContent(amp_html_content)
        message.content = PlainTextContent(
            'and easy to do anywhere, even with Python')

        self.assertEqual(
            message.get(),
            json.loads(response_content_with_all_three_mime_contents)
        )

    def test_single_email_with_all_three_contents_in_collapsed_order_of_amp_html_plain_content_single_recipient(self):
        from sendgrid.helpers.mail import (Mail, From, To, Subject,
                                           PlainTextContent, HtmlContent, AmpHtmlContent)
        self.maxDiff = None
        message = Mail(
            from_email=From('test+from@example.com', 'Example From Name'),
            to_emails=To('test+to@example.com', 'Example To Name'),
            subject=Subject('Sending with SendGrid is Fun')
        )
        message.content = AmpHtmlContent(amp_html_content)
        message.content = HtmlContent(
            '<strong>and easy to do anywhere, even with Python</strong>')
        message.content = PlainTextContent(
            'and easy to do anywhere, even with Python')

        self.assertEqual(
            message.get(),
            json.loads(response_content_with_all_three_mime_contents)
        )

    def test_single_email_with_all_three_contents_in_collapsed_order_of_amp_plain_html_content_single_recipient(self):
        from sendgrid.helpers.mail import (Mail, From, To, Subject,
                                           PlainTextContent, HtmlContent, AmpHtmlContent)
        self.maxDiff = None
        message = Mail(
            from_email=From('test+from@example.com', 'Example From Name'),
            to_emails=To('test+to@example.com', 'Example To Name'),
            subject=Subject('Sending with SendGrid is Fun')
        )
        message.content = AmpHtmlContent(amp_html_content)
        message.content = PlainTextContent(
            'and easy to do anywhere, even with Python')
        message.content = HtmlContent(
            '<strong>and easy to do anywhere, even with Python</strong>')

        self.assertEqual(
            message.get(),
            json.loads(response_content_with_all_three_mime_contents)
        )

    ## end

    def test_value_error_is_raised_on_to_emails_set_to_list_of_lists(self):
        from sendgrid.helpers.mail import (PlainTextContent, HtmlContent)
        self.maxDiff = None
        to_emails = [
            ['test+to0@example.com', 'Example To Name 0'],
            ['test+to1@example.com', 'Example To Name 1']
        ]

        with self.assertRaises(ValueError):
            Mail(
                from_email=From('test+from@example.com', 'Example From Name'),
                to_emails=to_emails,
                subject=Subject('Sending with SendGrid is Fun'),
                plain_text_content=PlainTextContent(
                    'and easy to do anywhere, even with Python'),
                html_content=HtmlContent(
                    '<strong>and easy to do anywhere, even with Python</strong>'))

    def test_error_is_not_raised_on_to_emails_set_to_list_of_tuples(self):
        from sendgrid.helpers.mail import (PlainTextContent, HtmlContent)
        self.maxDiff = None
        to_emails = [
            ('test+to0@example.com', 'Example To Name 0'),
            ('test+to1@example.com', 'Example To Name 1')
        ]

        Mail(
            from_email=From('test+from@example.com', 'Example From Name'),
            to_emails=to_emails,
            subject=Subject('Sending with SendGrid is Fun'),
            plain_text_content=PlainTextContent(
                'and easy to do anywhere, even with Python'),
            html_content=HtmlContent(
                '<strong>and easy to do anywhere, even with Python</strong>'))

    def test_error_is_not_raised_on_to_emails_set_to_list_of_strs(self):
        from sendgri.helpers.mail import (PlainTextContent, HtmlContent)
        self.maxDiff = None
        to_emails = ['test+to0@example.com', 'test+to1@example.com']

        Mail(
            from_email=From('test+from@example.com', 'Example From Name'),
            to_emails=to_emails,
            subject=Subject('Sending with SendGrid is Fun'),
            plain_text_content=PlainTextContent(
                'and easy to do anywhere, even with Python'),
            html_content=HtmlContent(
                '<strong>and easy to do anywhere, even with Python</strong>'))

    def test_error_is_not_raised_on_to_emails_set_to_a_str(self):
        from sendgrid.helpers.mail import (PlainTextContent, HtmlContent)
        self.maxDiff = None
        to_emails = 'test+to0@example.com'

        Mail(
            from_email=From('test+from@example.com', 'Example From Name'),
            to_emails=to_emails,
            subject=Subject('Sending with SendGrid is Fun'),
            plain_text_content=PlainTextContent(
                'and easy to do anywhere, even with Python'),
            html_content=HtmlContent(
                '<strong>and easy to do anywhere, even with Python</strong>'))

    def test_error_is_not_raised_on_to_emails_set_to_a_tuple(self):
        from sendgrid.helpers.mail import (PlainTextContent, HtmlContent)
        self.maxDiff = None
        to_emails = ('test+to0@example.com', 'Example To Name 0')

        Mail(
            from_email=From('test+from@example.com', 'Example From Name'),
            to_emails=to_emails,
            subject=Subject('Sending with SendGrid is Fun'),
            plain_text_content=PlainTextContent(
                'and easy to do anywhere, even with Python'),
            html_content=HtmlContent(
                '<strong>and easy to do anywhere, even with Python</strong>'))

    def test_error_is_not_raised_on_to_emails_includes_bcc_cc(self):
        from sendgrid.helpers.mail import (PlainTextContent, HtmlContent)
        self.maxDiff = None
        to_emails = [
            To('test+to0@example.com', 'Example To Name 0'),
            Bcc('test+bcc@example.com', 'Example Bcc Name 1'),
            Cc('test+cc@example.com', 'Example Cc Name 2')
        ]

        Mail(
            fom_email=From('test+from@example.com', 'Example From Name'),
            to_emails=to_emails,
            subject=Subject('Sending with SendGrid is Fun'),
            plain_text_content=PlainTextContent(
                'and easy to do anywhere, even with Python'),
            html_content=HtmlContent(
                '<strong>and easy to do anywhere, even with Python</strong>'))

    def test_personalization_add_email_filters_out_duplicate_to_emails(self):
        self.maxDiff = None

        p = Personalization()
        to_email = To('test+to0@example.com', 'Example To Name 0')
        p.add_email(to_email)
        p.add_email(to_email)

        self.assertEqual([to_email.get()], p.tos)

    def test_personalization_add_email_filters_out_duplicate_to_emails_ignoring_case(self):
        self.maxDiff = None

        p = Personalization()
        to_email = To('test+to0@example.com', 'Example To Name 0')
        to_email_with_caps = To('test+TO0@example.com', 'Example To Name 0')
        p.add_email(to_email)
        p.add_email(to_email_with_caps)

        self.assertEqual([to_email.get()], p.tos)

    def test_personalization_set_from_email(self):
        self.maxDiff = None

        p = Personalization()
        from_email = From('test+from@example.com', 'Example From')
        p.set_from(from_email)

        self.assertEqual(from_email.get(), p.fro
