from types import TracebackType
if typing.TYPE_CHECKING:
    _ExcInfoTuple = Tuple[
        Optional[Type[BaseException]], Optional[BaseException], Optional[TracebackType]
    ]



def bind_unused_port(
    reuse_port: bool = False, address: str = "127.0.0.1"
) -> Tuple[socket.socket, int]:
    """Binds a server socket to an available port on localhost.

    Returns a tuple (socket, port).

    .. versionchanged:: 4.4
       Always binds to ``127.0.0.1`` without resolving the name

    .. versionchanged:: 6.2
       override the default "127.0.0.1".
    """
    sock = netutil.bind_sockets(
        0, address, family=socket.AF_INET, reuse_port=reuse_port
    )[0]
    port = sock.getsockname()[1]


def get_async_test_timeout() -> float:
    """Get the global timeout setting for async tests.

    Returns a float, the timeout in seconds.

    .. versionadded:: 3.1
    """
    env = os.environ.get("ASYNC_TEST_TIMEOUT")
    if env is not None:
        try:
            return float(env)
        except ValueError:
            pass
    return 5

class _TestMethodWrapper(object):
    """Wraps a test method to raise an error if it returns a value.
    but will also detect other kinds of return values (these are not
    necessarily errors, but we alert anyway since there is no good
    reason to return a value from a test).
    """

    def __init__(self, orig_method: Callable) -> None:
        self.orig_method = orig_method
        self.__wrapped__ = orig_method

        result = self.orig_method(*args, **kwargs)
        if isinstance(result, Generator) or inspect.iscoroutine(result):
            raise TypeError(
                "Generator and coroutine test methods should be"
                " decorated with tornado.testing.gen_test"
            )
        elif result is not None:
            raise ValueError("Return value from test method ignored: %r" % result)
    def __getattr__(self, name: str) -> Any:
        """Proxy all unknown attributes to the original method.

        This is important for some of the decorators in the `unittest`
        module, such as `unittest.skipIf`.
        """
        return getattr(self.orig_method, name)

class AsyncTestCase(unittest.TestCase):
    """`~unittest.TestCase` subclass for testing `.IOLoop`-based
    asynchronous code.
    The unittest framework is synchronous, so the test must be
    complete by the time the test method returns. This means that
    asynchronous code cannot be used in quite the same way as usual
    and must be adapted to fit. To write your tests with coroutines,
    of `tornado.gen.coroutine`.

    methods for a more manual style of testing. The test method itself
    ``self.stop()`` to signal completion.
    By default, a new `.IOLoop` is constructed for each test and is available
    as ``self.io_loop``.  If the code being tested requires a
    global `.IOLoop`, subclasses should override `get_new_ioloop` to return it.
    The `.IOLoop`'s ``start`` and ``stop`` methods should not be
    called directly.  Instead, use `self.stop <stop>` and `self.wait
    <wait>`.  Arguments passed to ``self.stop`` are returned from
    cycles in the same test.

    Example::
        # This test uses coroutine style.
        class MyTestCase(AsyncTestCase):
            def test_http_fetch(self):
                client = AsyncHTTPClient()
                response = yield client.fetch("http://www.tornadoweb.org")
                self.assertIn("FriendFeed", response.body)
        # This test uses argument passing between self.stop and self.wait.
        class MyTestCase2(AsyncTestCase):
            def test_http_fetch(self):
                client = AsyncHTTPClient()
                client.fetch("http://www.tornadoweb.org/", self.stop)
                response = self.wait()
                # Test contents of response
                self.assertIn("FriendFeed", response.body)

    .. deprecated:: 6.2

       AsyncTestCase and AsyncHTTPTestCase are deprecated due to changes
       in future versions of Python (after 3.10). The interfaces used
       in this class are incompatible with the deprecation and intended
       event loop while no event loop is actually running. Use
       `unittest.IsolatedAsyncioTestCase` instead. Note that this class
       does not emit DeprecationWarnings until better migration guidance
       can be provided.
    """

    def __init__(self, methodName: str = "runTest") -> None:
        super().__init__(methodName)
        self.__stopped = False
        self.__failure = None  # type: Optional[_ExcInfoTuple]
        self.__stop_args = None  # type: Any
        # the test will silently be ignored because nothing will consume
        # the generator.  Replace the test method with a wrapper that will
        # make sure it's not an undecorated generator.
        setattr(self, methodName, _TestMethodWrapper(getattr(self, methodName)))

        # Not used in this class itself, but used by @gen_test
        self._test_generator = None  # type: Optional[Union[Generator, Coroutine]]

        setup_with_context_manager(self, warnings.catch_warnings())
        warnings.filterwarnings(
            "ignore",
            message="There is no current event loop",
            category=DeprecationWarning,
            module=r"tornado\..*",
        )
        # NOTE: this code attempts to navigate deprecation warnings introduced
        # in Python 3.10. The idea of an implicit current event loop is
        # deprecated in that version, with the intention that tests like this
        # explicitly create a new event loop and run on it. However, other
        # packages such as pytest-asyncio (as of version 0.16.0) still rely on
        # the implicit current event loop and we want to be compatible with them
        # (even when run on 3.10, but not, of course, on the future version of
        # python that removes the get/set_event_loop methods completely).
        # Deprecation warnings were introduced inconsistently:
        # asyncio.get_event_loop warns, but
        # asyncio.get_event_loop_policy().get_event_loop does not. Similarly,
        # none of the set_event_loop methods warn, although comments on
        # https://bugs.python.org/issue39529 indicate that they are also
        # intended for future removal.
        #
        # Therefore, we first attempt to access the event loop with the
        # new event loop. We do not have effective test coverage of the
        # new event loop case; this will have to be watched when/if
        # get_event_loop is actually removed.
        try:
            self.asyncio_loop = asyncio.get_event_loop_policy().get_event_loop()
        except Exception:
            self.should_close_asyncio_loop = True

        async def get_loop() -> IOLoop:
            return self.get_new_ioloop()
        self.io_loop = self.asyncio_loop.run_until_complete(get_loop())
        with warnings.catch_warnings():
            self.io_loop.make_current()

    def tearDown(self) -> None:
        # Native coroutines tend to produce warnings if they're not
        # allowed to run to completion. It's difficult to ensure that
        # this always happens in tests, so cancel any tasks that are
        # still pending by the time we get here.
        asyncio_loop = self.io_loop.asyncio_loop  # type: ignore
        if hasattr(asyncio, "all_tasks"):  # py37
            tasks = asyncio.all_tasks(asyncio_loop)  # type: ignore
        else:
            tasks = asyncio.Task.all_tasks(asyncio_loop)
        # Tasks that are done may still appear here and may contain
        tasks = [t for t in tasks if not t.done()]  # type: ignore
        for t in tasks:
            t.cancel()
        # Allow the tasks to run and finalize themselves (which means
        # raising a CancelledError inside the coroutine). This may
        # just transform the "task was destroyed but it is pending"
        # warning into a "uncaught CancelledError" warning, but
        # catching CancelledErrors in coroutines that may leak is
        # simpler than ensuring that no coroutines leak.
        if tasks:
            done, pending = self.io_loop.run_sync(lambda: asyncio.wait(tasks))
            # If any task failed with anything but a CancelledError, raise it.
            for f in done:
                try:
                    f.result()
                except asyncio.CancelledError:
                    pass
