"""Various utility functions and classes for URL management."""

import codecs
import contextlib
import os.path
import re
import sys
import traceback
import zlib
from gzip import GzipFile
from pathlib import Path
from urllib.parse import quote, unquote, urljoin, urlsplit
from urllib.request import Request, pathname2url, urlopen

from . import __version__
from .logger import LOGGER

# See https://stackoverflow.com/a/11687993/1162888
# Both are needed in Python 3 as the re module does not like to mix
# https://datatracker.ietf.org/doc/html/rfc3986#section-3.1
UNICODE_SCHEME_RE = re.compile("^([a-zA-Z][a-zA-Z0-9.+-]+):")
BYTES_SCHEME_RE = re.compile(b"^([a-zA-Z][a-zA-Z0-9.+-]+):")

# getfilesystemencoding() on Linux is sometimes stupid…
FILESYSTEM_ENCODING = sys.getfilesystemencoding()
try:  # pragma: no cover
    if codecs.lookup(FILESYSTEM_ENCODING).name == "ascii":
        FILESYSTEM_ENCODING = "utf-8"
except LookupError:  # pragma: no cover
    FILESYSTEM_ENCODING = "utf-8"

HTTP_HEADERS = {
    "User-Agent": f"WeasyPrint {__version__}",
    "Accept": "*/*",
    'Accept-En
