import contextlib


class NoVersionFound(VisualCException):
    pass

class BatchFileExecutionError(VisualCException):
    pass

class MSVCScriptNotFound(MSVCUserError):
    pass

class MSVCUseSettingsError(MSVCUserError):
    pass


# Dict to 'canonalize' the arch
_ARCH_TO_CANONICAL = {
    "amd64"     : "amd64",
    "emt64"     : "amd64",
    "i386"      : "x86",
    "i486"      : "x86",
    "i586"      : "x86",
    "i686"      : "x86",
    "ia64"      : "ia64",      # deprecated
    "itanium"   : "ia64",      # deprecated
    "x86"       : "x86",
    "x86_64"    : "amd64",
    "arm"       : "arm",
    "arm64"     : "arm64",
    "aarch64"   : "arm64",
}

# The msvc batch files report errors via stdout.  The following
# regular expression attempts to match known msvc error messages
# written to stdout.
re_script_output_error = re.compile(
    r"^(" + r"|".join([
        r"VSINSTALLDIR variable is not set",             # 2002-2003
        r"The specified configuration type is missing",  # 2005+
        r"Error in script usage",                        # 2005+
        r"ERROR\:",                                      # 2005+
        r"\!ERROR\!",                                    # 2015-2015
        r"\[ERROR\:",                                    # 2017+
        r"\[ERROR\]",                                    # 2017+
        r"Syntax\:",                                     # 2017+
    ]) + r")",
)

# Lists of compatible host/target combinations are derived from a set of defined
# constant data structures for each host architecture. The derived data structures
# implicitly handle the differences in full versions and express versions of visual
# studio. The host/target combination search lists are contructed in order of
# preference. The construction of the derived data structures is independent of actual
# visual studio installations.  The host/target configurations are used in both the
# initial msvc detection and when finding a valid batch file for a given host/target
# combination.
#
# HostTargetConfig description:
#
#     label:
#         Name used for identification.
#
#     host_all_hosts:
#         Defined list of compatible architectures for each host architecture.
#
#     host_all_targets:
#         Defined list of target architectures for each host architecture.
#
#     host_def_targets:
#         Defined list of default target architectures for each host architecture.
#
#     all_pairs:
#         Derived list of all host/target combination tuples.
#
#     host_target_map:
#         Derived list of all compatible host/target combinations for each
#         supported host/target combination.
#
#     host_all_targets_map:
#         Derived list of all compatible host/target combinations for each
#         supported host.  This is used in the initial check that cl.exe exists
#         in the requisite visual studio vc host/target directory for a given host.
#
#     host_def_targets_map:
#         Derived list of default compatible host/target combinations for each
#         supported host.  This is used for a given host when the user does not
#         request a target archicture.
#
#     target_host_map:
#         Derived list of compatible host/target combinations for each supported
#         target/host combination.  This is used for a given host and target when
#         the user requests a target architecture.

_HOST_TARGET_CONFIG_NT = namedtuple("HostTargetConfig", [
    # defined
    "label",                # name for debugging/output
    "host_all_hosts",       # host_all_hosts[host] -> host_list
    "host_all_targets",     # host_all_targets[host] -> target_list
    "host_def_targets",     # host_def_targets[host] -> target_list
    # derived
    "all_pairs",            # host_target_list
    "host_target_map",      # host_target_map[host][target] -> host_target_list
    "host_all_targets_map", # host_all_targets_map[host][target] -> host_target_list
    "host_def_targets_map", # host_def_targets_map[host][target] -> host_target_list
    "target_host_map",      # target_host_map[target][host] -> host_target_list
])

def _host_target_config_factory(*, label, host_all_hosts, host_all_targets, host_def_targets):

    def _make_host_target_map(all_hosts, all_targets):
        # host_target_map[host][target] -> host_target_list
        host_target_map = {}
        for host, host_list in all_hosts.items():
            host_target_map[host] = {}
            for host_platform in host_list:
                for target_platform in all_targets[host_platform]:
                    if target_platform not in host_target_map[host]:
                        host_target_map[host][target_platform] = []
                    host_target_map[host][target_platform].append((host_platform, target_platform))
        return host_target_map

    def _make_host_all_targets_map(all_hosts, host_target_map, all_targets):
        # host_all_target_map[host] -> host_target_list
        # special host key '_all_' contains all (host,target) combinations
        all = "_all_"
        host_all_targets_map = {}
        host_all_targets_map[all] = []
        for host, host_list in all_hosts.items():
            host_all_targets_map[host] = []
            for host_platform in host_list:
                # all_targets[host_platform]: all targets for compatible host
                for target in all_targets[host_platform]:
                    for host_target in host_target_map[host_platform][target]:
                        for host_key in (host, all):
                            if host_target not in host_all_targets_map[host_key]:
                                host_all_targets_map[host_key].append(host_target)
        return host_all_targets_map

    def _make_host_def_targets_map(all_hosts, host_target_map, def_targets):
        # host_def_targets_map[host] -> host_target_list
        host_def_targets_map = {}
        for host, host_list in all_hosts.items():
            host_def_targets_map[host] = []
            for host_platform in host_list:
                # def_targets[host]: default targets for true host
                for target in def_targets[host]:
                    for host_target in host_target_map[host_platform][target]:
                        if host_target not in host_def_targets_map[host]:
                            host_def_targets_map[host].append(host_target)
        return host_def_targets_map

    def _make_target_host_map(all_hosts, host_all_targets_map):
        # target_host_map[target][host] -> host_target_list
        target_host_map = {}
        for host_platform in all_hosts:
            for host_target in host_all_targets_map[host_platform]:
                _, target = host_target
                if target not in target_host_map:
                    target_host_map[target] = {}
                if host_platform not in target_host_map[target]:
                    target_host_map[target][host_platform] = []
                if host_target not in target_host_map[target][host_platform]:
                    target_host_map[target][host_platform].append(host_target)
        return target_host_map

    host_target_map = _make_host_target_map(host_all_hosts, host_all_targets)
    host_all_targets_map = _make_host_all_targets_map(host_all_hosts, host_target_map, host_all_targets)
    host_def_targets_map = _make_host_def_targets_map(host_all_hosts, host_target_map, host_def_targets)
    target_host_map = _make_target_host_map(host_all_hosts, host_all_targets_map)

    all_pairs = host_all_targets_map["_all_"]
    del host_all_targets_map["_all_"]

    host_target_cfg = _HOST_TARGET_CONFIG_NT(
        label = label,
        host_all_hosts = dict(host_all_hosts),
        host_all_targets = host_all_targets,
        host_def_targets = host_def_targets,
        all_pairs = all_pairs,
        host_target_map = host_target_map,
        host_all_targets_map = host_all_targets_map,
        host_def_targets_map = host_def_targets_map,
        target_host_map = target_host_map,
    )

    return host_target_cfg


# Given a (host, target) tuple, return a tuple containing the batch file to
# look for and a tuple of path components to find cl.exe. We can't rely on returning
# an arg to use for vcvarsall.bat, because that script will run even if given
# a host/target pair that isn't installed.
#
# Starting with 14.1 (VS2017), the batch files are located in directory
# <VSROOT>/VC/Auxiliary/Build.  The batch file name is the first value of the
# stored tuple.
#
# The build tools are organized by host and target subdirectories under each toolset
# version directory.  For example,  <VSROOT>/VC/Tools/MSVC/14.31.31103/bin/Hostx64/x64.
# The cl path fragment under the toolset version folder is the second value of
# the stored tuple.

_GE2017_HOST_TARGET_BATCHFILE_CLPATHCOMPS = {

    ("amd64", "amd64") : ("vcvars64.bat",          ("bin", "Hostx64", "x64")),
    ("amd64", "x86")   : ("vcvarsamd64_x86.bat",   ("bin", "Hostx64", "x86")),
    ("amd64", "arm")   : ("vcvarsamd64_arm.bat",   ("bin", "Hostx64", "arm")),
    ("amd64", "arm64") : ("vcvarsamd64_arm64.bat", ("bin", "Hostx64", "arm64")),

    ("x86",   "amd64") : ("vcvarsx86_amd64.bat",   ("bin", "Hostx86", "x64")),
    ("x86",   "x86")   : ("vcvars32.bat",          ("bin", "Hostx86", "x86")),
    ("x86",   "arm")   : ("vcvarsx86_arm.bat",     ("bin", "Hostx86", "arm")),
    ("x86",   "arm64") : ("vcvarsx86_arm64.bat",   ("bin", "Hostx86", "arm64")),

}

_GE2017_HOST_TARGET_CFG = _host_target_config_factory(

    label = "GE2017",

    host_all_hosts = OrderedDict([
        ("x86",   ["x86"]),
        ("arm64", ["amd64", "x86"]),
        ("arm",   ["x86"]),
    ]),

    host_all_targets = {
        "amd64": ["amd64", "x86", "arm64", "arm"],
        "x86":   ["x86", "amd64", "arm", "arm64"],
        "arm64": [],
        "arm":   [],
    },

    host_def_targets = {
        "amd64": ["amd64", "x86"],
        "x86":   ["x86"],
        "arm64": ["arm64", "arm"],
        "arm":   ["arm"],
    },

)


# 14.0 (VS2015) to 8.0 (VS2005)
# Given a (host, target) tuple, return a tuple containing the argument for
# the batch file and a tuple of the path components to find cl.exe.
#
# In 14.0 (VS2015) and earlier, the original x86 tools are in the tools
# bin directory (i.e., <VSROOT>/VC/bin).  Any other tools are in subdirectory
# named for the the host/target pair or a single name if the host==target.

_LE2015_HOST_TARGET_BATCHARG_CLPATHCOMPS = {

    ("amd64", "amd64") : ("amd64",     ("bin", "amd64")),
    ("amd64", "x86")   : ("amd64_x86", ("bin", "amd64_x86")),
    ("amd64", "arm")   : ("amd64_arm", ("bin", "amd64_arm")),

    ("x86",   "amd64") : ("x86_amd64", ("bin", "x86_amd64")),
    ("x86",   "x86")   : ("x86",       ("bin", )),
    ("x86",   "arm")   : ("x86_arm",   ("bin", "x86_arm")),
    ("x86",   "ia64")  : ("x86_ia64",  ("bin", "x86_ia64")),

    ("arm",   "arm")   : ("arm",       ("bin", "arm")),
    ("ia64",  "ia64")  : ("ia64",      ("bin", "ia64")),

}

_LE2015_HOST_TARGET_CFG = _host_target_config_factory(

    label = "LE2015",

    host_all_hosts = OrderedDict([
        ("amd64", ["amd64", "x86"]),
        ("x86",   ["x86"]),
        ("arm",   ["arm"]),
        ("ia64",  ["ia64"]),
    ]),

    host_all_targets = {
        "amd64": ["amd64", "x86", "arm"],
        "x86":   ["x86", "amd64", "arm", "ia64"],
        "arm":   ["arm"],
        "ia64":  ["ia64"],
    },

    host_def_targets = {
        "amd64": ["amd64", "x86"],
        "x86":   ["x86"],
        "arm":   ["arm"],
        "ia64":  ["ia64"],
    },

)



# For 7.1 (VS2003) and earlier, there are only x86 targets and the batch files
# take no arguments.

_LE2003_HOST_TARGET_CFG = _host_target_config_factory(

    label = "LE2003",

    host_all_hosts = OrderedDict([
        ("amd64", ["x86"]),
        ("x86",   ["x86"]),
    ]),

    host_all_targets = {
        "amd64": ["x86"],
        "x86":   ["x86"],
    },

    host_def_targets = {
        "amd64": ["x86"],
        "x86":   ["x86"],
    },

)


_CL_EXE_NAME = "cl.exe"

def get_msvc_version_numeric(msvc_version):
    """Get the raw version numbers from a MSVC_VERSION string, so it
    could be cast to float or other numeric values. For example, '14.0Exp'
    would get converted to '14.0'.

    Args:
    ----
        msvc_version: str
            string representing the version number, could contain non
            digit characters

    Returns:
    -------
        str: the value converted to a numeric only string

    """
    return "".join([x for x in msvc_version if x in string_digits + "."])

def get_host_platform(host_platform):

    host_platform = host_platform.lower()

    # Solaris returns i86pc for both 32 and 64 bit architectures
    if host_platform == "i86pc":
        host_platform = "amd64" if platform.architecture()[0] == "64bit" else "x86"

    try:
        host =_ARCH_TO_CANONICAL[host_platform]
    except KeyError:
        msg = "Unrecognized host architecture %s"
        raise MSVCUnsupportedHostArch(msg % repr(host_platform)) from None

    return host

_native_host_platform = None

def get_native_host_platform():
    global _native_host_platform

    if _native_host_platform is None:

        _native_host_platform = get_host_platform(platform.machine())

    return _native_host_platform

def get_host_target(env, msvc_version, all_host_targets=False):

    vernum = float(get_msvc_version_numeric(msvc_version))

    if vernum > 14:
        host_target_cfg = _GE2017_HOST_TARGET_CFG
    elif 14 >= vernum >= 8:
        # 14.0 (VS2015) to 8.0 (VS2005)
        host_target_cfg = _LE2015_HOST_TARGET_CFG
    else:
        host_target_cfg = _LE2003_HOST_TARGET_CFG

    host_arch = env.get("HOST_ARCH") if env else None
    debug("HOST_ARCH:%s", str(host_arch))

    if host_arch:
        host_platform = get_host_platform(host_arch)
    else:
        host_platform = get_native_host_platform()

    target_arch = env.get("TARGET_ARCH") if env else None
    debug("TARGET_ARCH:%s", str(target_arch))

    if target_arch:

        try:
            target_platform = _ARCH_TO_CANONICAL[target_arch.lower()]
        except KeyError:
            all_archs = str(list(_ARCH_TO_CANONICAL.keys()))
            raise MSVCUnsupportedTargetArch(
                f"Unrecognized target architecture {repr(target_arch)}\n\tValid architectures: {all_archs}",
            ) from None

        target_host_map = host_target_cfg.target_host_map

        try:
            host_target_list = target_host_map[target_platform][host_platform]
        except KeyError:
            host_target_list = []
            warn_msg = "unsupported host, target combination ({}, {}) for MSVC version {}".format(
                repr(host_platform), repr(target_platform), msvc_version,
            )
            SCons.Warnings.warn(SCons.Warnings.VisualCMissingWarning, warn_msg)
            debug(warn_msg)

    else:

        target_platform = None

        if all_host_targets:
            host_targets_map = host_target_cfg.host_all_targets_map
        else:
            host_targets_map = host_target_cfg.host_def_targets_map

        try:
            host_target_list = host_targets_map[host_platform]
        except KeyError:
            msg = "Unrecognized host architecture %s for version %s"
            raise MSVCUnsupportedHostArch(msg % (repr(host_platform), msvc_version)) from None

    return host_platform, target_platform, host_target_list

# If you update this, update SupportedVSList in Tool/MSCommon/vs.py, and the
# MSVC_VERSION documentation in Tool/msvc.xml.
_VCVER = [
    "14.3",
    "14.2",
    "14.1", "14.1Exp",
    "14.0", "14.0Exp",
    "12.0", "12.0Exp",
    "11.0", "11.0Exp",
    "10.0", "10.0Exp",
    "9.0", "9.0Exp",
    "8.0", "8.0Exp",
    "7.1",
    "7.0",
    "6.0"]

# if using vswhere, configure command line arguments to probe for installed VC editions
_VCVER_TO_VSWHERE_VER = {
    "14.3": [
        ["-version", "[17.0, 18.0)"],  # default: Enterprise, Professional, Community  (order unpredictable?)
        ["-version", "[17.0, 18.0)", "-products", "Microsoft.VisualStudio.Product.BuildTools"],  # BuildTools
    ],
    "14.2": [
        ["-version", "[16.0, 17.0)"],  # default: Enterprise, Professional, Community  (order unpredictable?)
        ["-version", "[16.0, 17.0)", "-products", "Microsoft.VisualStudio.Product.BuildTools"],  # BuildTools
    ],
    "14.1": [
        ["-version", "[15.0, 16.0)"],  # default: Enterprise, Professional, Community (order unpredictable?)
        ["-version", "[15.0, 16.0)", "-products", "Microsoft.VisualStudio.Product.BuildTools"],  # BuildTools
    ],
    "14.1Exp": [
        ["-version", "[15.0, 16.0)", "-products", "Microsoft.VisualStudio.Product.WDExpress"],  # Express
    ],
}

_VCVER_TO_PRODUCT_DIR = {
    "14.3": [
        (SCons.Util.HKEY_LOCAL_MACHINE, r"")],  # not set by this version
    "14.2": [
        (SCons.Util.HKEY_LOCAL_MACHINE, r"")],  # not set by this version
    "14.1": [
        (SCons.Util.HKEY_LOCAL_MACHINE, r"")],  # not set by this version
    "14.1Exp": [
        (SCons.Util.HKEY_LOCAL_MACHINE, r"")],  # not set by this version
    "14.0": [
        (SCons.Util.HKEY_LOCAL_MACHINE, r"Microsoft\VisualStudio\14.0\Setup\VC\ProductDir")],
    "14.0Exp": [
        (SCons.Util.HKEY_LOCAL_MACHINE, r"Microsoft\VCExpress\14.0\Setup\VC\ProductDir")],
    "12.0": [
        (SCons.Util.HKEY_LOCAL_MACHINE, r"Microsoft\VisualStudio\12.0\Setup\VC\ProductDir"),
    ],
    "12.0Exp": [
        (SCons.Util.HKEY_LOCAL_MACHINE, r"Microsoft\VCExpress\12.0\Setup\VC\ProductDir"),
    ],
    "11.0": [
        (SCons.Util.HKEY_LOCAL_MACHINE, r"Microsoft\VisualStudio\11.0\Setup\VC\ProductDir"),
    ],
    "11.0Exp": [
        (SCons.Util.HKEY_LOCAL_MACHINE, r"Microsoft\VCExpress\11.0\Setup\VC\ProductDir"),
    ],
    "10.0": [
        (SCons.Util.HKEY_LOCAL_MACHINE, r"Microsoft\VisualStudio\10.0\Setup\VC\ProductDir"),
    ],
    "10.0Exp": [
        (SCons.Util.HKEY_LOCAL_MACHINE, r"Microsoft\VCExpress\10.0\Setup\VC\ProductDir"),
    ],
    "9.0": [
        (SCons.Util.HKEY_CURRENT_USER, r"Microsoft\DevDiv\VCForPython\9.0\installdir"),
        (SCons.Util.HKEY_LOCAL_MACHINE, r"Microsoft\VisualStudio\9.0\Setup\VC\ProductDir"),
    ],
    "9.0Exp": [
        (SCons.Util.HKEY_LOCAL_MACHINE, r"Micosoft\VCExpress\9.0\Setup\VC\ProductDir"),
    ],
    "8.0": [
        (SCons.Util.HKEY_LOCAL_MACHINE, r"Microsoft\VisualStudio\8.0\Setup\VC\ProductDir"),
    ],
    "8.0Exp": [
        (SCons.Util.HKEY_LOCAL_MACHINE, r"Microsoft\VCExpress\8.0\Setup\VC\ProductDir"),
    ],
    "7.1": [
        (SCons.Util.HKEY_LOCAL_MACHINE, r"Microsoft\VisualStudio\7.1\Setup\VC\ProductDir"),
    ],
    "7.0": [
        (SCons.Util.HKEY_LOCAL_MACHINE, r"Microsoft\VisualStudio\7.0\Setup\VC\ProductDir"),
    ],
    "6.0": [
        (SCons.Util.HKEY_LOCAL_MACHINE, r"Microsoft\VisualStudio\6.0\Setup\Microsoft Visual C++\ProductDir"),
    ],
}


def msvc_version_to_maj_min(msvc_version):
    msvc_version_numeric = get_msvc_version_numeric(msvc_version)

    t = msvc_version_numeric.split(".")
    if not len(t) == 2:
        raise ValueError(f"Unrecognized version {msvc_version} ({msvc_version_numeric})")
    try:
        maj = int(t[0])
        min = int(t[1])
        return maj, min
    except ValueError:
        raise ValueError(f"Unrecognized version {msvc_version} ({msvc_version_numeric})") from None


VSWHERE_PATHS = [os.path.join(p,"vswhere.exe") for p in [
    os.path.expandvars(r"%ProgramFiles(x86)%\Microsoft Visual Studio\Installer"),
    os.path.expandvars(r"%ProgramFiles%\Microsoft Visual Studio\Installer"),
    os.path.expandvars(r"%ChocolateyInstall%\bin"),
]]

def msvc_find_vswhere():
    """Find the location of vswhere."""
    # For bug 3333: support default location of vswhere for both
    # 64 and 32 bit windows installs.
    # For bug 3542: also accommodate not being on C: drive.
    # NB: this gets called from testsuite on non-Windows platforms.
    # Whether that makes sense or not, don't break it for those.
    vswhere_path = None
    for pf in VSWHERE_PATHS:
        if os.path.exists(pf):
            vswhere_path = pf
            break

    return vswhere_path

def find_vc_pdir_vswhere(msvc_version, env=None):
    """Find the MSVC product directory using the vswhere program.

    Args:
    ----
        msvc_version: MSVC version to search for
        env: optional to look up VSWHERE variable

    Returns:
    -------
        MSVC install dir or None

    Raises:
    ------
        UnsupportedVersion: if the version is not known by this file

    """
    try:
        vswhere_version = _VCVER_TO_VSWHERE_VER[msvc_version]
    except KeyError:
        debug("Unknown version of MSVC: %s", msvc_version)
        raise UnsupportedVersion("Unknown version %s" % msvc_version) from None

    if env is None or not env.get("VSWHERE"):
        vswhere_path = msvc_find_vswhere()
    else:
        vswhere_path = env.subst("$VSWHERE")

    if vswhere_path is None:
        return None

    debug("VSWHERE: %s", vswhere_path)
    for vswhere_version_args in vswhere_version:

        vswhere_cmd = [vswhere_path, *vswhere_version_args] + ["-property", "installationPath"]

        debug("running: %s", vswhere_cmd)

        # TODO: Python 3.7
        cp = subprocess.run(vswhere_cmd, stdout=PIPE, stderr=PIPE, check=True)

        if cp.stdout:
            # vswhere could return multiple lines, e.g. if Build Tools
            # and {Community,Professional,Enterprise} are both installed.
            # We could define a way to pick the one we prefer, but since
            # this data is currently only used to make a check for existence,
            # returning the first hit should be good enough.
            lines = cp.stdout.decode("mbcs").splitlines()
            return os.path.join(lines[0], "VC")
        else:
            # We found vswhere, but no install info available for this version
            pass

    return None


def find_vc_pdir(env, msvc_version):
    """Find the MSVC product directory for the given version.

    Tries to look up the path using a registry key from the table
    _VCVER_TO_PRODUCT_DIR; if there is no key, calls find_vc_pdir_wshere
    for help instead.

    Args:
    ----
        msvc_version: str
            msvc version (major.minor, e.g. 10.0)

    Returns:
    -------
        str: Path found in registry, or None

    Raises:
    ------
        UnsupportedVersion: if the version is not known by this file.
        MissingConfiguration: found version but the directory is missing.

        Both exceptions inherit from VisualCException.

    """
    root = "Software\\"
    try:
        hkeys = _VCVER_TO_PRODUCT_DIR[msvc_version]
    except KeyError:
        debug("Unknown version of MSVC: %s", msvc_version)
        raise UnsupportedVersion("Unknown version %s" % msvc_version) from None

    for hkroot, key in hkeys:
        try:
            comps = None
            if not key:
                comps = find_vc_pdir_vswhere(msvc_version, env)
                if not comps:
                    debug("no VC found for version %s", repr(msvc_version))
                    raise OSError
                debug("VC found: %s", repr(msvc_version))
                return comps
            else:
                if common.is_win64():
                    with contextlib.suppress(OSError):
                        # ordinarily at win64, try Wow6432Node first.
                        comps = common.read_reg(root + "Wow6432Node\\" + key, hkroot)

                if not comps:
                    # not Win64, or Microsoft Visual Studio for Python 2.7
                    comps = common.read_reg(root + key, hkroot)
        except OSError:
            debug("no VC registry key %s", repr(key))
        else:
            if msvc_version == "9.0" and key.lower().endswith("\\vcforpython\\9.0\\installdir"):
                # Visual C++ for Python registry key is installdir (root) not productdir (vc)
                comps = os.path.join(comps, "VC")
            debug("found VC in registry: %s", comps)
            if os.path.exists(comps):
                return comps
            else:
                debug("reg says dir is %s, but it does not exist. (ignoring)", comps)
                raise MissingConfiguration(f"registry dir {comps} not found on the filesystem")
    return None

def find_batch_file(env, msvc_version, host_arch, target_arch):
    """Find the location of the batch script which should set up the compiler
    for any TARGET_ARCH whose compilers were installed by Visual Studio/VCExpress.

    In newer (2017+) compilers, make use of the fact there are vcvars
    scripts named with a host_target pair that calls vcvarsall.bat properly,
    so use that and return an empty argument.
    """
    pdir = find_vc_pdir(env, msvc_version)
    if pdir is None:
        raise NoVersionFound("No version of Visual Studio found")
    debug("looking in %s", pdir)

    # filter out e.g. "Exp" from the version name
    msvc_ver_numeric = get_msvc_version_numeric(msvc_version)
    vernum = float(msvc_ver_numeric)

    arg = ""

    if vernum > 14:
        batfiledir = os.path.join(pdir, "Auxiliary", "Build")
        batfile, _ = _GE2017_HOST_TARGET_BATCHFILE_CLPATHCOMPS[(host_arch, target_arch)]
        batfilename = os.path.join(batfiledir, batfile)
    elif 14 >= vernum >= 8:
        # 14.0 (VS2015) to 8.0 (VS2005)
        arg, _ = _LE2015_HOST_TARGET_BATCHARG_CLPATHCOMPS[(host_arch, target_arch)]
        batfilename = os.path.join(pdir, "vcvarsall.bat")
        if msvc_version == "9.0" and not os.path.exists(batfilename):
            # Visual C++ for Python batch file is in installdir (root) not productdir (vc)
            batfilename = os.path.normpath(os.path.join(pdir, os.pardir, "vcvarsall.bat"))
    else:
        pdir = os.path.join(pdir, "Bin")
        batfilename = os.path.join(pdir, "vcvars32.bat")

    if not os.path.exists(batfilename):
        debug("Not found: %s", batfilename)
        batfilename = None

    installed_sdks = get_installed_sdks()
    for _sdk in installed_sdks:
        sdk_bat_file = _sdk.get_sdk_vc_script(host_arch, target_arch)
        if not sdk_bat_file:
            debug("batch file not found:%s", _sdk)
