
http_method_probably_has_body = dict.fromkeys(
    ("GET", "HEAD", "DELETE", "TRACE"), False)
http_method_probably_has_body.update(
    dict.fromkeys(("POST", "PUT"), True))

_LATIN_ENCODINGS = (
    "ascii", "latin-1", "latin", "latin_1", "l1", "latin1",
    "iso-8859-1", "iso8859_1", "iso_8859_1", "iso8859", "8859",
    )

class BaseRequest:
    ## The limit after which request bodies should be stored on disk
    ## if they are read in (under this, and the request body is stored
    ## in memory):
    request_body_tempfile_limit = 10*1024

    _charset = None

    def __init__(self, environ, charset=None, unicode_errors=None,
                 decode_param_names=None, **kw) -> None:

        if type(environ) is not dict:
            raise TypeError(
                f"WSGI environ must be a dict; you passed {environ!r}")
        if unicode_errors is not None:
            warnings.warn(
                "You unicode_errors={!r} to the Request constructor.  Passing a "
                "``unicode_errors`` value to the Request is no longer "
                "supported in WebOb 1.2+.  This value has been ignored ".format(
                    unicode_errors),
                DeprecationWarning,
                )
        if decode_param_names is not None:
            warnings.warn(
                "You passed decode_param_names={!r} to the Request constructor. "
                "Passing a ``decode_param_names`` value to the Request "
                "is no longer supported in WebOb 1.2+.  This value has "
                "been ignored ".format(decode_param_names),
                DeprecationWarning,
                )
        if not _is_utf8(charset):
            raise DeprecationWarning(
                "You passed charset=%r to the Request constructor. As of "
                "WebOb 1.2, if your application needs a non-UTF-8 request "
                "charset, please construct the request without a charset or "
                "with a charset of 'None',  then use ``req = "
                "req.decode(charset)``" % charset,

            )
        d = self.__dict__
        d["environ"] = environ
        if kw:
            cls = self.__class__
            if "method" in kw:
                # set method first, because .body setters
                # depend on it for checks
                self.method = kw.pop("method")
            for name, value in kw.items():
                if not hasattr(cls, name):
                    raise TypeError(
                        f"Unexpected keyword: {name}={value!r}")
                setattr(self, name, value)

    if PY3: # pragma: no cover
        def encget(self, key, default=NoDefault, encattr=None):
            val = self.environ.get(key, default)
            if val is NoDefault:
                raise KeyError(key)
            if val is default:
                return default
            if not encattr:
                return val
            encoding = getattr(self, encattr)
            if encoding in _LATIN_ENCODINGS: # shortcut
                return val
            return bytes_(val, "latin-1").decode(encoding)
    else:
        def encget(self, key, default=NoDefault, encattr=None):
            val = self.environ.get(key, default)
            if val is NoDefault:
                raise KeyError(key)
            if val is default:
                return default
            if encattr is None:
                return val
            encoding = getattr(self, encattr)
            return val.decode(encoding)

    def encset(self, key, val, encattr=None):
        encoding = getattr(self, encattr) if encattr else "ascii"
        if PY3: # pragma: no cover
            self.environ[key] = bytes_(val, encoding).decode("latin-1")
        else:
            self.environ[key] = bytes_(val, encoding)

    @property
    def charset(self):
        if self._charset is None:
            charset = detect_charset(self._content_type_raw)
            if _is_utf8(charset):
                charset = "UTF-8"
            self._charset = charset
        return self._charset

    @charset.setter
    def charset(self, charset):
        if _is_utf8(charset):
            charset = "UTF-8"
        if charset != self.charset:
            raise DeprecationWarning("Use req = req.decode(%r)" % charset)

    def decode(self, charset=None, errors="strict"):
        charset = charset or self.charset
        if charset == "UTF-8":
            return self
        # cookies and path are always utf-8
        t = Transcoder(charset, errors)

        new_content_type = CHARSET_RE.sub('; charset="UTF-8"',
                                          self._content_type_raw)
        content_type = self.content_type
        r = self.__class__(
            self.environ.copy(),
            query_string=t.transcode_query(self.query_string),
            content_type=new_content_type,
        )

        if content_type == "application/x-www-form-urlencoded":
            r.body = bytes_(t.transcode_query(native_(r.body)))
            return r
        elif content_type != "multipart/form-data":
            return r

        fs_environ = self.environ.copy()
        fs_environ.setdefault("CONTENT_LENGTH", "0")
        fs_environ["QUERY_STRING"] = ""
        if PY3: # pragma: no cover
            fs = cgi.FieldStorage(fp=self.body_file,
                                  environ=fs_environ,
                                  keep_blank_values=True,
                                  encoding=charset,
                                  errors=errors)
        else:
            fs = cgi.FieldStorage(fp=self.body_file,
                                  environ=fs_environ,
                                  keep_blank_values=True)


        fout = t.transcode_fs(fs, r._content_type_raw)

        # this order is important, because setting body_file
        # resets content_length
        r.body_file = fout
        r.content_length = fout.tell()
        fout.seek(0)
        return r


    # this is necessary for correct warnings depth for both
    # BaseRequest and Request (due to AdhocAttrMixin.__setattr__)
    _setattr_stacklevel = 2

    def _body_file__get(self):
        """Input stream of the request (wsgi.input).
        Setting this property resets the content_length and seekable flag
        (unlike setting req.body_file_raw).
        """
        if not self.is_body_readable:
            return io.BytesIO()
        r = self.body_file_raw
        clen = self.content_length
        if not self.is_body_seekable and clen is not None:
            # we need to wrap input in LimitedLengthFile
            # but we have to cache the instance as well
            # otherwise this would stop working
            # (.remaining counter would reset between calls):
            env = self.environ
            wrapped, raw = env.get("webob._body_file", (0,0))
            if raw is not r:
                wrapped = LimitedLengthFile(r, clen)
                wrapped = io.BufferedReader(wrapped)
                env["webob._body_file"] = wrapped, r
            r = wrapped
        return r

    def _body_file__set(self, value):
        if isinstance(value, bytes):
            warn_deprecation(
                "Please use req.body = b'bytes' or req.body_file = fileobj",
                "1.2",
                self._setattr_stacklevel,
            )
        self.content_length = None
        self.body_file_raw = value
        self.is_body_seekable = False
        self.is_body_readable = True
    def _body_file__del(self):
        self.body = b""
    body_file = property(_body_file__get,
                         _body_file__set,
                         _body_file__del,
                         doc=_body_file__get.__doc__)
    body_file_raw = environ_getter("wsgi.input")
    @property
    def body_file_seekable(self):
        """Get the body of the request (wsgi.input) as a seekable file-like
        object. Middleware and routing applications should use this
        attribute over .body_file.

        If you access this value, CONTENT_LENGTH will also be updated.
        """
        if not self.is_body_seekable:
            self.make_body_seekable()
        return self.body_file_raw

    url_encoding = environ_getter("webob.url_encoding", "UTF-8")
    scheme = environ_getter("wsgi.url_scheme")
    method = environ_getter("REQUEST_METHOD", "GET")
    http_version = environ_getter("SERVER_PROTOCOL")
    content_length = converter(
        environ_getter("CONTENT_LENGTH", None, "14.13"),
        parse_int_safe, serialize_int, "int")
    remote_user = environ_getter("REMOTE_USER", None)
    remote_addr = environ_getter("REMOTE_ADDR", None)
    query_string = environ_getter("QUERY_STRING", "")
    server_name = environ_getter("SERVER_NAME")
    server_port = converter(
        environ_getter("SERVER_PORT"),
        parse_int, serialize_int, "int")

    script_name = environ_decoder("SCRIPT_NAME", "", encattr="url_encoding")
    path_info = environ_decoder("PATH_INFO", encattr="url_encoding")

    # bw compat
    uscript_name = script_name
    upath_info = path_info

    _content_type_raw = environ_getter("CONTENT_TYPE", "")

    def _content_type__get(self):
        """Return the content type, but leaving off any parameters (like
        charset, but also things like the type in ``application/atom+xml;
        type=entry``).

        If you set this property, you can include parameters, or if
        you don't include any parameters in the value then existing
        parameters will be preserved.
        """
        return self._content_type_raw.split(";", 1)[0]
    def _content_type__set(self, value=None):
        if value is not None:
            value = str(value)
            if ";" not in value:
                content_type = self._content_type_raw
                if ";" in content_type:
                    value += ";" + content_type.split(";", 1)[1]
        self._content_type_raw = value

    content_type = property(_content_type__get,
                            _content_type__set,
                            _content_type__set,
                            _content_type__get.__doc__)

    _headers = None

    def _headers__get(self):
        """All the request headers as a case-insensitive dictionary-like
        object.
        """
        if self._headers is None:
            self._headers = EnvironHeaders(self.environ)
        return self._headers

    def _headers__set(self, value):
        self.headers.clear()
        self.headers.update(value)

    headers = property(_headers__get, _headers__set, doc=_headers__get.__doc__)

    @property
    def client_addr(self):
        """The effective client IP address as a string.  If the
        ``HTTP_X_FORWARDED_FOR`` header exists in the WSGI environ, this
        attribute returns the client IP address present in that header
        (e.g. if the header value is ``192.168.1.1, 192.168.1.2``, the value
        will be ``192.168.1.1``). If no ``HTTP_X_FORWARDED_FOR`` header is
        present in the environ at all, this attribute will return the value
        of the ``REMOTE_ADDR`` header.  If the ``REMOTE_ADDR`` header is
        unset, this attribute will return the value ``None``.

        .. warning::

           It is possible for user agents to put someone else's IP or just
           any string in ``HTTP_X_FORWARDED_FOR`` as it is a normal HTTP
           header. Forward proxies can also provide incorrect values (private
           IP addresses etc).  You cannot "blindly" trust the result of this
           method to provide you with valid data unless you're certain that
           ``HTTP_X_FORWARDED_FOR`` has the correct values.  The WSGI server
           must be behind a trusted proxy for this to be true.
        """
        e = self.environ
        xff = e.get("HTTP_X_FORWARDED_FOR")
        addr = xff.split(",")[0].strip() if xff is not None else e.get("REMOTE_ADDR")
        return addr

    @property
    def host_port(self):
        """The effective server port number as a string.  If the ``HTTP_HOST``
        header exists in the WSGI environ, this attribute returns the port
        number present in that header. If the ``HTTP_HOST`` header exists but
        contains no explicit port number: if the WSGI url scheme is "https" ,
        this attribute returns "443", if the WSGI url scheme is "http", this
        attribute returns "80" .  If no ``HTTP_HOST`` header is present in
        the environ at all, this attribute will return the value of the
        ``SERVER_PORT`` header (which is guaranteed to be present).
        """
        e = self.environ
        host = e.get("HTTP_HOST")
        if host is not None:
            if ":" in host:
                host, port = host.split(":", 1)
            else:
                url_scheme = e["wsgi.url_scheme"]
                port = "443" if url_scheme == "https" else "80"
        else:
            port = e["SERVER_PORT"]
        return port

    @property
    def host_url(self):
        """The URL through the host (no path)."""
        e = self.environ
        scheme = e.get("wsgi.url_scheme")
        url = scheme + "://"
        host = e.get("HTTP_HOST")
        if host is not None:
            if ":" in host:
                host, port = host.split(":", 1)
            else:
                port = None
        else:
            host = e.get("SERVER_NAME")
            port = e.get("SERVER_PORT")
        if scheme == "https":
            if port == "443":
                port = None
        elif scheme == "http" and port == "80":
            port = None
        url += host
        if port:
            url += ":%s" % port
        return url

    @property
    def application_url(self):
        """The URL including SCRIPT_NAME (no PATH_INFO or query string)."""
        bscript_name = bytes_(self.script_name, self.url_encoding)
        return self.host_url + url_quote(bscript_name, PATH_SAFE)

    @property
    def path_url(self):
        """The URL including SCRIPT_NAME and PATH_INFO, but not QUERY_STRING."""
        bpath_info = bytes_(self.path_info, self.url_encoding)
        return self.application_url + url_quote(bpath_info, PATH_SAFE)

    @property
    def path(self):
        """The path of the request, without host or query string."""
        bscript = bytes_(self.script_name, self.url_encoding)
        bpath = bytes_(self.path_info, self.url_encoding)
        return url_quote(bscript, PATH_SAFE) + url_quote(bpath, PATH_SAFE)

    @property
    def path_qs(self):
        """The path of the request, without host but with query string."""
        path = self.path
        qs = self.environ.get("QUERY_STRING")
        if qs:
            path += "?" + qs
        return path

    @property
    def url(self):
        """The full request URL, including QUERY_STRING."""
        url = self.path_url
        qs = self.environ.get("QUERY_STRING")
        if qs:
            url += "?" + qs
        return url

    def relative_url(self, other_url, to_application=False):
        """Resolve other_url relative to the request URL.

        If ``to_application`` is True, then resolve it relative to the
        URL with only SCRIPT_NAME
        """
        if to_application:
            url = self.application_url
            if not url.endswith("/"):
                url += "/"
        else:
            url = self.path_url
        return urlparse.urljoin(url, other_url)

    def path_info_pop(self, pattern=None):
        """'Pops' off the next segment of PATH_INFO, pushing it onto
        SCRIPT_NAME, and returning the popped segment.  Returns None if
        there is nothing left on PATH_INFO.

        Does not return ``''`` when there's an empty segment (like
        ``/path//path``); these segments are just ignored.

        Optional ``pattern`` argument is a regexp to match the return value
        before returning. If there is no match, no changes are made to the
        request and None is returned.
        """
        path = self.path_info
        if not path:
            return None
        slashes = ""
        while path.startswith("/"):
            slashes += "/"
            path = path[1:]
        idx = path.find("/")
        if idx == -1:
            idx = len(path)
        r = path[:idx]
        if pattern is None or re.match(pattern, r):
            self.script_name += slashes + r
            self.path_info = path[idx:]
            return r
        return None

    def path_info_peek(self):
        """Returns the next segment on PATH_INFO, or None if there is no
        next segment.  Doesn't modify the environment.
        """
        path = self.path_info
        if not path:
            return None
        path = path.lstrip("/")
        return path.split("/", 1)[0]

    def _urlvars__get(self):
        """Return any *named* variables matched in the URL.

        Takes values from ``environ['wsgiorg.routing_args']``.
        Systems like ``routes`` set this value.
        """
        if "paste.urlvars" in self.environ:
            return self.environ["paste.urlvars"]
        elif "wsgiorg.routing_args" in self.environ:
            return self.environ["wsgiorg.routing_args"][1]
        else:
            result = {}
            self.environ["wsgiorg.routing_args"] = ((), result)
            return result
