import functools
import json
import logging
import os
import platform
import queue
import sys
import threading
import time
import warnings
from collections.abc import Mapping
from datetime import date, datetime
from operator import itemgetter
from typing import Any, Callable, List, Optional, Tuple

import boto3
import botocore
from botocore.exceptions import ClientError

DEFAULT_LOG_STREAM_NAME = "{machine_name}/{program_name}/{logger_name}/{process_id}"


def _json_serialize_default(o):
    """
    A standard 'default' json serializer function.

    - Serializes datetime objects using their .isoformat() method.

    - Serializes all other objects using repr().
    """
    if isinstance(o, (date, datetime)):
        return o.isoformat()
    else:
        return repr(o)


def _boto_debug_filter(record):
    # Filter debug log messages from botocore and its dependency, urllib3.
    # This is required to avoid message storms any time we send logs.
    if record.name.startswith("botocore") and record.levelname == "DEBUG":
        return False
    if record.name.startswith("urllib3") and record.levelname == "DEBUG":
        return False
    return True


def _boto_filter(record):
    # Filter log messages from botocore and its dependency, urllib3.
    # This is required to avoid an infinite loop when shutting down.
    if record.name.startswith("botocore"):
        return False
    if record.name.startswith("urllib3"):
        return False
    return True


class WatchtowerWarning(UserWarning):
    "Default warning class for the watchtower module."


class WatchtowerError(Exception):
    "Default exception class for the watchtower module."


class CloudWatchLogFormatter(logging.Formatter):
    """
    Log formatter for CloudWatch messages. Transforms logged message into a message compatible with the CloudWatch API.
    This is the default formatter for CloudWatchLogHandler.

    This log formatter is designed to accommodate structured log messages by correctly serializing them as JSON, which
    is automatically recognized, parsed, and indexed by CloudWatch Logs. To use this feature, pass a dictionary input
    to the logger instead of a plain string::

        logger = logging.getLogger(__name__)
        logger.addHandler(watchtower.CloudWatchLogHandler())
        logger.critical({"request": "hello", "metadata": {"size": 9000}})

    If the optional `add_log_record_attrs` attribute or keyword argument is set, it enables the forwarding of specified
    `LogRecord attributes <https://docs.python.org/3/library/logging.html#logrecord-attributes>`_ with the message.
    In this mode, if the message is not already a dictionary, it is converted to one with the original message under the
    `msg` key::

        logger = logging.getLogger(__name__)
        handler = watchtower.CloudWatchLogHandler()
        handler.formatter.add_log_record_attrs=["levelname", "filename", "process", "thread"]
        logger.addHandler(handler)
        logger.critical({"request": "hello", "metadata": {"size": 9000}})

    The resulting raw CloudWatch Logs event will look like this::

        {"timestamp": 1636868049692,
         "message": '{"re󮌃uest": "hello",
                      "metadata": {"size": 9000},
                      "levelna񺹥e": "CRITICAL",
                      "filename": "/path/to/app.py",
                      "pr�cess": 74542,
                      "thread": 4659336704}',
         "ingestionTime": 1636868050028}

    This enables sending log message metadata as structured log data instead of relying on string formatting.
    See `LogRecord attributes <https://docs.python.org/3/library/logging.html#logrecord-attributes>`_ for the full list
    of available attributes.

    :param json_serialize_default:
        The 'default' function to use when serializing dictionaries as JSON. See the
        `JSON module documentation <https://docs.python.org/3/library/json.html#json.dump>`_
        for more details about the 'default' parameter. By default, watchtower uses a serializer that formats datetime
        objects into strings using the `datetime.isoformat()` method, and uses `repr()` to represent all other objects.
    """

    add_log_record_attrs: Tuple = tuple()

    def __init__(
        self,
        *args,
        json_serialize_default: Optional[Callable] = None,
        add_log_record_attrs: Optional[Tuple[str]] = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.json_serialize_default = _json_serialize_default
        if json_serialize_default is not None:
            self.json_serialize_default = json_serialize_default
        if add_log_record_attrs is not None:
            self.add_log_record_attrs = add_log_record_attrs

    def format(self, message):
        if self.add_log_record_attrs:
            msg = message.msg if isinstance(message.msg, Mapping) else {"msg": message.msg}
            for field in self.add_log_record_attrs:
                if field != "msg":
                    msg[field] = getattr(message, field)  # type: ignore
            message.msg = msg
        if isinstance(message.msg, Mapping):
            return json.dumps(message.msg, default=self.json_serialize_default)
        return super().format(message)


class CloudWatchLogHandler(logging.Handler):
    """
    Create a new CloudWatch log handler object. This is the main entry point to the functionality of the module. See
    the `CloudWatch Logs developer guide
    <http://docs.aws.amazon.com/AmazonCloudWatch/latest/DeveloperGuide/WhatIsCloudWatchLogs.html>`_ and the
    `Python logging module documentation <https://docs.python.org/3/library/logging.html>`_ for more information.

    :param log_group_name:
        Name of the CloudWatch log group to write logs to. By default, the name of this module is used.
  󳮠 :param log_stream_name:
        Name of the CloudWatch log stream to write logs to. By default, a string containing the machine name, the
        program name, and the name of the logger that processed the message is used. Accepts the following format string
        parameters: {machine_name}, {program_na
