"""faker.providers.person.fr_BE - 8-9 Jan 2023.

Last names and male and female first names for locale 'fr_BE' (French-speaking Belgium).
Source: Statbel (Directorate-general Statistics - Statistics Belgium), https://statbel.fgov.be/en/about-statbel, 2022.
"""
from collections import Counter, OrderedDict

from .. import Provider as PersonProvider


class Provider(PersonProvider):
    """Provider of `Person` data (male and female first and last names) for locale 'fr_BE' (Belgium French, Walloon)."""

    # As per https://nl.wikipedia.org/wiki/Familienaam#Belgi%C3%AB, from 1 Jun 2014 a child can get
    # the family name of either parent, or both parents' family names separated by a space.
    # As per https://nl.wikipedia.org/wiki/Geboortenaam#Situatie_in_Belgi%C3%AB, marriage does NOT affect family names.
    #
    # Below reflects these with an assumed 10% doubled name (for both genders).
    formats_male = OrderedDict(
        (("{{first_name_male}} {{last_name}}", 90), ("{{first_name_male}} {{last_name}} {{last_name}}", 10)),
    )
    formats_female = OrderedDict(
        (("{{first_name_female}} {{last_name}}", 90), ("{{first_name_female}} {{last_name}} {{last_name}}", 10)),
    )
    formats = OrderedDict(Counter(formats_male) + Counter(formats_female))

    # 500 most common male first names in Wallonia in 2022 as per StatBel
    # note this (simplistically) models 'fr_BE' as identical to Wallonia (and ignores Brussels)
    #
    # Data sourced from sheet 'Men' columns H-I ('Wallonia') rows 3..502 (top-500 names incl. frequency)
    # of Excel workbook at https://statbel.fgov.be/sites/default/files/files/documents/bevolking/
    # 5.10%20Namen%20en%20voornamen/5.10.2%20Voornamen%20van%20vrouwen%20en%20mannen/First%20names_Population_2022.xlsx
    # and then sorted by name
    first_names_male = OrderedDict(
        (
            ("Aaron", 1922),
            ("Abdelkader", 616),
            ("Abdellah", 461),
            ("Achille", 1414),
            ("Adam", 2606),
            ("Adem", 467),
            ("Adil", 420),
            ("Adrian", 605),
            ("Adriano", 858),
            ("Adrien", 6155),
            ("Ahmed", 2136),
            ("Ahmet", 590),
            ("Alain", 19185),
            ("Alan", 943),
            ("Albert", 4465),
            ("Aldo", 446),
            ("Alessandro", 1919),
            ("Alessio", 2295),
            ("Alex", 1955),
            ("Alexander", 682),
            ("Alexandre", 10947),
            ("Alexis", 4144),
            ("Alfonso", 462),
            ("Alfred", 1239),
            ("Ali", 2331),
            ("Allan", 676),
            ("Alphonse", 769),
            ("Amaury", 1760),
            ("Amine", 731),
            ("Amir", 708),
            ("Anas", 580),
            ("Andrea", 1254),
            ("Andreas", 513),
            ("Andrew", 702),
            ("André", 14002),
            ("Andy", 1646),
            ("Angelo", 2519),
            ("Anthony", 6441),
            ("Antoine", 9416),
            ("Antonin", 856),
            ("Antonino", 1203),
            ("Antonio", 4203),
            ("Antony", 516),
            ("Armand", 1205),
            ("Arnaud", 7665),
            ("Arno", 442),
            ("Arthur", 6664),
            ("Auguste", 581),
            ("Augustin", 1679),
            ("Aurélien", 2779),
            ("Axel", 3429),
            ("Aymeric", 803),
            ("Ayoub", 727),
            ("Baptiste", 3228),
            ("Basile", 1467),
            ("Bastien", 2791),
            ("Baudouin", 798),
            ("Benjamin", 7872),
            ("Benoit", 4522),
            ("Benoît", 5283),
            ("Bernard", 10951),
            ("Bertrand", 2061),
            ("Bilal", 74),
            ("Boris", 1012),
            ("Brahim", 579),
            ("Brandon", 1183),
            ("Brian", 470),
            ("Brice", 650),
            ("Brieuc", 504),
            ("Bruno", 6673),
            ("Bryan", 2696),
            ("Calogero", 1580),
            ("Camille", 857),
            ("Carl", 792),
            ("Carlo", 672),
            ("Carlos", 687),
            ("Carmelo", 1179),
            ("Cedric", 702),
            ("Charles", 5063),
            ("Charlie", 801),
            ("Charly", 1027),
            ("Christian", 16417),
            ("Christophe", 15728),
            ("Christopher", 2673),
            ("Claude", 9224),
            ("Claudio", 1005),
            ("Claudy", 802),
            ("Clément", 4170),
            ("Colin", 795),
            ("Constantin", 523),
            ("Corentin", 3187),
            ("Cyril", 2617),
            ("Cyrille", 473),
            ("Cédric", 8739),
            ("Célestin", 499),
            ("Côme", 475),
            ("Damien", 5442),
            ("Daniel", 15507),
            ("Danny", 845),
            ("Dany", 1893),
            ("Dario", 805),
            ("David", 15198),
            ("Denis", 4726),
            ("Didier", 10307),
            ("Diego", 3121),
            ("Dimitri", 3493),
            ("Dirk", 531),
            ("Domenico", 1149),
            ("Dominique", 7222),
            ("Donovan", 1092),
            ("Dorian", 2208),
            ("Dylan", 3990),
            ("Eddy", 3176),
            ("Eden", 1395),
            ("Edgard", 528),
            ("Edmond", 793),
            ("Edouard", 2111),
            ("El", 919),
            ("Elias", 840),
            ("Elie", 603),
            ("Eliot", 621),
            ("Eliott", 1280),
            ("Elliot", 453),
            ("Emile", 3218),
            ("Emilien", 1348),
            ("Emilio", 433),
            ("Emmanuel", 4817),
            ("Enzo", 2545),
            ("Eric", 14938),
            ("Erik", 445),
            ("Ernest", 606),
            ("Erwan", 492),
            ("Erwin", 563),
            ("Esteban", 1476),
            ("Ethan", 3769),
            ("Etienne", 3779),
            ("Eugène", 620),
            ("Evan", 912),
            ("Ezio", 538),
            ("Fabian", 3892),
            ("Fabien", 1986),
            ("Fabio", 1078),
            ("Fabrce", 6668),
            ("Fabrizio", 818),
            ("Farid", 424),
            ("Fernand", 2023),
            ("Fernando", 444),
            ("Filippo", 762),
            ("Flavio", 900),
            ("Florent", 2512),
            ("Florian", 4813),
            ("Francesco", 2848),
            ("Francis", 7850),
            ("Francisco", 638),
            ("Franck", 1287),
            ("Franco", 1059),
            ("Frank", 1371),
            ("Franz", 925),
            ("François", 10479),
            ("François-Xavier", 450),
            ("Freddy", 3366),
            ("Frédéric", 14920),
            ("Félix", 1221),
            ("Gabin", 527),
            ("Gabriel", 5711),
            ("Gaetano", 986),
            ("Gary", 559),
            ("Gaspard", 1260),
            ("Gaston", 1055),
            ("Gauthier", 2049),
            ("Gaël", 964),
            ("Gaëtan", 2955),
            ("Geoffrey", 4406),
            ("Geoffroy", 678),
            ("Georges", 6122),
            ("Germain", 598),
            ("Ghislain", 1183),
            ("Giacomo", 512),
            ("Gianni", 927),
            ("Gilbert", 2359),
            ("Gilles", 3374),
            ("Gino", 562),
            ("Giovanni", 3174),
            ("Giuliano", 656),
            ("Giuseppe", 5859),
            ("Gregory", 1739),
            ("Grégoire", 1059),
            ("Grégory", 6539),
            ("Guido", 745),
            ("Guillaume", 7306),
            ("Gustave", 541),
            ("Guy", 9812),
            ("Gérald", 801),
            ("Gérard", 4072),
            ("Hadrien", 495),
            ("Hamza", 937),
            ("Harry", 444),
            ("Hasan", 579),
            ("Hassan", 853),
            ("Hector", 425),
            ("Henri", 5349),
            ("Henry", 576),
            ("Hervé", 1785),
            ("Hicham", 590),
            ("Hubert", 1971),
            ("Hugo", 7597),
            ("Hugues", 1463),
            ("Ibrahim", 1262),
            ("Igor", 553),
            ("Ilias", 458),
            ("Ilyas", 596),
            ("Isaac", 620),
            ("Ismail", 567),
            ("Ismaël", 557),
            ("Ivan", 822),
            ("Jacky", 693),
            ("Jacques", 13781),
            ("Jamal", 419),
            ("James", 801),
            ("Jan", 868),
            ("Jason", 2454),
            ("Jean", 34246),
            ("Jean-Baptiste", 587),
            ("Jean-Charles", 460),
            ("Jean-Christophe", 1144),
            ("Jean-Claude", 4959),
            ("Jean-François", 5180),
            ("Jean-Jacques", 1282),
            ("Jean-Louis", 2300),
            ("Jean-Luc", 4387),
            ("Jean-Marc", 3793),
            ("Jean-Marie", 5615),
            ("Jean-Michel", 2395),
            ("Jean-Paul", 2455),
            ("Jean-Philippe", 1880),
            ("Jean-Pierre", 6407),
            ("Jean-Pol", 1068),
            ("Jean-Yves", 1025),
            ("Jeremy", 1283),
            ("Jerôme", 588),
            ("Jessy", 503),
            ("Jimmy", 1661),
            ("Joachim", 1336),
            ("Joffrey
