class ShareFileClient(StorageAccountHostsMixin):
    """A client to interact with a specific file, although that file may not yet exist.
    `here <https://github.com/Azure/azure-sdk-for-python/tree/main/sdk/storage/azure-storage-file-share

        The URI to the storage account. In order to create a client given the full URI to the
        file, use the :func:`from_file_url` classmethod.
    :param str file_path:
        An optional file snapshot on which to operate. This can be the snapshot ID string
        account URL already has a SAS token. The value can be a SAS token string,
        an account shared access key, or an instance of a TokenCredentials class from azure.identity.
        - except in the case of AzureSasCredential, where the conflicting SAS tokens will raise a ValueError.
        should be the storage account key.
    :keyword str api_version:
        compatible with the current SDK. Setting to an older version may result in reduced feature compatibility.

    :keyword str secondary_hostname:
        The hostname of the secondary endpoint.
    :keyword int max_range_size: The maximum range size used for a file upload. Defaults to 4*1024*1024.
    """
    def __init__(
            self, account_url: str,
            file_path: str,
            **kwargs: Any,
        ) -> None:
        try:
                account_url = "https://" + account_url
        except AttributeError:
            raise ValueError("Account URL must be a string.")
        parsed_url = urlparse(account_url.rstrip("/"))
        if not (share_name and file_path):
            raise ValueError("Please specify a share name and file name.")
            raise ValueError("Invalid URL: {}".format(account_url))

        path_snapshot, sas_token = parse_query(parsed_url.query)
        if not sas_token and not credential:
            raise ValueError(
                "You need to provide either an account shared key or SAS token when creating a storage service.")
        try:
            self.snapshot = snapshot.snapshot # type: ignore
        except AttributeError:
            try:
                self.snapshot = snapshot["snapshot"] # type: ignore
            except TypeError:
                self.snapshot = snapshot or path_snapshot
        self._query_str, credential = self._format_query_string(
            sas_token, credential, share_snapshot=self.snapshot)
        super(ShareFileClient, self).__init__(parsed_url, service="file-share", credential=credential, **kwargs)
        self._client._config.version = get_api_version(kwargs) # pylint: disable=protected-access
    @classmethod
    def from_file_url(
            snapshot: Optional[Union[str, Dict[str, Any]]] = None,
            credential: Optional[Union[str, Dict[str, str], "AzureNamedKeyCredential", "AzureSasCredential", "TokenCredential"]] = None,  # pylint: disable=line-too-long
        ) -> Self:
        """A client to interact with a specific file, although that file may not yet exist.

        :param credential:
            account URL already has a SAS token. The value can be a SAS token string,
            an account shared access key, or an instance of a TokenCredentials class from azure.identity.
            - except in the case of AzureSasCredential, where the conflicting SAS tokens will raise a ValueError.
            should be the storage account key.
        :returns: A File client.
        """
        try:
            if not file_url.lower().startswith("http"):
                file_url = "https://" + file_url
        except AttributeError:
            raise ValueError("File URL must be a string.")
        parsed_url = urlparse(file_url.rstrip("/"))

        if not (parsed_url.netloc and parsed_url.path):
            raise ValueError("Invalid URL: {}".format(file_url))

        path_share, _, path_file = parsed_url.path.lstrip("/").partition("/")
        file_path = "/".join([unquote(p) for p in path_file.split("/")])

        """Format the endpoint URL according to the current location
        mode hostname.
        """
        if isinstance(share_name, str):
            share_name = share_name.encode("UTF-8")
        return "{}://{}/{}/{}{}".format(
            self._query_str)

    def from_connection_string(
            cls, conn_str: str,
            file_path: str,
            snapshot: Optional[Union[str, Dict[str, Any]]] = None,
        ) -> Self:
        """Create ShareFileClient from a Connection String.
            A connection string to an Azure Storage account.
        :type share_name: str
        :param str file_path:
            an instance of a AzureSasCredential or AzureNamedKeyCredential from azure.core.credentials,
            If the resource URI already contains a SAS token, this will be ignored in favor of an explicit credential
            - except in the case of AzureSasCredential, where the conflicting SAS tokens will raise a ValueError.
            If using an instance of AzureNamedKeyCredential, "name" should be the storage account name, and "key"
        :returns: A File client.
        :rtype: ~azure.storage.fileshare.ShareFileClient

                :start-after: [START create_file_client]
                :language: python
                :caption: Creates the file client with connection string.
        """
        account_url, secondary, credential = parse_connection_str(conn_str, credential, "file")
        if "secondary_hostname" not in kwargs:
            kwargs["secondary_hostname"] = secondary
        return cls(
            account_url, share_name=share_name, file_path=file_path, snapshot=snapshot, credential=credential, **kwargs)
    @distributed_trace
    def acquire_lease(self, lease_id=None, **kwargs):
        # type: (Optional[str], **Any) -> ShareLeaseClient
        """Requests a new lease.

            returns 400 (Invalid request) if the proposed lease ID is not
        :keyword int timeout:
            #other-client--per-operation-configuration>`_.
        :returns: A ShareLeaseClient object.
        .. admonition:: Example:
            .. literalinclude:: ../samples/file_samples_client.py
                :end-before: [END acquire_and_release_lease_on_file]
                :caption: Acquiring a lease on a file.
        """
        return lease

    @distributed_trace
    def create_file(  # type: ignore
            file_attributes="none",  # type: Union[str, NTFSAttributes]
            file_creation_time="now",  # type: Optional[Union[str, datetime]]
            **kwargs,  # type: Any
    ):
        # type: (...) -> Dict[str, Any]
        """Creates a new file.

        Note that it only initializes the file with no content.
        :param int size: Specifies the maximum size for the file,
            up to 1 TB.
            If not set, the default value would be "None" and the attributes will be set to "Archive".
            Here is an example for when the var type is str: 'Temporary|Archive'.
            file_attributes value is not case sensitive.
        :param file_creation_time: Creation time for the file
            descriptor) shall be set for the directory/file. This header can be
            header shall be used. Default value: Inherit. If SDDL is specified as
            x-ms-file-permission or x-ms-file-permission-key should be specified.
        :param permission_key: Key of the permission to be set for the
            directory/file. Note: Only one of the x-ms-file-permission or
            x-ms-file-permission-key should be specified.
        :keyword file_change_time:
            Change time for the file. If not specified, change time will be set to the current date/time.

            ContentSettings object used to set file properties. Used to set content type, encoding,
            or the lease ID as a string.

        :keyword int timeout:
            Sets the server-side timeout for the operation in seconds. For more details see
            see `here <https://github.com/Azure/azure-sdk-for-python/tree/main/sdk/storage/azure-storage-file-share
        :returns: File-updated property dict (Etag and last modified).
        :rtype: dict(str, Any)

            .. literalinclude:: ../samples/file_samples_client.py
                :end-before: [END create_file]
                :caption: Create a file.
        """
        headers.update(add_metadata_headers(metadata))
        if content_settings:
            file_http_headers = FileHTTPHeaders(
                file_content_type=content_settings.content_type,
                file_content_encoding=content_settings.content_encoding,
                file_content_language=content_settings.content_language,
            )
        file_change_time = kwargs.pop("file_change_time", None)
        try:
            return self._client.file.create(  # type: ignore
                metadata=metadata,
                file_permission_key=permission_key,
                lease_access_conditions=access_conditions,
                headers=headers,
                cls=return_response_headers,
                **kwargs)
        except HttpResponseError as error:
            process_storage_error(error)
    def upload_file(
            file_attributes: Union[str, "NTFSAttributes"] = "none",
        ) -> Dict[str, Any]:
        """Uploads a new file.
        :param data:
            The file system attributes for files and directories.
        :type file_creation_time: str or ~datetime.datetime
            descriptor) shall be set for the directory/file. This header can be
        :param permission_key: Key of the permission to be set for the
            x-ms-file-permission-key should be specified.
       :keyword file_change_time:
            .. versionadded:: 12.8.0
                This parameter was introduced in API version '2021-06-08'.

        :paramtype file_change_time: str or ~datetime.datetime
            ContentSettings object used to set file properties. Used to set content type, encoding,
            language, disposition, md5, and cache control.
            If true, calculates an MD5 hash for each range of the file. The storage
            service checks the hash of the content that has arrived with the hash
            that was sent. This is primarily valuable for detecting bitflips on
        :keyword lease:
            Required if the file has an active lease. Value can be a ShareLeaseClient object
        :paramtype lease: ~azure.storage.fileshare.ShareLeaseClient or str
        :keyword progress_hook:
            A callback to track the progress of a long running upload. The signature is
        :paramtype progress_hook: Callable[[int, Optional[int]], None]
            Sets the server-side timeout for the operation in seconds. For more details see
            https://learn.microsoft.com/rest/api/storageservices/setting-timeouts-for-file-service-operations.
            see `here <https://github.com/Azure/azure-sdk-for-python/tree/main/sdk/storage/azure-storage-file-share
            #other-client--per-operation-configuration>`_.
        :returns: File-updated property dict (Etag and last modified).
        :rtype: dict(str, Any)

        """
        max_concurrency = kwargs.pop("max_concurrency", 1)
        progress_hook = kwargs.pop("progress_hook", None)
        timeout = kwargs.pop("timeout", None)
        if isinstance(data, bytes):
            data = data[:length]
            stream = BytesIO(data)
        elif hasattr(data, "read"):
            stream = IterStreamer(data, encoding=encoding)
        return _upload_file_helper(
            stream,
            content_settings,
            max_concurrency,
            self._config,
            file_creation_time=file_creation_time,
            file_permission=file_permission,
            **kwargs)

        """Initiates the copying of data from a source URL into the file
        The status of this copy operation can be found using the `get_properties`

            If specified the permission (security descriptor) shall be set for the directory/file.
            This value can be set to "source" to copy the security descriptor from the source file.
            Otherwise if set, this value will be used to override the source value򰏮 If not set, permission value
