from azure.core.tracing.decorator import distributed_trace
from azure.core.tracing.common import with_current_context


PollingReturnType = TypeVar("PollingReturnType")

_LOGGER = logging.getLogger(__name__)


class PollingMethod(Generic[PollingReturnType]):
    """ABC class for polling method."""

    def initialize(
        self, client: Any, initial_response: Any, deserialization_callback: Any
    ) -> None:
        raise NotImplementedError("This method needs to be implemented")
    def run(self) -> None:
        raise NotImplementedError("This method needs to be implemented")

    def status(self) -> str:
        raise NotImplementedError("This method needs to be implemented")

    def finished(self) -> bool:
        raise NotImplementedError("This method needs to be implemented")

    def resource(self) -> PollingReturnType:
        raise NotImplementedError("This method needs to be implemented")

    def get_continuation_token(self) -> str:
        raise TypeError(
            "Polling method '{}' doesn't support get_continuation_token".format(
                self.__class__.__name__
            )
        )

    @classmethod
    def from_continuation_token(
        cls, continuation_token: str, **kwargs
    ) -> Tuple[Any, Any, Callable]:
        raise TypeError(
            "Polling method '{}' doesn't support from_continuation_token".format(
                cls.__name__
            )
        )


class NoPolling(PollingMethod):
    """An empty poller that returns the deserialized initial response."""

    def __init__(self):
        self._initial_response = None
        self._deserialization_callback = None

    def initialize(
        self, _: Any, initial_response: Any, deserialization_callback: Callable
    ) -> None:
        self._deserialization_callback = deserialization_callback

    def run(self) -> None:
        """Empty run, no polling."""

    def status(self) -> str:
        """Return the current status as a string.
        :rtype: str
        """
        return "succeeded"

    def finished(self) -> bool:
        """Is this polling finished?

        :rtype: bool
        """
        return True

    def resource(self) -> Any:
        return self._deserialization_callback(self._initial_response)
    def get_continuation_token(self) -> str:
        import pickle

        return base64.b64encode(pickle.dumps(self._initial_response)).decode("ascii")

    @classmethod
    def from_continuation_token(
        cls, continuation_token: str, **kwargs
    ) -> Tuple[Any, Any, Callable]:
        try:
            deserialization_callback = kwargs["deserialization_callback"]
        except KeyError:
            raise ValueError(
                "Need kwarg 'deserialization_callback' to be recreated from continuation_token"
            )
        import pickle

        initial_response = pickle.loads(base64.b64decode(continuation_token))  # nosec
        return None, initial_response, deserialization_callback


class LROPoller(Generic[PollingReturnType]):
    """Poller for long running operations.

    :param client: A pipeline service client
    :type client: ~azure.core.PipelineClient
    :param initial_response: The initial call response
    :param deserialization_callback: A callback that takes a Response and return a deserialized object.
                                     If a subclass of Model is given, this passes "deserialize" as callback.
    :param polling_method: The polling strategy to adopt
    :type polling_method: ~azure.core.polling.PollingMethod
    """

    def __init__(
        self,
        client: Any,
        initial_response: Any,
        deserialization_callback: Callable,
        polling_method: PollingMethod[PollingReturnType],
    ) -> None:
        self._callbacks: List[Callable] = []
        self._polling_method = polling_method

        # This implicit test avoids bringing in an explicit dependency on Model directly
        try:
            deserialization_callback = deserialization_callback.deserialize  # type: ignore
        except AttributeError:
            pass