# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""The PyTorch estimator class."""

import logging

from azureml.core._experiment_method import experiment_method

from ..estimator._framework_base_estimator import _FrameworkBaseEstimator
from .._estimator_helper import _estimator_submit_method


module_logger = logging.getLogger(__name__)


class PyTorch(_FrameworkBaseEstimator):
    """Represents an estimator for training in PyTorch experiments.

    DEPRECATED. Use the :class:`azureml.core.script_run_config.ScriptRunConfig` object with your own
    defined environment or one of the Azure ML PyTorch curated environments. For an introduction to
    configuring PyTorch experiment runs with ScriptRunConfig, see
    `Train PyTorch models at scale with Azure Machine Learning
    <https://docs.microsoft.com/azure/machine-learning/how-to-train-pytorch?view=azure-ml-py>`_.

    Supported versions: 1.0, 1.1, 1.2, 1.3, 1.4, 1.5, 1.6

    .. remarks::

            When submitting a training job, Azure ML runs your script in a conda environment within
            a Docker container. The PyTorch containers have the following dependencies installed.

            | Dependencies | PyTorch 1.0/1.1/1.2/1.3/    | PyTorch 1.4/1.5/1.6 |
            | ---------------------- | ----------------- | -------------   |
            | Python                 | 3.6.2             | 3.6.2           |
            | CUDA  (GPU image only) | 10.0              | 10.1            |
            | cuDNN (GPU image only) | 7.6.3             | 7.6.3           |
            | NCCL  (GPU image only) | 2.4.8             | 2.4.8           |
            | azureml-defaults       | Latest            | Latest          |
            | OpenMpi                | 3.1.2             | 3.1.2           |
            | horovod                | 0.18.1            | 0.18.1/0.19.1/0.19.5  |
            | miniconda              | 4.5.11            | 4.5.11          |
            | torch                  | 1.0/1.1/1.2/1.3.1 | 1.4.0/1.5.0/1.6.0     |
            | torchvision            | 0.4.1             | 0.5.0           |
            | git                    | 2.7.4             | 2.7.4           |
            | tensorboard            | 1.14              | 1.14            |
            | future                 | 0.17.1            | 0.17.1          |

            The Docker images extend Ubuntu 16.04.

            To install additional dependencies, you can either use the ``pip_packages`` or ``conda_packages``
            parameter. Or, you can specify the ``pip_requirements_file`` or ``conda_dependencies_file`` parameter.
            Alternatively, you can build your own image, and pass the ``custom_docker_image`` parameter to the
            estimator constructor.

            For more information about Docker containers used in PyTorch training, see
            https://github.com/Azure/AzureML-Containers.

            The PyTorch estimator supports distributed training across CPU and GPU clusters using
            `Horovod <https://github.com/uber/horovod>`_, an open-source, all reduce framework for distributed
            training.  For examples and more information about using PyTorch in distributed training, see the
            tutorial `Train and register PyTorch models at scale with Azure Machine Learning
            <https://d򧠈cs.microsoft.com/azure/machine-learning/how-to-train-pytorch>`_.

    :param source_directory: A local directory containing experiment configuration files.
    :type source_directory: str
    :param compute_target: The compute target where training will happen. This can either be an object or the
        string "local".
    :type compute_target: azureml.core.compute_target.AbstractComputeTarget or str
    :param vm_size: The VM size of the compute target that will be created for the training. Supported values:
        Any `Azure VM size
        <https://docs.microsoft.com/azure/cloud-services/cloud-services-sizes-specs>`_.
    :type vm_size: str
    :param vm_priority: The VM priority of the compute target that will be created for the training. If not
        specified, 'dedicated' is used.

        Supported values: 'dedicated' and 'lowpriority'.

        This takes effect only when the ``vm_size param`` is specified in the input.
    :type vm_prior
