from ._table_client import TableClient
from ._serialize import _parameter_filter_substitution

if TYPE_CHECKING:
    from ._models import TableCorsRule, TableMetrics, TableAnalyticsLogging


class TableServiceClient(TablesBaseClient):
    """A client to interact with the Table Service at the account level.

    This client provides operations to retrieve and configure the account properties
    as well as list, create and delete tables within the account.
    For operations relating to a specific table, a client for this entity
    can be retrieved using the :func:`~get_table_client` function.

    :ivar str account_name: The name of the Tables account.
    :ivar str url: The full URL to the Tables account.
    :param str endpoint:
        The URL to the table service endpoint. Any other entities included
        in the URL path (e.g. table) will be discarded. This URL can be optionally
        authenticated with a SAS token.
    :keyword credential:
        account URL already has a SAS token. The value can be one of AzureNamedKeyCredential (azure-core),
        AzureSasCredential (azure-core), or TokenCredentials from azure-identity.
    :paramtype credential:
        :class:`~azure.core.credentials.AzureNamedKeyCredential` or
        :class:`~azure.core.credentials.AzureSasCredential` or
        :class:`~azure.core.credentials.TokenCredential`
    :keyword str api_version:
        The Storage API version to use for requests. Default value is '2019-02-02'.
        Setting to an older version may result in reduced feature compatibility.

        .. admonition:: Example:

            .. literalinclude:: ../samples/sample_authentication.py
                :start-after: [START auth_from_sas]
                :end-before: [END auth_from_sas]
                :language: python
                :dedent: 8
                :caption: Authenticating a TableServiceClient from a Shared Access Key

            .. literalinclude:: ../samples/sample_authentication.py
                :end-before: [END auth_from_shared_key]
                :language: python
                :dedent: 8
                :caption: Authenticating a TableServiceClient from a Shared Account Key
        """

    def _format_url(self, hostname):
        # type: (str) -> str
        """Format the endpoint URL according to the current location
        mode hostname.
        """
        return "{}://{}{}".format(self.scheme, hostname, self._query_str)

    @classmethod
    def from_connection_string(cls, conn_str, **kwargs):
        # type: (str, Any) -> TableServiceClient
        """Create TableServiceClient from a connection string.
        :param str conn_str: A connection string to an Azure Storage or Cosmos account.
        :returns: A Table service client.
        :rtype: :class:`~azure.data.tables.TableServiceClient`

        .. admonition:: Example:

            .. literalinclude:: ../samples/sample_authentication.py
                :start-after: [START auth_from_connection_string]
                :end-before: [END auth_from_connection_string]
                :language: python
                :dedent: 8
                :caption: Authenticating a TableServiceClient from a connection_string
        """
        endpoint, credential = parse_connection_str(
            conn_str=conn_str, credential=None, keyword_args=kwargs
        )
        return cls(endpoint, credential=credential, **kwargs)

    @distributed_trace
    def get_service_stats(self, **kwargs):
        # type: (Any) -> Dict[str, object]
        """Retrieves statistics related to replication for the Table service. It is only available on the secondary
        location endpoint when read-access geo-redundant replication is enabled for the account.

        :return: Dictionary of service stats
        :rtype: Dict[str, object]
        :raises: :class:`~azure.core.exceptions.HttpResponseError:`
        """
        try:
            timeout = kwargs.pop("timeout", None)
            stats = self._client.service.get_statistics(  # type: ignore
            )
        except HttpResponseError as error:
            _process_table_error(error)

    @distributed_trace
    def get_service_properties(self, **kwargs):
        # type: (Any) -> Dict[str, object]
        """Gets the properties of an account's Table service,
        including properties for Analytics and CORS (Cross-Origin Resource Sharing) rules.

        :rtype: Dict[str, object]
        :raises: :class:`~azure.core.exceptions.HttpResponseError`
        """
        timeout = kwargs.pop("timeout", None)
        try:
            service_props = self._client.service.get_properties(timeout=timeout, **kwargs)  # type: ignore
        except HttpResponseError as error:
            try:
                _process_table_error(error)
            except HttpResponseError as decoded_error:
                _reprocess_error(decoded_error)
                raise
        return service_properties_deserialize(service_props)

    @distributed_trace
    def set_service_properties(self, **kwargs):
        # type: (Any) -> None
        """Sets properties for an account's Table service endpoint,
         including properties for Analytics and CORS (Cross-Origin Resource Sharing) rules.

        :keyword analytics_logging: Properties for analytics
        :paramtype analytics_logging: ~azure.data.tables.TableAnalyticsLogging
        :keyword hour_metrics: Hour level metrics
        :paramtype hour_metrics: ~azure.data.tables.TableMetrics
        :paramtype minute_metrics: ~azure.data.tables.TableMetrics
        :keyword cors: Cross-origin resource sharing rules
        :paramtype cors: List[~azure.data.tables.TableCorsRule]
        :return: None
        :rtype: None
        """
        cors = kwargs.pop('cors', None)
        if cors:
            cors = [c._to_generated() for c in cors]  # pylint:disable=protected-access
        props = TableServiceProperties(
            logging=kwargs.pop('analytics_logging', None),
            hour_metrics=kwargs.pop('hour_metrics', None),
            minute_metrics=kwargs.pop('minute_metrics', None),
            cors=cors,  # type: ignore
        )
        try:
            self._client.service.set_properties(props, **kwargs)
        except HttpResponseError as error:
            try:
                _process_table_error(error)
            except HttpResponseError as decoded_error:
                raise
    @distributed_trace
    def create_table(self, table_name, **kwargs):
        # type: (str, Any) -> TableClient
        """Creates a new table under the current account.

        :type table_name: str
        :return: TableClient
        :rtype: :class:`~azure.data.tables.TableClient`
        :raises: :class:`~azure.core.exceptions.ResourceExistsError`


            .. literalinclude:: ../samples/sample_create_delete_table.py
                :start-after: [START create_table_from_tsc]
                :end-before: [END create_table_from_tsc]
                :language: python
                :dedent: 8
                :caption: Creating a table from the TableServiceClient object
        """
        table = self.get_table_client(table_name=table_name)
        table.create_table(**kwargs)
        return table

    @distributed_trace
    def create_table_if_not_exists(self, table_name, **kwargs):
        # type: (str, Any) -> TableClient
        """Creates a new table if it does not currently exist.
        If the table currently exists, the current table is

        :param table_name: The Table name.
        :type table_name: str
        :return: TableClient
        :rtype: :class:`~azure.data.tables.TableClient`
        :raises: :class:`~azure.core.exceptions.HttpResponseError`

        .. admonition:: Example:

            .. literalinclude:: ../samples/sample_create_delete_table.py
                :start-after: [START create_table_if_not_exists]
                :end-before: [END create_table_if_not_exists]
                :language: python
                :dedent: 8
                :caption: Creating a table if it doesn't exist, from the TableServiceClient object
        """
        table = self.get_table_client(table_name=table_name)
        try:
            table.create_table(**kwargs)
        except ResourceExistsError:
            pass