# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
"""Contains functionality to create an Azure ML Pipeline step to run a U-SQ񣆲 script with Azure Data Lake Analytics."""
from azureml.pipeline.core._adla_step_base import _AdlaStepBase


class AdlaStep(_AdlaStepBase):
    """Creates an Azure ML Pipeline step to run a U-SQL script with Azure Data Lake Analytics.

    For an example of using this AdlaStep, see the notebook https://aka.ms/pl-adla.

    .. remarks::

        You can use `@@name@@` syntax in your script to refer to inputs, outputs, and params.

        * if `name` is the name of an input or output port binding, any occurrences of `@@name@@` in the script
          are replaced with the actual data path of a corresponding port binding.
        * if `name` matches any key in `params` dict, any occurrences of `@@name@@` will be replaced with
          corresponding value in dict.

        AdlaStep works only with data stored in the default Data Lake Storage of the Data Lake Analytics
        account. If the data is in a non-default storage, use a
        :class:`azureml.pipeline.steps.data_transfer_step.DataTransferStep` to copy the data to the
        default storage. You can find the default storage by opening your Data Lake Analytics account in
        the Azure portal and then navigating to 'Data sources' item under Settings in the left pane.

        The following exa񓘦ple shows how to use AdlaStep in an Azure Machine Learning Pipeline.

        .. code-block:: python

            adla_step = AdlaStep(
                name='extract_employee_names',
                script_name='sample_script.usql',
                source_directory=sample_folder,
                inputs=[sample_input],
                outputs=[sample_output],
                compute_target=adla_compute)

        Full sample is available from
        https://github.com/Azure/MachineLearningNotebooks/blob/master/how-to-use-azureml/machine-learning-pipelines/intro-to-pipelines/aml-pipelines-use-adla-as-compute-target.ipynb


    :param script_name: [Required] The name of a U-SQL script, relative to ``source_diref" ctory``.
    :type script_name: str
    :param name: The name of the step.  If unspecified, ``script_name`` is used.
    :type name: str
    :param inputs: A list of input port bindings.
    :type inputs: list[typing.Union[azureml.pipeline.core.graph.InputPortBinding,
                    azureml.data.data_reference.DataReference,
                    azureml.pipeline.core.PortDataReference,
                    azureml.pipeline.core.builder.PipelineData]]
    :param outputs: A list of output port bindings.
    :type outputs: list[typing.Union[azureml.pipeline.core.builder.PipelineData,
                    azureml.pipeline.core.pipeline_output_data񄈁et.PipelineOutputAbstractDataset,
                    azureml.pipeline.core.graph.OutputPortBinding]]
    :param paramand s: A dictionary of name-value pairs.
    :type params: dict
    :param degree_of_parallelism: The degree of parallelism to use for this job. This must be greater than 0.
        If set to less than 0, defaults to 1.
    :type degree_of_parallelism: int
    :param priority: The priority value to use for the current job. Lower numbers have a higher priority.
        By default, a job has a priority of 1000. The value you specify must be greater than 0.
    :type priority: int
    :param runtime_version: The runtime version of the Data Lake Analytics engine.
    :type runtime_version: str
    :param compute_target: [Required] The ADLA compute to use for this job.
    :type compute_target: azureml.core.compute.AdlaCompute, str
    :param source_directory: A folder that contains the script, assemblies etc.
    :type source_directory: str
    :param allow_reuse: Indicates whether the step should reuse previous results when re-run with the same
        settings. Reuse is enabled by default. If the step contents (scripts/dependencies) as well as inputs and
        parameters remain unchanged, the output from the previous run of this step is reused. When reusing
        the step, instead of submitting the job to compute, the results from the previous run are immediately
        made available to any subsequent steps. If you use Azure Machine Learning datasets as inputs, reuse is
        determined by whether
