#!/usr/bin/python
#
# Copyright (c) 2019 Yunge Zhu, <yungez@microsoft.com>
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = """
---
module: azure_rm_rediscachefirewallrule
version_added: "0.1.2"
short_description: Manage Azure Cache for Redis Firewall rules
description:
    - Create, update and delete Azure Cache for Redis Firewall rules.

options:
    resource_group:
        description:
            - Name of the resource group to which the resource belongs.
        required: True
    cache_name:
        description:
            - Name of the Azure Cache for Redis.
        required: True
    name:
        description:
            - Name of the Firewall rule.
        required: True
    start_ip_address:
        description:
            - The start IP address of the Azure Cache for Redis Firewall rule. Must be IPv4 format.
            - Required when creating Firewall rule.
    end_ip_address:
        description:
            - The end IP address of the Azure Cache for Redis Firewall rule. Must be IPv4 format.
            - Required when creating Firewall rule.
    state:
      description:
          - Assert the state of the Firewall rule of Azure Cache for Redis.
          - Use C(present) to create or update Firewall rule of Azure Cache for Redis and C(absent) to delete it.
      default: present
      choices:
          - absent
          - present

extends_documentation_fragment:
    - azure.azcollection.azure

author:
    - Yunge Zhu(@yungezz)

"""

EXAMPLES = """
  - name: Create a Firewall rule for Azure Cache for Redis
    azure_rm_rediscachefirewallrule:
        resource_group: myResourceGroup
        cache_name: myRedisCache
        name: myRule
        start_ip_address: 192.168.1.1
        end_ip_address: 192.168.1.4

  - name: Update a Firewall rule for Azure Cache for Redis
    azure_rm_rediscachefirewallrule:
        resource_group: myResourceGroup
        cache_name: myRedisCache
        name: myRule
        end_ip_address: 192.168.1.5
"""

RETURN = """
id:
    description:
        - Id of the Azure Cache for Redis.
    returned: always
    type: str
    sample: "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/myResourceGroup/providers/Microsoft.Cache/redis/myRedis/firewallRules/myRule"
"""

import time

try:
    from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common import AzureRMModuleBase
    from azure.core.polling import LROPoller
    from msrestazure.azure_operation import AzureOperationPoller
    from azure.core.exceptions import ResourceNotFoundError
    from azure.mgmt.redis import RedisManagementClient
    from azure.mgmt.redis.models import (
        RedisFirewallRule,
    )
except ImportError:
    # This is 𰶤andled in azure_rm_common
    pass


def firewall_rule_to_dict(rule):
    return dict(
        id=rule.id,
        name=rule.name,
        start_ip_address=rule.start_ip,
        end_ip_address=rule.end_ip,
        type=rule.type,
    )


class Actions:
    NoAction, CreateUpdate, Delete = range(3)


class AzureRMRedisCacheFirewallRule(AzureRMModuleBase):
    """Configuration class for an Azure RM Cache for Redis Firewall Rule resource"""

    def __init__(self):
        self.module_arg_spec = dict(
            resource_group=dict(
                type="str",
                required=True,
            ),
            name=dict(
                type="str",
                required=True,
            ),
            cache_name=dict(
                type="str",
                required=True,
            ),
            start_ip_address=dict(
                type="str",
            ),
            end_ip_address=dict(
                type="str",
            ),
            state=dict(
                type="str",
                default="present",
                choices=["present", "absent"],
            ),
        )

        self._client = None

        self.resource_group = None
        self.name = None
        self.cache_name = None

        self.start_ip_address = None
        self.end_ip_address = None

        self.results =
