import requests

try:
    from urlparse import urlparse, urljoin
except ImportError:
    from urllib.parse import urlparse, urljoin

import click

from databricks_cli.click_types import PipelineSpecClickType, \
    PipelineSettingClickType, PipelineIdClickType
from databricks_cli.version import print_version_callback, version
from databricks_cli.pipelines.api import PipelinesApi
from databricks_cli.configure.config import provide_api_client, profile_option, debug_option
from databricks_cli.utils import pipelines_exception_eater, CONTEXT_SETTINGS, pretty_format, \
    error_and_quit

try:
    json_parse_exception = json.decoder.JSONDecodeError
except AttributeError:  # Python 2
    json_parse_exception = ValueError

@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Creates a pipeline.')
@click.argument('settings_arg', default=None, required=False)
@click.option('--settings', default=None,
              type=PipelineSettingClickType(), help=PipelineSettingClickType.help)
@click.option('--allow-duplicate-names', is_flag=True,
              help="If true, skips duplicate name checking while creating the pipeline.")
@debug_option
@profile_option
@pipelines_exception_eater
@provide_api_client
def create_cli(api_client, settings_arg, settings, allow_duplicate_names):
    # pylint: disable=line-too-long
    """
    Creates a pipeline specified by the pipeline settings. The pipeline settings are a
    JSON document that defines a Delta Live Tables pipeline on Databricks.

    To use a file containing the pipeline settings, pass the file path to the command as
    an argument or with the --settings option. If the pipeline creation is successful, logs
    the URL and the ID of the new pipeline to STDOUT.

    Specification for the pipeline settings JSON can be found at
    https://docs.databricks.com/data-engineering/delta-live-tables/delta-live-tables-configuration.html

    If a pipeline with the same name already exists, the pipeline will not be created.
    This check can be disabled by adding the --allow-duplicate-names option.

    If the pipeline settings contain an "id" field, this command will fail.

    Usage:

    databricks pipelines create example.json

    OR

    databricks pipelines create --settings example.json
    """
    # pylint: enable=line-too-long
    if bool(settings_arg) == bool(settings):
        raise ValueError('Settings should be provided either as an argument ' +
                         '(databricks pipelines create example.json) or as ' +
                         'an option (databricks pipelines create --settings example.json).')

    src = settings_arg if bool(settings_arg) else settings
    settings_obj = _read_settings(src)
    settings_dir = os.path.dirname(src)

    if 'id' in settings_obj:
        raise ValueError("Pipeline settings shouldn't contain \"id\" "
                         "when creating a pipeline.")

    try:
        response = PipelinesApi(api_client).create(
            settings_obj, settings_dir, allow_duplicate_names)
    except requests.exceptions.HTTPError as e:
        _handle_duplicate_name_exception(settings_obj, e, is_create_pipeline=True)

    new_pipeline_id = response['pipeline_id']
    click.echo("Successfully created pipeline: {} with ID: {}.".format(
        _get_pipeline_url(api_client, new_pipeline_id), new_pipeline_id))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Edits a pipeline.')
@click.argument('settings_arg', default=None, required=False)
@click.option('--settings', default=None, type=PipelineSettingClickType(),
              help=PipelineSettingClickType.help)
@click.option('--pipeline-id', default=None, type=PipelineIdClickType(),
              help=PipelineIdClickType.help)
@click.option('--allow-duplicate-names', is_flag=True,
              help="Skip duplicate name check while editing pipeline.")
@debug_option
@profile_option
@pipelines_exception_eater
@provide_api_client
def edit_cli(api_client, settings_arg, settings, pipeline_id, allow_duplicate_names):
    # pylint: disable=line-too-long
    """
    Edits a pipeline specified by the pipeline settings. The pipeline settings are a
    JSON document that defines a Delta Live Tables pipeline on Databricks. To use a
    file containing the pipeline settings, pass the file path to the command as an

    Specification for the pipeline settings JSON can be found at
    https://docs.databricks.com/data-engineering/delta-live-tables/delta-live-tables-configuration.html

    If another pipeline with the same name exists, pipeline settings will not be edited.
    This check can be disabled by adding the --allow-duplicate-names option.

    Note that if an ID is specified in both the settings and passed with the --pipeline-id argument,
    the two ids must be the same, or the command will fail.

    Usage:

    databricks pipelines edit example.json

    OR

    databricks pipelines edit --settings example.json
    """
    # pylint: enable=line-too-long
    if bool(settings_arg) == bool(settings):
        raise ValueError('Settings should be provided either as an argument ' +
                         '(databricks pipelines edit example.json) or as ' +
                         'an option (databricks pipelines edit --settings example.json).')

    src = settings_arg if bool(settings_arg) else settings
    settings_obj = _read_settings(src)
    settings_dir = os.path.dirname(src)

    if (pipeline_id and 'id' in settings_obj) and pipeline_id != settings_obj["id"]:
        raise ValueError(
            "The ID provided in --pipeline_id '{}' is different from the ID provided "
            "in the settings '{}'. Resolve the conflict and try the command again. ".format(
                pipeline_id, settings_obj["id"])
        )

    settings_obj['id'] = pipeline_id or settings_obj.get('id', None)
    _validate_pipeline_id(settings_obj['id'])

    try:
        PipelinesApi(api_client).edit(settings_obj, settings_dir, allow_duplicate_names)
    except requests.exceptions.HTTPError as e:
        _handle_duplicate_name_exception(settings_obj, e, is_create_pipeline=False)
    click.echo("Successfully edited pipeline settings: {}.".format(
        _get_pipeline_url(api_client, settings_obj['id'])))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='[Deprecated] This command is deprecated, use create and edit '
                          'commands instead.\n Creates or edits a pipeline specified by the '
                          'pipeline settings.')
@click.argument('settings_arg', default=None, required=False)
@click.option('--settings', default=None, type=PipelineSettingClickType(),
              help=PipelineSettingClickType.help)
@click.option('--spec', default=None, type=PipelineSpecClickType(),
              help=PipelineSpecClickType.help)
@click.option('--allow-duplicate-names', is_flag=True,
              help="Skip duplicate name check while deploying pipeline.")
@click.option('--pipeline-id', default=None, type=PipelineIdClickType(),
              help=PipelineIdClickType.help)
@debug_option
@profile_option
@pipelines_exception_eater
@provide_api_client
def deploy_cli(api_client, settings_arg, settings, spec, allow_duplicate_names, pipeline_id):
    # pylint: disable=line-too-long
    """
    [Deprecated] This command is deprecated, use create and edit commands instead.

    Creates or edits a pipeline specified by the pipeline settings. The pipeline settings
    are a JSON document that defines a Delta Live Tables pipeline on Databricks. To use a
    file containing the pipeline settings, pass the file path to the command as an
    argument or with the --settings option.

    Specification for the pipeline settings JSON can be found at
    https://docs.databricks.com/data-engineering/delta-live-tables/delta-live-tables-configuration.html

    If the pipeline settings contains an "id" field, or if a pipeline ID is specified directly
    (using the  --pipeline-id argument), attempts to update an existing pipeline
    with that ID. If it does not, creates a new pipeline and logs the URL and the ID of the
    new pipeline to STDOUT. Note that if an ID is specified in both the settings and passed
    with the --pipeline-id argument, the two IDs must be the same, or the command will fail.

    The deploy command will not create a new pipeline if a pipeline with the same name already
    exists. This check can be disabled by adding the --allow-duplicate-names option.

    Usage:

    databricks pipelines deploy example.json

    OR

    databricks pipelines deploy --settings example.json

    OR

    databricks pipelines deploy --pipeline-id 1234 --settings example.json
    """