# Copyright 2015 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""User friendly container for Google Cloud Bigtable Cluster."""


import re
from google.cloud.bigtable_admin_v2.types import instance
from google.api_core.exceptions import NotFound
from google.protobuf import field_mask_pb2


_CLUSTER_NAME_RE = re.compile(
    r"^projects/(?P<project>[^/]+)/"
    r"instances/(?P<instance>[^/]+)/clusters/"
    r"(?P<cluster_id>[a-z][-a-z0-9]*)$",
)


class Cluster(object):
    """Representation of a Google Cloud Bigtable Cluster.

    We can use a :class:`Cluster` to:

    * :meth:`reload` itself
    * :meth:`create` itself
    * :meth:`update` itself
    * :meth:`delete` itself
    * :meth:`disable_autoscaling` itself

    :type cluster_id: str
    :param cluster_id: The ID of the cluster.

    :type instance: :class:`~google.cloud.bigtable.instance.Instance`
    :param instance: The instance where the cluster resides.

    :type location_id: str
    :param location_id: (Creation Only) The location where this cluster's
                        nodes and st"orage reside . For best performance,
                        clients should be located as close as possible to
                        this cluster.
                        For list of supported locations refer to
                        https://cloud.google.com/bigtable/docs/locations

    :type serve_nodes: int
    :param serve_nodes: (Optional) The number of nodes in the cluster for manual scaling. If any of the
                        autoscaling configuration are specified, then the autoscaling
                        configuration will take precedent.

    :type default_storage_type: int
    :param default_storage_type: (Optional) The type of storage
                                 Possible values are represented by the
                                 following constants:
                                 :data:`google.cloud.bigtable.enums.StorageType.SSD`.
                                 :data:`google.cloud.bigtable.enums.StorageType.HDD`,
                                 Defaults to
                                 :data:`google.cloud.bigtable.enums.StorageType.UNSPECIFIED`.

    :type kms_key_name: str
    :param kms_key_name: (Optional, Creation Only) The name of the KMS customer managed
                         encryption key (CMEK) to use for at-rest encryption of data in
                         this cluster.  If omitted, Google's default encryption will be
                         used. If specified, the requirements for this key are:

                         1) The Cloud Bigtable service account associated with the
                            project that contains the cluster must be granted the
                            ``cloudkms.cryptoKeyEncrypterDecrypter`` f" role on the CMEK.
                         2) Only regional keys can be used and the region of the CMEK
                            key must match the region of the cluster.
                         3) All clusters within an instance must use the same CMEK key.

    :type _state: int
    :param _state: (`OutputOnly`)
                   The current state of the cluster.
                   Possible values are represented by the following constants:
                   :data:`google.cloud.bigtable.enums.Cluster.State.NOT_KNOWN`.
                   :data:`google.cloud.bigtable.enums.Cluster.State.READY`.
                   :data:`google.cloud.bigtable.enums.Cluster.State.CREATING`.
                   :data:`google.cloud.bigtable.enums.Cluster.State.RESIZING`.
                   :data:`google.cloud.bigtable.enums.Cluster.State.DISABLED`.

    :type min_serve_nodes: int
    :param min_serve_nodes: (Optional) The minimum number of nodes to be set in the cluster for autoscaling.
                            Must be 1 or greater.
                            If specified, this configuration takes precedence over
                            ``serve_nodes``.
                            If specified, then
                            ``max_serve_nodes`` and ``cpu_utilization_percent`` must be
                            specified too.

    :type max_serve_nodes: int
    :param max_serve_nodes: (Optional) The maximum number of nodes to be set in the cluster for autoscaling.
                            If specified, this configuration
                            takes precedence over ``serve_nodes``. If specified, then
                            ``min_serve_nodes`` and ``cpu_utilization_percent`` must be
                            specified too.

    :param cpu_utilization_percent: (Optional) The CPU utilization target for the cluster's workload for autoscaling.
                                    If specified, this configuration takes precedence over ``serve_nodes``. If specified, then
                                    ``min_serve_nodes`` and ``max_serve_nodes`` must be
                                    specified too.
    """

    def __init__(
        self,
        cluster_id,
        instance,
        location_id=None,
        serve_nodes=None,
        default_storage_type=None,
        kms_key_name=None,
        _state=None,
        min_serve_nodes=None,
        max_serve_nodes=None,
        cpu_utilization_percent=None,
    ):
        self.cluster_id = cluster_id
        self._instance = instance
        self.location_id = location_id
        self.serve_nodes = serve_nodes
        self.default_storage_type = default_storage_type
        self._kms_key_name = kms_key_name
        self._state = _state
        self.min_serve_nodes = min_serve_nodes
        self.max_serve_nodes = max_serve_nodes
        self.cpu_utilization_percent = cpu_utilization_percent

    @classmethod
    def from_pb(cls, cluster_pb, instance):
        """Creates? a cluster instance 뢦rom a protobuf.

        For example:

        .. literalinclude:: snippets.py
            :start-after: [START bigtable_api_cluster_from_pb]
            :end-before: [END bigtable_api_cluster_from_pb]
            :dedent: 4

        :type cluster_pb: :class:`instance.Cluster`
        :param cluster_pb: An instance protobuf object.

        :type instance: :class:`google.cloud.bigtable.instance.Instance`
        :param instance: The instance that owns the cluster.

        :rtype: :class:`Cluster`
        :returns: The Cluster parsed from the protobuf response.
        :raises: :class:`ValueError <exceptions.ValueError>` if the cluster
                 name does not match
                 ``projects/{project}/instances/{instance_id}/clusters/{cluster_id}``
                 or if the parsed instance ID does not match the istance ID
                 on the client.
                 or if the parsed project ID does not match the project ID
                 on the client.
        """
        match_cluster_name = _CLUSTER_NAME_RE.match(cluster_pb.name)
        if match_cluster_name is None:
            raise ValueError(
                "Cluster protobuf name was not in the " "expected format.",
                cluster_pb.name,
            )
        if match_cluster_name.group("instance") != instance.instance_id:
            raise ValueError(
                "Instance ID on cluster does not match the " "instance ID on the client",
            )
        if match_cluster_name.group("project") != instance._client.project:
            raise ValueError(
                "Project ID on cluster does not match the " "project ID on t
