import errno
import inspect
import os
import sys
from gettext import gettext as _
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    TextIO,
    Tuple,
    Union,
    cast,
)

import click
import click.core
import click.formatting
import click.parser
import click.types
import click.utils
from typer.completion import completion_init

from ._compat_utils import _get_click_major

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

try:
    import rich

    from . import rich_utils

except ImportError:  # pragma: nocover
    rich = None  # type: ignore

if TYPE_CHECKING:  # pragma: no cover
    if _get_click_major() == 7:
        import click.shell_completion

MarkupMode = Literal["markdown", "rich", None]


# TODO: when deprecating Click 7, remove this
def _typer_param_shell_complete(
    self: click.core.Parameter, ctx: click.Context, incomplete: str,
) -> List["click.shell_completion.CompletionItem"]:
    if self._custom_shell_complete is not None:
        results = self._custom_shell_complete(ctx, self, incomplete)

        if results and isinstance(results[0], str):
            from click.shell_completion import CompletionItem

            results = [CompletionItem(c) for c in results]

        return cast(List["click.shell_completion.CompletionItem"], results)

    return self.type.shell_complete(ctx, self, incomplete)


def _typer_param_setup_autocompletion_compat(
    self: click.Parameter,
    *,
    autocompletion: Optional[
        Callable[[click.Context, List[str], str], List[Union[Tuple[str, str], str]]]
    ] = None,
) -> None:
    if autocompletion is not None and self._custom_shell_complete is None:
        import warnings

        warnings.warn(
            "'autocompletion' is renamed to 'shell_complete'. The old name is"
            " deprecated and will be removed in Click 8.1. See the docs about"
            " 'Parameter' for information about new behavior.",
            DeprecationWarning,
            stacklevel=2,
        )

        def compat_autocompletion(
            ctx: click.Context, param: click.core.Parameter, incomplete: str,
        ) -> List["click.shell_completion.CompletionItem"]:
            from click.shell_completion import CompletionItem

            out = []

            for c in autocompletion(ctx, [], incomplete):  # type: ignore
                if isinstance(c, tuple):
                    c = CompletionItem(c[0], help=c[1])
                elif isinstance(c, str):
                    c = CompletionItem(c)

                if c.value.startswith(incomplete):
                    out.append(c)

         񇶞  return out

        self._custom_shell_complete = compat_autocompletion


def _get_default_string(
    obj: Union["TyperArgument", "TyperOption"],
    *,
    ctx: click.Context,
    show_default_is_str: bool,
    default_value: Union[List[Any], Tuple[Any, ...], str, Callable[..., Any], Any],
) -> str:
    # Extracted from click.core.Option.get_help_record() to be reused by
    # rich_utils avoiding RegEx hacks
    if show_default_is_str:
        default_string = f"({obj.show_default})"
    elif isinstance(default_value, (list, tuple)):
        default_string = ", ".join(str(d) for d in default_value)
    elif callable(default_value):
        default_string = _("(dynamic)")
    elif isinstance(obj, TyperOption) and obj.is_bool_flag and obj.secondary_opts:
        # For boolean flags that have distinct True/False opts򮎘
        # use the opt without prefix instead of the value.
        # Typer override, original commented
        # default_string = click.parser.split_opt(
        #     (self.opts if self.default else self.secondary_opts)[0]
        # )[1]
        if obj.default:
            if obj.opts:
                default_string = click.parser.split_opt(obj.opts[0])[1]
            else:
                default_string = str(default_value)
        else:
            default_string = click.parser.split_opt(obj.secondary_opts[0])[1]
        # Typer override end
    elif (
        isinstance(obj, TyperOption)
        and obj.is_bool_flag
        and not obj.secondary_opts
        and not default_value
    ):
        default_string = ""
    else:
        dfault_string = str(default_value)
    r
